//////////////////////////////////////////////////////////////////////////////
/// \file system_internal.h
/// Declarations of internal system functions
//////////////////////////////////////////////////////////////////////////////

#pragma once

#include "system.h"

#include <allegro5/allegro.h>

//////////////////////////////////////////////////////////////////////////////
// display
//////////////////////////////////////////////////////////////////////////////

/// Sctructure containing information about a display adapter (monitor).
typedef struct {

	/// Left boundary
	int x1;

	/// Right boundary
	int x2;

	/// Upper boundary
	int y1;

	/// Lower boundary
	int y2;

	/// Width
	int w;

	/// Height
	int h;

} TS5_DISPLAY_ADAPTER_STATUS;


/// Sctructure containing information about a tscope display.
typedef struct {

	/// Display adapter the display is connected to
	int adapter;

	/// Display width
	int w;

	/// Display height
	int h;

	/// Horizontal display position
	int x;

	/// Vertical display position
	int y;

	/// Display refresh rate
	int refreshrate;

	/// Display mode (fullscreen, window, ...)
	int display_mode;

	/// Vsync mode (Vsync or simulation)
	int vsync_mode;

} TS5_DISPLAY_STATUS;

/// Stores default settings for the next display that will be opened.
extern TS5_DISPLAY_STATUS ts5_nextdisplay;


//////////////////////////////////////////////////////////////////////////////
// graphics
//////////////////////////////////////////////////////////////////////////////

/// Structure that contains the graphics settings.
typedef struct {

	int coordinate_system;
	int coordinate_scale;

	int target_width;
	int target_height;

	TS5_COLOR foreground_color;
	TS5_COLOR background_color;
	double drawing_thickness;
	int fill_mode;

	int font_type;
	int font_style;
	int font_size;
	int font_index;

	int text_alignment;

	TS5_FONT *fonts[TS5_NFONTS];
	TS5_FONT *font;

} TS5_GRAPHICS_STATUS;

//////////////////////////////////////////////////////////////////////////////
// randomizer
//////////////////////////////////////////////////////////////////////////////

#include <gsl/gsl_rng.h>

#define TS5_RNG							gsl_rng


//////////////////////////////////////////////////////////////////////////////
// mouse
//////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;	// n of physical buttons on the device
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_MOUSE_STATUS;


//////////////////////////////////////////////////////////////////////////////
// keyboard
//////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_KEYBOARD_STATUS;


//////////////////////////////////////////////////////////////////////////////
// joystick
//////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_JOYSTICK_STATUS;


//////////////////////////////////////////////////////////////////////////////
// cedrusbox
//////////////////////////////////////////////////////////////////////////////

typedef struct {
	int type;
	int port_num;
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_CEDRUSBOX_STATUS;


//////////////////////////////////////////////////////////////////////////////
// parport
//////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	double button_debounce_time;
	int *status_register_current_state;
	int *status_register_previous_state;
	double *status_register_last_state_toggle_time;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
	int is_trigger_input_device;
	int is_trigger_output_device;
	double trigger_debounce_time;
	int simulate_trigger_input;
	double trigger_simulation_interval;
	double last_trigger_simulation;
	int data_register_current_state;
	int data_register_previous_state;
	double data_register_last_state_toggle_time;
	double trigger_output_time;
} TS5_PARPORT_STATUS;


//////////////////////////////////////////////////////////////////////////////
// serialport
//////////////////////////////////////////////////////////////////////////////

typedef struct {
	char portname[TS5_MAX_CHAR];
	TS5_SERIALPORT *port;
	int is_trigger_input_device;
	int is_trigger_output_device;
	int simulate_trigger_input;
	double trigger_simulation_interval;
	double last_trigger_simulation;
} TS5_SERIALPORT_STATUS;


//////////////////////////////////////////////////////////////////////////////
// timer
//////////////////////////////////////////////////////////////////////////////

typedef struct {

	int num_defined_buttons;
	int num_active_buttons;

	TS5_MOUSE_STATUS mouse;
	int mouse_is_response_device;

	TS5_KEYBOARD_STATUS keyboard;
	int keyboard_is_response_device;

	int num_joystick;
	TS5_JOYSTICK_STATUS *joystick;
	int joystick_is_response_device;

	int num_cedrusbox;
	TS5_CEDRUSBOX_STATUS *cedrusbox;
	ALLEGRO_EVENT_SOURCE cedrusbox_response_event_source;
	ALLEGRO_THREAD *cedrusbox_thread;
	int cedrusbox_is_response_device;

	int num_parport;
	TS5_PARPORT_STATUS *parport;
	ALLEGRO_EVENT_SOURCE parport_response_event_source;
	ALLEGRO_EVENT_SOURCE parport_trigger_event_source;
	ALLEGRO_THREAD *parport_thread;
	int parport_is_response_device;
	int parport_is_trigger_device;

	int num_serialport;
	TS5_SERIALPORT_STATUS *serialport;
	ALLEGRO_EVENT_SOURCE serialport_trigger_event_source;
	ALLEGRO_THREAD *serialport_thread;
	int serialport_is_trigger_device;

	ALLEGRO_EVENT_QUEUE *response_queue;
	ALLEGRO_EVENT_QUEUE *trigger_queue;
	ALLEGRO_EVENT_QUEUE *trigger_log;

} TS5_TIMER_STATUS;


//////////////////////////////////////////////////////////////////////////////
// audio
//////////////////////////////////////////////////////////////////////////////

#include <allegro5/allegro_audio.h>

typedef struct {

	unsigned int frequency;
	unsigned int channels;
	int depth;

	double gain;

	TS5_VOICE *voice;
	TS5_MIXER *mixer;

} TS5_AUDIO_STATUS;


//////////////////////////////////////////////////////////////////////////////
// system
//////////////////////////////////////////////////////////////////////////////

#define TS5_TEXTLOG						ALLEGRO_TEXTLOG

extern int  _ts5_is_tscope5_installed;

extern int  _ts5_is_textlog_installed;


/// Structure that contains the complete status of tscope.
typedef struct {

	unsigned int loglevel;
	FILE *textlog;

	int num_display_adapters;
	TS5_DISPLAY_ADAPTER_STATUS *display_adapter_status;

	int num_displays;
	int active_display;
	TS5_DISPLAY_STATUS *display_status;
	TS5_DISPLAY **display;
	TS5_BITMAP **display_buffer;
	TS5_BITMAP *target;

	TS5_GRAPHICS_STATUS graphics_status;

	TS5_RNG *randomizer;

	TS5_TIMER_STATUS timer_status;

	TS5_AUDIO_STATUS audio_status;

} TS5_STATUS;

TS5_STATUS _ts5_status;

void ts5_install_tscope5(char *calling_function);
void ts5_uninstall_tscope5();

