//////////////////////////////////////////////////////////////////////////////
/// \file audio.c
/// Definitions of audio functions
/// \example audio1.c read and play a sample from a file
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Audio parameters
/// The parameters of the audio output device (sound card) can be set
/// using the functions below. 
///
/// The audio gain (amplification factor) can be changed at any time. 
/// All other functions functions in this group have to be called before
/// the audio subsystem is loaded (i.e. before the first sample is loaded and played)
/// and their settings remain unchanged until the end of the program.
///
/// The parameter functions also define the settings for samples that
/// are created with ts5_alloc_sample().
//@{
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/audio.h"
#include "../include/tscope5/audio_internal.h"
#include "../include/tscope5/system_internal.h"


//////////////////////////////////////////////////////////////////////////////
/// Set the requested sample rate for the audio system.
///
/// \param frequency The sample rate. Can be TS5_11025, TS5_22050 or TS5_44100.
///
/// \return The previous requested sample rate.
///
/// This function has to be called before the (automatic) installation of the audio system.
///
/// Once the audio system is installed, the sample rate of the system cannot be changed.
///
/// This is only a request, the audio driver can choose non-matching values. Query the audio system
/// after installation with ts5_get_audio_frequency() for the actual sample rate.
///
/// This function also influences the sample rate of samples made by ts5_alloc_sample().
///
/// The default frequency is TS5_44100.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_set_audio_frequency(unsigned int frequency)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_set_audio_frequency");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_frequency(%u)\n", frequency);

	unsigned int retval = _ts5_status.audio_status.frequency;

	if (frequency!=TS5_11025 && frequency!=TS5_22050 && frequency!=TS5_44100) {
		ts5_fatal("ts5_set_audio_frequency: unknonwn frequency requested\n");
	}

	_ts5_status.audio_status.frequency = frequency;

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_frequency: set sample rate to %u (was %u)\n",
				_ts5_status.audio_status.frequency, retval);

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the sample rate of the audio system.
///
/// \return The sample rate of the audio system as an unsigned integer.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_get_audio_frequency()
{
	ts5_check_audio("ts5_get_audio_frequency");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_audio_frequency()\n");

	return al_get_voice_frequency(_ts5_status.audio_status.voice);
}


//////////////////////////////////////////////////////////////////////////////
/// Set the requested number of channels for the audio system.
///
/// \param channels The number of channels. Can be TS5_MONO or TS5_STEREO.
///
/// \return The previous requested number of channels.
///
/// This function has to be called before the (automatic) installation of the audio system.
///
/// Once the audio system is installed, the number of channels of the system cannot be changed.
///
/// This is only a request, the audio driver can choose non-matching values. Query the audio system
/// after installation with ts5_get_audio_channels() for the actual number of channels.
///
/// This function also influences the number of channels of samples made by ts5_alloc_sample().
///
/// The default number of channels is TS5_STEREO.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_set_audio_channels(unsigned int channels)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_set_audio_channels");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_channels(%d)\n", channels);

	unsigned int retval = _ts5_status.audio_status.channels;

	if (channels!=TS5_MONO && channels!=TS5_STEREO) {
		ts5_fatal("ts5_set_audio_channels: unknonwn number of channels requested\n");
	}

	_ts5_status.audio_status.channels = channels;

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_channels: set number of channels to %u (was %u)\n",
				_ts5_status.audio_status.channels, retval);

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the number of channels of the audio system.
///
/// \return The number of channels of the audio system.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_get_audio_channels()
{
	ts5_check_audio("ts5_get_audio_channels");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_audio_channels()\n");

	ALLEGRO_CHANNEL_CONF channels = al_get_voice_channels(_ts5_status.audio_status.voice);

	unsigned int retval;

	if (channels == ALLEGRO_CHANNEL_CONF_1) {
		retval =  TS5_MONO;
	}
	else if (channels == ALLEGRO_CHANNEL_CONF_2) {
		retval =  TS5_STEREO;
	}
	else {
		ts5_fatal("ts5_get_audio_channels: got an unknonwn number of channels\n");
		retval =  0;
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the sample depth for the audio system.
///
/// \param depth The sample depth. Can be TS5_INTEGER or TS5_FLOAT.
///
/// \return The previous requested sample depth.
///
/// This function has to be called before the (automatic) installation of the audio system.
///
/// Once the audio system is installed, the sample depth of the system cannot be changed.
///
/// This is only a request, the audio driver can choose non-matching values. Query the audio system
/// after installation with ts5_get_audio_depth() for the actual sample depth.
///
/// This function also influences the sample depth of samples made by ts5_alloc_sample().
///
/// The default sample depth is TS5_FLOAT.
///
/// \todo currently has no effect. The voice will be integer and the mixer floating point.
//////////////////////////////////////////////////////////////////////////////
int ts5_set_audio_depth(int depth)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_set_audio_depth");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_depth(%d)\n", depth);

	int retval = _ts5_status.audio_status.depth;

	if (depth!=TS5_INTEGER && depth!=TS5_FLOAT) {
		ts5_fatal("ts5_set_audio_depth: unknonwn sample depth requested\n");
	}

	_ts5_status.audio_status.depth = depth;

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_depth: set sample depth to %d (was %d)\n",
				_ts5_status.audio_status.depth, retval);

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the sample depth of the audio system.
///
/// \return The sample depth of the audio system.
//////////////////////////////////////////////////////////////////////////////
int ts5_get_audio_depth()
{
	ts5_check_audio("ts5_get_audio_depth");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_audio_depth()\n");

	ALLEGRO_AUDIO_DEPTH depth = al_get_voice_depth(_ts5_status.audio_status.voice);

	if (depth != TS5_INTEGER && depth !=TS5_FLOAT) {
		ts5_fatal("ts5_get_audio_depth: got an unknonwn sample depth\n");
	}

	return depth;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the gain (amplification factor) of the audio system.
///
/// \param gain The gain. 0.0 = silent, 0.5=half, 1.0 = untouched, 2.0 = double, ...
///
/// \return The previous gain.
///
/// The default gain is 1.0.
//////////////////////////////////////////////////////////////////////////////
double ts5_set_audio_gain(double gain)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_set_audio_gain");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_gain(%f)\n", gain);

	double retval = _ts5_status.audio_status.gain;

	if (gain<0.0) {
		ts5_fatal("ts5_set_audio_gain: gain should be >= 0.0\n");
	}

	_ts5_status.audio_status.gain = gain;

	if (!al_set_mixer_gain(_ts5_status.audio_status.mixer, gain)) {
		ts5_fatal("ts5_set_audio_gain: could not change audio gain\n");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_audio_gain: set gain to %f (was %f)\n",
				_ts5_status.audio_status.gain, retval);

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the gain of the audio system.
///
/// \return The gain of the audio system.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_audio_gain()
{
	ts5_check_audio("ts5_get_audio_gain");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_audio_gain()\n");

	return _ts5_status.audio_status.gain;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Loading samples 
/// Samples can either be read from a file, or empty samples can be generated.
/// 
/// Empty samples can be used for recording vocal responses (tbd).
///
/// Most samples (samples that contain less than 2^24 sample values) 
/// can be read completely into memory using ts5_read_sample(). 
/// Longer samples should be loaded with ts5_read_long_sample(). 
/// These will be streamed (i.e. they will be divided in parts that are read when necessary).
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Create an empty sample.
///
/// \param length Length of the sample in seconds.
///
/// \return A pointer to the newly created sample.
///
/// Sample frequency, channels and depth are set by ts5_set_audio_frequency(),
/// ts5_set_audio_channels() and ts5_set_audio_depth()
//////////////////////////////////////////////////////////////////////////////
TS5_SAMPLE *ts5_alloc_sample(double length)
{
	ts5_check_audio("ts5_alloc_sample");
	ts5_log(TS5_LOGLEVEL_2, "ts5_alloc_sample(%f)\n", length);

	unsigned int samples = length * _ts5_status.audio_status.frequency;

	ALLEGRO_CHANNEL_CONF channels;

	if (_ts5_status.audio_status.channels==TS5_MONO) {
		channels=ALLEGRO_CHANNEL_CONF_1;
	}
	else if (_ts5_status.audio_status.channels==TS5_STEREO) {
		channels=ALLEGRO_CHANNEL_CONF_2;
	}
	else {
		ts5_fatal("ts5_alloc_sample: unknonwn channel configuration requested\n");
	}

	unsigned int sample_size = al_get_channel_count(channels) *
								al_get_audio_depth_size(_ts5_status.audio_status.depth);

	unsigned int bytes = samples * sample_size;

	void *buff = NULL;
	buff = al_malloc(bytes);

	if (!buff) {
		ts5_fatal("ts5_alloc_sample: could not allocate sample buffer\n");
	}

	TS5_SAMPLE *sample = NULL;
	sample = (TS5_SAMPLE *)al_malloc(sizeof(TS5_SAMPLE));

	if (!sample) {
		ts5_fatal("ts5_alloc_sample: could not allocate sample\n");
	}

	sample->sample = al_create_sample(buff, samples, _ts5_status.audio_status.frequency,
						_ts5_status.audio_status.depth, channels, true);

	if (!sample->sample) {
		ts5_fatal("ts5_alloc_sample: could not create sample\n");
	}

	sample->instance = al_create_sample_instance(NULL);

	if (!sample->instance) {
		ts5_fatal("ts5_alloc_sample: could not create sample instance\n");
	}

	if (!al_set_sample(sample->instance, sample->sample)) {
  		ts5_fatal("ts5_alloc_sample: could not attach sample to sample instance\n");
  	}

	if (!al_attach_sample_instance_to_mixer(sample->instance, _ts5_status.audio_status.mixer)) {
  		ts5_fatal("ts5_alloc_sample: could not attach sample instance to mixer\n");
	}

	al_set_sample_instance_playing(sample->instance, false);

	sample->issample = true;
	sample->isfilestream = false;
	sample->ismemorystream = false;
	sample->stream = NULL;

	return sample;
}


//////////////////////////////////////////////////////////////////////////////
/// Open a sample from a file.
///
/// \param file Path to the sample file.
///
/// \return A pointer to the newly created sample.
///
/// Available sample types are: wav, flac, ogg, it, mod, s3m, xm.
///
/// The file is read from disk all at once.
//////////////////////////////////////////////////////////////////////////////
TS5_SAMPLE *ts5_read_sample(const char *file)
{
	ts5_check_audio("ts5_read_sample");
	ts5_log(TS5_LOGLEVEL_2, "ts5_read_sample(%s)\n", file);

	TS5_SAMPLE *sample = NULL;
	sample = (TS5_SAMPLE *)al_malloc(sizeof(TS5_SAMPLE));

	if (!sample) {
		ts5_fatal("ts5_read_sample: could not allocate sample\n");
	}

	sample->issample = true;
	sample->isfilestream = false;
	sample->ismemorystream = false;

	sample->sample = NULL;
	sample->sample = al_load_sample(file);

	if (!sample->sample) {
		ts5_fatal("ts5_read_sample: could not read sample\n");
	}

	sample->instance = NULL;
	sample->instance = al_create_sample_instance(NULL);

	if (!sample->instance) {
		ts5_fatal("ts5_read_sample: could not create sample instance\n");
	}

	if (!al_set_sample(sample->instance, sample->sample)) {
  		ts5_fatal("ts5_read_sample: could not attach sample to sample instance\n");
  	}

	if (!al_attach_sample_instance_to_mixer(sample->instance, _ts5_status.audio_status.mixer)) {
  		ts5_fatal("ts5_read_sample: could not attach sample instance to mixer\n");
	}

	al_set_sample_instance_playing(sample->instance, false);

	sample->stream = NULL;

	return sample;
}


//////////////////////////////////////////////////////////////////////////////
/// Open a long sample from a file.
///
/// \param file Path to the sample file.
///
/// \return A pointer to the newly created sample.
///
/// Available sample types are: wav, flac, ogg, it, mod, s3m, xm.
///
/// The file is read from disk as it is needed.
//////////////////////////////////////////////////////////////////////////////
TS5_SAMPLE *ts5_read_long_sample(const char *file)
{
	ts5_check_audio("ts5_read_long_sample");
	ts5_log(TS5_LOGLEVEL_2, "ts5_read_long_sample(%s)\n", file);

	TS5_SAMPLE *sample = NULL;
	sample = (TS5_SAMPLE *)al_malloc(sizeof(TS5_SAMPLE));

	if (!sample) {
		ts5_fatal("ts5_read_long_sample: could not allocate sample\n");
	}

	sample->issample = false;
	sample->isfilestream = true;
	sample->ismemorystream = false;

	sample->sample = NULL;
	sample->instance = NULL;

	sample->stream = NULL;
	sample->stream = al_load_audio_stream(file, 4, 2048);

	if (!sample->stream) {
		ts5_fatal("ts5_read_long_sample: could not read sample\n");
	}

	if (!al_attach_audio_stream_to_mixer(sample->stream, _ts5_status.audio_status.mixer)) {
  		ts5_fatal("ts5_read_long_sample: could not attach sample stream to mixer\n");
	}

	al_set_audio_stream_playing(sample->stream, false);
	al_set_audio_stream_playmode(sample->stream, ALLEGRO_PLAYMODE_ONCE);

	return sample;
}


//////////////////////////////////////////////////////////////////////////////
/// Write a sample.
///
/// \param file Path to the sample file.
/// \param sample Pointer to the sample that will be written.
///
/// Only .wav is supported. The file extension should be .wav
//////////////////////////////////////////////////////////////////////////////
void ts5_write_sample(const char *file, TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_write_sample");
	ts5_log(TS5_LOGLEVEL_2, "ts5_write_sample(%s,%p)\n", file, sample);

	if (!sample->issample) {
		ts5_fatal("ts5_write_sample: writing of streams is not implemented yet\n");
	}

	if (!al_save_sample(file, sample->sample)) {
		ts5_fatal("ts5_write_sample: could not write sample\n");
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Free the memory used by a sample.
///
/// \param sample Pointer to the sample that will be freed.
///
/// This function should be called at the end of the program for
/// each sample allocated or read by the user.
//////////////////////////////////////////////////////////////////////////////
void ts5_free_sample(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_free_sample");
	ts5_log(TS5_LOGLEVEL_2, "ts5_free_sample(%p)\n", sample);

	if (sample->issample) {
		al_set_sample(sample->instance, NULL);
		al_destroy_sample(sample->sample);

		al_detach_sample_instance(sample->instance);
	 	al_destroy_sample_instance(sample->instance);
	}
	else if (sample->isfilestream) {
		al_detach_audio_stream(sample->stream);
		al_destroy_audio_stream(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_free_sample: memory streams not implemented yet\n");
	}

	al_free(sample);
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Playing samples
/// Audio samples can be played using the functions below. 
///
/// Playing a sample will not block the flow of your program.
/// Once a sample starts playing the function call following
/// ts5_play_sample() will be executed immediately.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Start playing a sample.
///
/// \param sample Pointer to the sample that will be played.
//////////////////////////////////////////////////////////////////////////////
void ts5_play_sample(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_play_sample");
	ts5_log(TS5_LOGLEVEL_5, "ts5_play_sample(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_play_sample: sample pointer is null\n");
	}

	if (sample->issample) {

		if (!al_set_sample_instance_playing(sample->instance, 1)) {
			ts5_fatal("ts5_play_sample: could not play sample\n");
		}
	}
	else if (sample->isfilestream) {

		if (!al_set_audio_stream_playing(sample->stream, 1)) {
			ts5_fatal("ts5_play_sample: could not play sample\n");
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_play_sample: memory streams not implemented yet\n");
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Pause playing a sample.
///
/// \param sample Pointer to the sample that will be paused.
//////////////////////////////////////////////////////////////////////////////
void ts5_pause_sample(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_pause_sample");
	ts5_log(TS5_LOGLEVEL_5, "ts5_pause_sample(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_pause_sample: sample pointer is null\n");
	}

	if (sample->issample) {

		unsigned int position = al_get_sample_instance_position(sample->instance);

		if (!al_set_sample_instance_playing(sample->instance, 0)) {
			ts5_fatal("ts5_pause_sample: could not pause sample\n");
		}

		if (!al_set_sample_instance_position(sample->instance, position)) {
		  	ts5_fatal("ts5_pause_sample: could not change sample position\n");
		}
	}
	else if (sample->isfilestream) {

		if (!al_set_audio_stream_playing(sample->stream, 0)) {
			ts5_fatal("ts5_pause_sample: could not pause sample\n");
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("memory streams not implemented yet\n");
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Stop playing a sample.
///
/// \param sample Pointer to the sample that will be stopped.
//////////////////////////////////////////////////////////////////////////////
void ts5_stop_sample(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_stop_sample");
	ts5_log(TS5_LOGLEVEL_5, "ts5_stop_sample(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_stop_sample: sample pointer is null\n");
	}

	if (sample->issample) {

		if (!al_set_sample_instance_playing(sample->instance, 0)) {
			ts5_fatal("ts5_stop_sample: could not stop sample\n");
		}

		if (!al_set_sample_instance_position(sample->instance, 0)) {
			ts5_fatal("ts5_stop_sample: could not stop sample\n");
		}
	}
	else if (sample->isfilestream) {

		al_drain_audio_stream(sample->stream);

		if (!al_rewind_audio_stream(sample->stream)) {
			ts5_fatal("ts5_stop_sample: could not rewind sample\n");
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_stop_sample: memory streams not implemented yet\n");
	}
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Sample parameters
/// For each sample the parameters can be queried/set using the functions below:
///
/// - Sample position, speed, gain, pan and playmode can be adjusted at any time, 
/// even when the sample is playing.
/// - Sample frequency, channels, depth and length are defined when the 
/// sample is read or allocated.  
///
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Query whether a sample is playing.
///
/// \param sample Pointer to the sample that will queried.
///
/// \return 1 if the sample is playing, 0 if not.
//////////////////////////////////////////////////////////////////////////////
int ts5_get_sample_status(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_status");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_sample_status(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_status: sample pointer is null\n");
	}

	int retval = 0;

	if (sample->issample) {
		retval =  al_get_sample_instance_playing(sample->instance);
	}
	else if (sample->isfilestream) {
		retval =  al_get_audio_stream_playing(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_status: memory streams not implemented yet\n");
	}
	else {
		ts5_fatal("ts5_get_sample_status: sample type unknown\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the sample rate of a sample.
///
/// \param sample Pointer to the sample that will queried.
///
/// \return The sample rate of the sample.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_get_sample_frequency(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_frequency");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_frequency(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_frequency: sample pointer is null\n");
	}

	unsigned int retval = 0;

	if (sample->issample) {
		retval =  al_get_sample_frequency(sample->sample);
	}
	else if (sample->isfilestream) {
		retval =  al_get_audio_stream_frequency(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_frequency: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the number of channels of a sample.
///
/// \param sample Pointer to the sample that will queried.
///
/// \return The number of channels of the sample.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_get_sample_channels(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_channels");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_channels(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_channels: sample pointer is null\n");
	}

	ALLEGRO_CHANNEL_CONF channels;

	if (sample->issample) {
		channels = al_get_sample_channels(sample->sample);
	}
	else if (sample->isfilestream) {
		channels = al_get_audio_stream_channels(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_channels: memory streams not implemented yet\n");
	}

	unsigned int retval = 0;

	if (channels == ALLEGRO_CHANNEL_CONF_1) {
		retval =  TS5_MONO;
	}
	else if (channels == ALLEGRO_CHANNEL_CONF_2) {
		retval =  TS5_STEREO;
	}
	else {
		ts5_fatal("ts5_get_sample_channels: got an unknonwn number of channels\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the sample depth of a sample.
///
/// \param sample Pointer to the sample that will queried.
///
/// \return The sample depth of a sample.
//////////////////////////////////////////////////////////////////////////////
int ts5_get_sample_depth(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_depth");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_depth(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_depth: sample pointer is null\n");
	}

	ALLEGRO_AUDIO_DEPTH depth;

	if (sample->issample) {
		depth = al_get_sample_depth(sample->sample);
	}
	else if (sample->isfilestream) {
		depth = al_get_audio_stream_depth(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_depth: memory streams not implemented yet\n");
	}

	int retval = 0;

	if (depth == ALLEGRO_AUDIO_DEPTH_INT16) {
		retval =  TS5_INTEGER;
	}
	else if (depth == ALLEGRO_AUDIO_DEPTH_FLOAT32) {
		retval =  TS5_FLOAT;
	}
	else {
		ts5_fatal("ts5_get_audio_depth: got an unknonwn sample depth\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the length of a sample.
///
/// \param sample Pointer to the sample that will queried.
///
/// \return The length of the sample.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_get_sample_length(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_length");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_length(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_length: sample pointer is null\n");
	}

	unsigned int retval = 0;

	if (sample->issample) {
		retval =  al_get_sample_length(sample->sample);
	}
	else if (sample->isfilestream) {
		retval =  al_get_audio_stream_length(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_length: memory streams not implemented yet\n");
	}

	return retval;
}



//////////////////////////////////////////////////////////////////////////////
/// Set the playback position of a sample.
///
/// \param sample Pointer to the sample that will be adjusted.
/// \param position The new playback position.
///
/// \return The old playback position.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_set_sample_position(TS5_SAMPLE *sample, unsigned int position)
{
	ts5_check_audio("ts5_set_sample_position");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_sample_position(%p,%u)\n", sample, position);

	if (!sample) {
		ts5_fatal("ts5_set_sample_position: sample pointer is null\n");
	}

	unsigned int retval= 0;

	if (sample->issample) {

		unsigned int max = al_get_sample_length(sample->sample);

		if (position > max) {
			ts5_fatal("ts5_set_sample_position: position should be",
						" <= the length of the sample (%d)\n", max);
		}

		retval =  al_get_sample_instance_position(sample->instance);

		if (!(al_set_sample_instance_position(sample->instance, position))) {
			ts5_fatal("ts5_set_sample_position: could not change sample position\n");
		}
	}
	else if (sample->isfilestream) {

		unsigned int freq = al_get_audio_stream_frequency(sample->stream);
		unsigned int max = al_get_audio_stream_length_secs(sample->stream) * freq;

		if (position > max) {
			ts5_fatal("ts5_set_sample_position: position should be",
						" <= the length of the sample (%d)\n", max);
		}

		retval =  al_get_audio_stream_position_secs(sample->stream) * freq;

		if (!(al_seek_audio_stream_secs(sample->stream, position / freq))) {
			ts5_fatal("ts5_set_sample_position: could not change sample position\n");
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_set_sample_position: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the playback position of a sample.
///
/// \param sample Pointer to the sample that will be queried.
///
/// \return The current playback position.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_get_sample_position(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_position");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_position(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_length: sample pointer is null\n");
	}

	unsigned int retval = 0;

	if (sample->issample) {
		retval =  al_get_sample_instance_position(sample->instance);
	}
	else if (sample->isfilestream) {
		unsigned int freq = al_get_audio_stream_frequency(sample->stream);
		retval =  al_get_audio_stream_position_secs(sample->stream) * freq;
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_position: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the playback speed of a sample.
///
/// \param sample Pointer to the sample that will be adjusted.
/// \param speed The new playback speed (1.0 is normal speed).
///
/// \return The old playback speed.
//////////////////////////////////////////////////////////////////////////////
double ts5_set_sample_speed(TS5_SAMPLE *sample, double speed)
{
	ts5_check_audio("ts5_set_sample_speed");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_sample_speed(%p,%f)\n", sample, speed);

	if (!sample) {
		ts5_fatal("ts5_set_sample_speed: sample pointer is null\n");
	}

	if (speed < 0) {
		ts5_fatal("ts5_set_sample_speed: speed should be >0 (is %f)\n", speed);
	}

	double retval =  0.0;

	if (sample->issample) {

		retval =  al_get_sample_instance_speed(sample->instance);

		if (!(al_set_sample_instance_speed(sample->instance, speed))) {
			ts5_fatal("ts5_set_sample_speed: could not change sample speed (%f)\n", speed);
		}
	}
	else if (sample->isfilestream) {

		retval =  al_get_audio_stream_speed(sample->stream);

		if (!(al_set_audio_stream_speed(sample->stream, speed))) {
			ts5_fatal("ts5_set_sample_speed: could not change sample speed (%f)\n", speed);
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the playback speed of a sample.
///
/// \param sample Pointer to the sample that will be queried.
///
/// \return The current playback speed.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_sample_speed(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_speed");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_speed(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_speed: sample pointer is null\n");
	}

	double retval =  0.0;

	if (sample->issample) {
		retval =  al_get_sample_instance_speed(sample->instance);
	}
	else if (sample->isfilestream) {
		retval =  al_get_audio_stream_speed(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_speed: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the playback gain of a sample.
///
/// \param sample Pointer to the sample that will be adjusted.
/// \param gain The new playback gain (1.0 is normal gain).
///
/// \return The old gain position.
//////////////////////////////////////////////////////////////////////////////
double ts5_set_sample_gain(TS5_SAMPLE *sample, double gain)
{
	ts5_check_audio("ts5_set_sample_gain");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_sample_gain(%p,%f)\n", sample, gain);

	if (!sample) {
		ts5_fatal("ts5_set_sample_gain: sample pointer is null\n");
	}

	if (gain < 0) {
		ts5_fatal("ts5_set_sample_gain: gain should be >0 (is %f)\n", gain);
	}

	double retval =  0.0;

	if (sample->issample) {

		retval =  al_get_sample_instance_gain(sample->instance);

		if (!(al_set_sample_instance_gain(sample->instance, gain))) {
			ts5_fatal("ts5_set_sample_gain: could not change sample gain (%f)\n", gain);
		}
	}
	else if (sample->isfilestream) {

		retval =  al_get_audio_stream_gain(sample->stream);

		if (!(al_set_audio_stream_gain(sample->stream, gain))) {
			ts5_fatal("ts5_set_sample_gain: could not change sample gain (%f)\n", gain);
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the playback gain of a sample.
///
/// \param sample Pointer to the sample that will be queried.
///
/// \return The current playback gain.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_sample_gain(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_gain");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_gain(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_gain: sample pointer is null\n");
	}

	double retval =  0.0;

	if (sample->issample) {
		retval =  al_get_sample_instance_gain(sample->instance);
	}
	else if (sample->isfilestream) {
		retval =  al_get_audio_stream_gain(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_gain: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the playback pan of a sample.
///
/// \param sample Pointer to the sample that will be adjusted.
/// \param pan The new playback pan  (-1.0 is left, 0.0 is normal, 1.0 is right).
///
/// \return The old playback pan.
//////////////////////////////////////////////////////////////////////////////
double ts5_set_sample_pan(TS5_SAMPLE *sample, double pan)
{
	ts5_check_audio("ts5_set_sample_pan");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_sample_pan(%p,%f)\n", sample, pan);

	if (!sample) {
		ts5_fatal("ts5_set_sample_pan: sample pointer is null\n");
	}

	if (pan < -1.0 || pan > 1) {
		ts5_fatal("ts5_set_sample_pan: pan should be between -1.0 and 1.0 (is %f)\n", pan);
	}

	double retval =  0.0;

	if (sample->issample) {

		retval =  al_get_sample_instance_pan(sample->instance);

		if (!(al_set_sample_instance_pan(sample->instance, pan))) {
			ts5_fatal("ts5_set_sample_pan: could not change sample pan (%f)\n", pan);
		}
	}
	else if (sample->isfilestream) {

		retval =  al_get_audio_stream_pan(sample->stream);

		if (!(al_set_audio_stream_pan(sample->stream, pan))) {
			ts5_fatal("ts5_set_sample_pan: could not change sample pan (%f)\n", pan);
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_set_sample_pan: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the playback pan of a sample.
///
/// \param sample Pointer to the sample that will be queried.
///
/// \return The current playback pan.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_sample_pan(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_pan");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_pan(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_pan: sample pointer is null\n");
	}

	double retval =  0.0;

	if (sample->issample) {
		retval =  al_get_sample_instance_pan(sample->instance);
	}
	else if (sample->isfilestream) {
		retval =  al_get_audio_stream_pan(sample->stream);
	}
	else if (sample->ismemorystream) {
		ts5_fatal("memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the playmode of a sample.
///
/// \param sample Pointer to the sample that will be adjusted.
/// \param playmode The new playmode. Can be TS5_PLAY_ONCE or TS5_PLAY_LOOP.
///
/// \return The old playmode.
//////////////////////////////////////////////////////////////////////////////
int ts5_set_sample_playmode(TS5_SAMPLE *sample, int playmode)
{
	ts5_check_audio("ts5_set_sample_playmode");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_sample_playmode(%p,%d)\n", sample, playmode);

	if (!sample) {
		ts5_fatal("ts5_set_sample_playmode: sample pointer is null\n");
	}

	if (playmode != TS5_PLAY_ONCE && playmode != TS5_PLAY_LOOP) {
		ts5_fatal("ts5_set_sample_playmode: playmode should be TS5_PLAY_ONCE or TS5_PLAY_LOOP\n");
	}

	int retval = 0;

	if (sample->issample) {

		retval =  al_get_sample_instance_playmode(sample->instance);

		if (!(al_set_sample_instance_playmode(sample->instance, playmode))) {
			ts5_fatal("ts5_set_sample_playmode: could not change sample playmode\n");
		}
	}
	else if (sample->isfilestream) {

		retval =  al_get_audio_stream_playmode(sample->stream) - 3;

		if (!(al_set_audio_stream_playmode(sample->stream, playmode))) {
			ts5_fatal("ts5_set_sample_playmode: could not change sample playmode\n");
		}
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_set_sample_playmode: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the playmode of a sample.
///
/// \param sample Pointer to the sample that will be queried.
///
/// \return The current playmode.
//////////////////////////////////////////////////////////////////////////////
int ts5_get_sample_playmode(TS5_SAMPLE *sample)
{
	ts5_check_audio("ts5_get_sample_playmode");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_sample_playmode(%p)\n", sample);

	if (!sample) {
		ts5_fatal("ts5_get_sample_playmode: sample pointer is null\n");
	}

	int retval = 0;

	if (sample->issample) {
		retval =  al_get_sample_instance_playmode(sample->instance);
	}
	else if (sample->isfilestream) {
		retval =  al_get_audio_stream_playmode(sample->stream) - 3;
	}
	else if (sample->ismemorystream) {
		ts5_fatal("ts5_get_sample_playmode: memory streams not implemented yet\n");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////

