//////////////////////////////////////////////////////////////////////////////
/// \file audio_internal.c
/// Definitions of internal audio functions
//////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/audio_internal.h"
#include "../include/tscope5/system_internal.h"

#include <allegro5/allegro_acodec.h>

/// Is the audio subsystem installed?
int  _ts5_is_audio_installed = 0;


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each audio function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether the audio subsystem is installed. If not, the audio subsystem is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_audio(char *calling_function)
{
	if (!_ts5_is_audio_installed) {
		ts5_install_audio(calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_audio\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Install the audio subsystem.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// This function is called automatically if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_install_audio(char *calling_function)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5(calling_function);
	}

	if (!_ts5_is_audio_installed) {

		_ts5_is_audio_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 audio\n", calling_function);

		if (!al_install_audio()) {
				ts5_fatal("%s: could not install Tscope5 audio\n", calling_function);
		}

		if (!al_init_acodec_addon()) {
				ts5_fatal("%s: could not install Tscope5 audio codec addon\n", calling_function);
		}

		ALLEGRO_CHANNEL_CONF channels;

		if (_ts5_status.audio_status.channels==TS5_MONO) {
			channels=ALLEGRO_CHANNEL_CONF_1;
		}
		else if (_ts5_status.audio_status.channels==TS5_STEREO) {
			channels=ALLEGRO_CHANNEL_CONF_2;
		}
		else {
			ts5_fatal("%s: unknonwn channel configuration requested\n", calling_function);
		}

		/*
		ALLEGRO_AUDIO_DEPTH depth;
		if (_ts5_status.audio_status.depth==TS5_INTEGER) {
			depth=ALLEGRO_AUDIO_DEPTH_INT16;
		}
		else if (_ts5_status.audio_status.depth==TS5_FLOAT) {
			depth=ALLEGRO_AUDIO_DEPTH_FLOAT32;
		}
		else {
			ts5_fatal("ts5_install_audio: unknonwn sample depth requested\n");
		}
		*/

		_ts5_status.audio_status.voice = al_create_voice(_ts5_status.audio_status.frequency,
											ALLEGRO_AUDIO_DEPTH_INT16, channels);

		if (!_ts5_status.audio_status.voice) {
			ts5_fatal("%s: could not create voice\n", calling_function);
		}

		_ts5_status.audio_status.mixer = al_create_mixer(_ts5_status.audio_status.frequency,
											ALLEGRO_AUDIO_DEPTH_FLOAT32, channels);

		if (!_ts5_status.audio_status.mixer) {
			ts5_fatal("%s: could not create mixer\n", calling_function);
		}

		if (!al_attach_mixer_to_voice(_ts5_status.audio_status.mixer,
				_ts5_status.audio_status.voice)) {
			ts5_fatal("%s: could not attach mixer to voice\n", calling_function);
		}

		atexit(ts5_uninstall_audio);
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Uninstall the audio subsystem.
///
/// This function is called automatically at the end of the program.
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_audio()
{
	if (_ts5_is_audio_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 audio\n");

		al_detach_mixer(_ts5_status.audio_status.mixer);
		al_detach_voice(_ts5_status.audio_status.voice);

		al_destroy_mixer(_ts5_status.audio_status.mixer);
		al_destroy_voice(_ts5_status.audio_status.voice);

		_ts5_status.audio_status.mixer = NULL;
		_ts5_status.audio_status.voice = NULL;

		al_uninstall_audio();

		_ts5_is_audio_installed = 0;
	}
}
