//////////////////////////////////////////////////////////////////////////////
/// \file bitmaps.c 
/// Definitions of bitmaps functions
/// \example bitmaps1.c read a bitmap from a file, default coordinates
/// \example bitmaps2.c read a bitmap from a file, relative coordinate scale
/// \example bitmaps3.c read a bitmap from a file, display coordinate system
/// \example bitmaps4.c read a bitmap from a file, relative coordinate scale + display coordinate system
/// \example bitmaps5.c prepare stimuli using memory bitmaps
//////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/bitmaps.h"
#include "../include/tscope5/display.h"
#include "../include/tscope5/graphics.h"
#include "../include/tscope5/bitmaps_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/graphics_internal.h"


//////////////////////////////////////////////////////////////////////////////
/// @name Loading bitmaps
/// Bitmaps can either be read from a file, or empty bitmaps can be generated.
/// 
/// Text/primitives output that is normally drawn on the display can
/// be redirected to a bitmap using ts5_set_drawing_target().
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Create a bitmap.
///
/// \param w Width of the bitmap.
/// \param h Height of the bitmap.
///
/// \return A pointer to the newly created bitmap.
///
/// The default background color of the newly created bitmap is the same as the display.
//////////////////////////////////////////////////////////////////////////////
TS5_BITMAP *ts5_alloc_bitmap(double w, double h)
{
	ts5_check_bitmaps("ts5_alloc_bitmap");
	ts5_log(TS5_LOGLEVEL_2, "ts5_alloc_bitmap(%f,%f)\n", w, h);

	if (_ts5_status.graphics_status.coordinate_scale==TS5_RELATIVE_COORDINATES) {
		w *= _ts5_status.graphics_status.target_width;
		h *= _ts5_status.graphics_status.target_height;
	}

	TS5_BITMAP *map;
	map = al_create_bitmap(w, h);

	if (map == NULL) {
		ts5_fatal("ts5_alloc_bitmap: could not make bitmap\n");
	}

	TS5_BITMAP *oldval = ts5_set_drawing_target(map);
	ts5_clear_display();
	ts5_set_drawing_target(oldval);

	return map;
}


//////////////////////////////////////////////////////////////////////////////
/// Open a bitmap from a file.
///
/// \param file Path to the bitmap file.
///
/// \return A pointer to the newly created bitmap.
///
/// Available bitmap types are: bmp, pcx, tga, jpeg and png.
//////////////////////////////////////////////////////////////////////////////
TS5_BITMAP *ts5_read_bitmap(const char *file)
{
	ts5_check_bitmaps("ts5_read_bitmap");
	ts5_log(TS5_LOGLEVEL_2, "ts5_read_bitmap(%s)\n", file);

	TS5_BITMAP *map;
	map = al_load_bitmap(file);

	if (map == NULL) {
		ts5_fatal("ts5_read_bitmap: could not read bitmap\n");
	}

	al_convert_bitmap(map);

	return map;
}


//////////////////////////////////////////////////////////////////////////////
/// Write a bitmap.
///
/// \param file Path to the bitmap file.
/// \param map Pointer to the bitmap that will be written.
///
/// The extension specifies the file type. Available bitmap types are: bmp, pcx, tga, jpeg and png.
///
/// Useful for saving screenshots of the experiment.
///
/// Useful for creating complex stimuli beforehand.
//////////////////////////////////////////////////////////////////////////////
void ts5_write_bitmap(const char *file, TS5_BITMAP *map)
{
	ts5_check_bitmaps("ts5_write_bitmap");
	ts5_log(TS5_LOGLEVEL_2, "ts5_write_bitmap(%s,%p)\n", file, map);

	if (!al_save_bitmap(file, map)) {
		ts5_fatal("ts5_write_bitmap: could not write bitmap\n");
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Free the memory used by a bitmap.
///
/// \param map pointer to the bitmap that will be freed.
///
/// This function should be called at the end of the program for
/// each bitmap allocated or read by the user.
//////////////////////////////////////////////////////////////////////////////
void ts5_free_bitmap(TS5_BITMAP *map)
{
	ts5_check_bitmaps("ts5_free_bitmap");
	ts5_log(TS5_LOGLEVEL_2, "ts5_free_bitmap(%p)\n", map);

	al_destroy_bitmap(map);
	map = NULL;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Query bitmap dimensions
/// The dimensions of a bitmap can be queried using the functions below:
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Get the size of a bitmap in pixels.
///
/// \param map The bitmap that will be queried.
/// \param w Variable that will store the width.
/// \param h Variable that will store the height.
///
/// You can pass NULL for values you are not interested in.
///
/// ts5_alloc_bitmap(), ts5_read_bitmap(), primitives.c, textio.c
//////////////////////////////////////////////////////////////////////////////
void ts5_get_bitmap_size(TS5_BITMAP *map, double *w, double *h)
{
	ts5_check_bitmaps("ts5_get_bitmap_size");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_bitmap_size(%p,%p,%p)\n", map, w, h);

	if (!map) {
		ts5_fatal("ts5_get_bitmap_size: bitmap pointer is null\n");
	}

	double ww, hh;

	ww =  al_get_bitmap_width(map);
	hh = al_get_bitmap_height(map);

	if (_ts5_status.graphics_status.coordinate_scale==TS5_RELATIVE_COORDINATES) {
		ww /= _ts5_status.graphics_status.target_width;
		hh /= _ts5_status.graphics_status.target_height;
	}

	if (w) {
		*w = ww;
	}

	if (h) {
		*h = hh;
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Get the width of a bitmap in pixels.
///
/// \param map The bitmap that will be queried.
///
/// \return The width of the bitmap.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_bitmap_width(TS5_BITMAP *map)
{
	ts5_check_bitmaps("ts5_get_bitmap_width");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_bitmap_width(%p)\n", map);

	double w;
	ts5_get_bitmap_size(map, &w, NULL);

	return w;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the height of a bitmap in pixels.
///
/// \param map The bitmap that will be queried.
///
/// \return The height of the bitmap.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_bitmap_height(TS5_BITMAP *map)
{
	ts5_check_bitmaps("ts5_get_bitmap_height");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_bitmap_width(%d)\n", map);

	double h;
	ts5_get_bitmap_size(map, NULL, &h);

	return h;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Drawing bitmaps on the display
/// 
/// The destination coordinates of the bitmap drawing functions below are 
/// interpreted dependent on the coordinate system used. 
/// - When using the Cartesian coordinate system, bitmaps are centered around the destination coordinates. 
/// - When using the Display coordinate system, the destination coordinates correspond to the upper left corner of the bitmap.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Draw a bitmap.
///
/// \param map The bitmap that will be drawn.
/// \param dx Horizontal position where it will be drawn.
/// \param dy Vertical position where it will be drawn.
//////////////////////////////////////////////////////////////////////////////
void ts5_draw_bitmap(TS5_BITMAP *map, double dx, double dy)
{
	ts5_check_bitmaps2("ts5_draw_bitmap", map);
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_bitmap(%p,%f,%f)\n", map, dx, dy);

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		dx = ts5_relative_to_absolute_coordinate_x(dx);
		dy = ts5_relative_to_absolute_coordinate_y(dy);
	}

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		dx = ts5_cartesian_to_display_coordinate_x(dx) - al_get_bitmap_width(map) / 2.0;
		dy = ts5_cartesian_to_display_coordinate_y(dy) - al_get_bitmap_height(map) / 2.0;
	}

	al_draw_bitmap(map, dx, dy, 0);
}


//////////////////////////////////////////////////////////////////////////////
/// Draw a flipped bitmap.
///
/// \param map The bitmap that will be drawn.
/// \param dx Horizontal position where it will be drawn.
/// \param dy Vertical position where it will be drawn.
/// \param flip Direction of the flip. Can be TS5_FLIP_HORIZONTAL, TS5_FLIP_VERTICAL or TS5_FLIP_BOTH.
//////////////////////////////////////////////////////////////////////////////
void ts5_draw_flipped_bitmap(TS5_BITMAP *map, double dx, double dy, const int flip)
{
	ts5_check_bitmaps2("ts5_draw_flipped_bitmap", map);
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_flipped_bitmap(%p,%f,%f,%d)\n", map, dx, dy, flip);

	if (flip != TS5_FLIP_HORIZONTAL && flip != TS5_FLIP_VERTICAL && flip != TS5_FLIP_BOTH) {
		ts5_fatal("ts5_draw_flipped_bitmap: flip should be",
					" TS5_FLIP_HORIZONTAL, TS5_FLIP_VERTICAL or TS5_FLIP_BOTH\n");
	}

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		dx = ts5_relative_to_absolute_coordinate_x(dx);
		dy = ts5_relative_to_absolute_coordinate_y(dy);
	}

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		dx = ts5_cartesian_to_display_coordinate_x(dx) - al_get_bitmap_width(map) / 2.0;
		dy = ts5_cartesian_to_display_coordinate_y(dy) - al_get_bitmap_height(map) / 2.0;
	}

	al_draw_bitmap(map, dx, dy, flip);
}


//////////////////////////////////////////////////////////////////////////////
/// Draw a tinted bitmap.
///
/// \param map The bitmap that will be drawn.
/// \param tint The tint color. All colors in the bitmap are multiplied with the color specified in tint.
/// \param dx Horizontal position where it will be drawn.
/// \param dy Vertical position where it will be drawn.
//////////////////////////////////////////////////////////////////////////////
void ts5_draw_tinted_bitmap(TS5_BITMAP *map, TS5_COLOR tint, double dx, double dy)
{
	ts5_check_bitmaps2("ts5_draw_tinted_bitmap", map);
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_tinted_bitmap(%p,(%f,%f,%f,%f),%f,%f)\n",
				map, tint.r, tint.g, tint.b, tint.a, dx, dy);

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		dx = ts5_relative_to_absolute_coordinate_x(dx);
		dy = ts5_relative_to_absolute_coordinate_y(dy);
	}

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		dx = ts5_cartesian_to_display_coordinate_x(dx) - al_get_bitmap_width(map) / 2.0;
		dy = ts5_cartesian_to_display_coordinate_y(dy) - al_get_bitmap_height(map) / 2.0;
	}

	al_draw_tinted_bitmap(map, tint, dx, dy, 0);
}


//////////////////////////////////////////////////////////////////////////////
/// Draw a region of a bitmap.
///
/// \param map The bitmap that will be drawn.
/// \param sx1 Horizontal coordinate of the first corner or the region.
/// \param sy1 Vertical coordinate of the first corner or the region.
/// \param sx2 Horizontal coordinate of the second corner or the region.
/// \param sy2 Vertical coordinate of the second corner or the region.
/// \param dx Horizontal position where it will be drawn.
/// \param dy Vertical position where it will be drawn.
//////////////////////////////////////////////////////////////////////////////
void ts5_draw_bitmap_region(TS5_BITMAP *map, double sx1, double sy1, double sx2, double sy2,
								double dx, double dy)
{
	ts5_check_bitmaps2("ts5_draw_bitmap_region", map);
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_bitmap_region(%p,%f,%f,%f,%f,%f,%f)\n",
				map, sx1, sy1, sx2, sy2, dx, dy);

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		dx = ts5_relative_to_absolute_coordinate_x(dx);
		dy = ts5_relative_to_absolute_coordinate_y(dy);
		TS5_BITMAP *oldval = ts5_set_drawing_target (map);
		sx1 = ts5_relative_to_absolute_coordinate_x(sx1);
		sx2 = ts5_relative_to_absolute_coordinate_x(sx2);
		sy1 = ts5_relative_to_absolute_coordinate_y(sy1);
		sy2 = ts5_relative_to_absolute_coordinate_y(sy2);
		ts5_set_drawing_target(oldval);
	}

	double width = abs(sx2 - sx1);
	double height = abs(sy2 - sy1);
	double x = sx1 < sx2 ? sx1 : sx2;
	double y = sy1 < sy2 ? sy1 : sy2;

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		dx = ts5_cartesian_to_display_coordinate_x(dx - width / 2.0);
		dy = ts5_cartesian_to_display_coordinate_y(height / 2.0 - dy);
		x = x + al_get_bitmap_width(map) / 2.0;
		y = y + al_get_bitmap_height(map) / 2.0;
	}

	al_draw_bitmap_region(map, x, y, width, height, dx, dy, 0);
}


//////////////////////////////////////////////////////////////////////////////
/// Draw a rotated bitmap.
///
/// \param map The bitmap that will be drawn.
/// \param cx Horizontal coordinate of the center of rotation.
/// \param cy Certical coordinate of the center of rotation.
/// \param angle The angle of rotation in degrees (positive is clockwise, negative is counter clockwise).
/// \param dx Horizontal position where it will be drawn.
/// \param dy Vertical position where it will be drawn.
///
/// The center of rotation (cx,cy) will be drawn at (dx,dy).
//////////////////////////////////////////////////////////////////////////////
void ts5_draw_rotated_bitmap(TS5_BITMAP *map, double cx, double cy, double angle, double dx, double dy)
{
	ts5_check_bitmaps2("ts5_draw_rotated_bitmap", map);
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_rotated_bitmap(%p,%f,%f,%f,%f,%f)\n", map, cx, cy, angle, dx, dy);

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		dx = ts5_relative_to_absolute_coordinate_x(dx);
		dy = ts5_relative_to_absolute_coordinate_y(dy);
		TS5_BITMAP *oldval = ts5_set_drawing_target (map);
		cx = ts5_relative_to_absolute_coordinate_x(cx);
		cy = ts5_relative_to_absolute_coordinate_y(cy);
		ts5_set_drawing_target(oldval);
	}

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		dx = ts5_cartesian_to_display_coordinate_x(dx);
		dy = ts5_cartesian_to_display_coordinate_y(dy);
		cx = al_get_bitmap_width(map) / 2.0 + cx;
		cy = al_get_bitmap_height(map) / 2.0 - cy;
	}

	angle = angle * TS5_PI / 180.0;

	al_draw_rotated_bitmap(map, cx, cy, dx, dy, angle, 0);
}


//////////////////////////////////////////////////////////////////////////////
/// Draw a scaled bitmap.
///
/// \param map The bitmap that will be drawn.
/// \param xratio The scaling factor in the horizontal direction.
/// \param yratio The scaling factor in the vertical direction.
/// \param dx Horizontal position where it will be drawn.
/// \param dy Vertical position where it will be drawn.
//////////////////////////////////////////////////////////////////////////////
void ts5_draw_scaled_bitmap(TS5_BITMAP *map, double xratio, double yratio, double dx, double dy)
{
	ts5_check_bitmaps2("ts5_draw_scaled_bitmap", map);
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_scaled_bitmap(%p,%f,%f,%f,%f)\n", map, xratio, yratio, dx, dy);

	double h = al_get_bitmap_height(map);
	double w = al_get_bitmap_width(map);

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		dx = ts5_relative_to_absolute_coordinate_x(dx);
		dy = ts5_relative_to_absolute_coordinate_y(dy);
	}

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		dx = ts5_cartesian_to_display_coordinate_x(dx) - w * xratio / 2.0;
		dy = ts5_cartesian_to_display_coordinate_y(dy) - h * yratio / 2.0;
	}

	al_draw_scaled_bitmap(map, 0.0, 0.0, w, h, dx, dy, w * xratio, h * yratio, 0);
}

