//////////////////////////////////////////////////////////////////////////////
/// \file cedrusbox_internal.c
/// Definitions of internal cedrusbox functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/cedrusbox.h"
#include "../include/tscope5/cedrusbox_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"

/// Is the cedrusbox subsystem installed?
int  _ts5_is_cedrusbox_installed = 0;

#ifdef ALLEGRO_WINDOWS
	#include "cedrusbox_windows_internal.c"

#else
	#include "cedrusbox_posix_internal.c"

#endif


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each cedrusbox function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether the cedrusbox subsystem is installed. If not, the cedrusbox subsystem is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_cedrusbox(char *calling_function)
{
	if (!_ts5_is_cedrusbox_installed) {
		ts5_install_cedrusbox(calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_cedrusbox\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Install the cedrusbox subsystem.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// This function is called automatically if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_install_cedrusbox(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_cedrusbox_installed) {

		_ts5_is_cedrusbox_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 cedrusbox\n", calling_function);
	}
	else {
		goto cedrusbox_is_installed;
	}

	// scan the serial ports for possible cedrusboxes
	// this is OS-dependent
	int num_serial_devices=0;

	#ifdef ALLEGRO_WINDOWS
		num_serial_devices = ts5_install_cedrusbox_windows(calling_function);

	#else
		num_serial_devices = ts5_install_cedrusbox_posix(calling_function);

	#endif

	// are the serial devices we found cedrusboxes?
	int i;
	int device_type[TS5_CEDRUSBOX_MAXPORT];

	for (i=0; i<num_serial_devices; i++) {

		char serial_buff[TS5_MAX_CHAR];

		// flush
		ts5_fflush_cedrusbox(i);

		// set xid mode
		ts5_write_cedrusbox(i, "c10", 3);

		// get product name
	//	unsigned int bytes_read;
	//	ts5_write_cedrusbox(i, "_d1", 3);
	//	bytes_read = ts5_read_cedrusbox(i, serial_buff, 80);
	//
	//	unsigned int j;
	//	for (j=0; j<bytes_read; j++) {
	//		if (serial_buff[j]==13) {
	//			serial_buff[j]=32;
	//		}
	//	}
	//	serial_buff[bytes_read] = 0;
	//	j=0;
	//	while (serial_buff[j]!=82 && j<bytes_read) {
	//		j++;
	//	}
	//
	//	ts5_log(TS5_LOGLEVEL_1,"ts5_install_cedrusbox: cedrusbox %s\n", &serial_buff[j]);

		// get the device ID
		ts5_fflush_cedrusbox(i);
		ts5_write_cedrusbox(i, "_d2", 3);
		ts5_read_cedrusbox(i, serial_buff, 6);
		device_type[i] = serial_buff[0] - 47; // lumina 1, SV1 2, RBX 3

		if (device_type[i]==3) {
			ts5_fflush_cedrusbox(i);
			ts5_write_cedrusbox(i, "_d3", 6);
			ts5_read_cedrusbox(i, serial_buff, 1);
			device_type[i] += serial_buff[0] - 49;
		}

		switch (device_type[i]) {

			case TS5_CEDRUSBOX_LUMINA:
				sprintf(serial_buff, "lumina");
				break;

			case TS5_CEDRUSBOX_SV1:
				sprintf(serial_buff, "sv1");
				break;

			case TS5_CEDRUSBOX_RB530:
				sprintf(serial_buff, "rb530");
				break;

			case TS5_CEDRUSBOX_RB730:
				sprintf(serial_buff, "rb730");
				break;

			case TS5_CEDRUSBOX_RB830:
				sprintf(serial_buff, "rb830");
				break;

			case TS5_CEDRUSBOX_RB834:
				sprintf(serial_buff, "rb834");
				break;

			default:

				device_type[i] = TS5_CEDRUSBOX_NODEVICE;

				#ifdef ALLEGRO_WINDOWS
					if (!CloseHandle(_ts5_cedrusbox_handle[i])) {
						ts5_log(TS5_LOGLEVEL_1, "%s: could not remove unknown device\n", calling_function);
					}

				#else
					tcdrain(_ts5_cedrusbox_fd[i]);
					tcsetattr(_ts5_cedrusbox_fd[i], TCSANOW, &_ts5_cedrusbox_oldoptions[i]);

					if (close(_ts5_cedrusbox_fd[i]) == -1) {
						ts5_log(TS5_LOGLEVEL_1, "%s: could not remove unknown device\n", calling_function);
					}
				#endif

				continue;
		}

		ts5_log(TS5_LOGLEVEL_1,"%s: cedrusbox ID %d (%s)\n",
					calling_function, device_type[i], serial_buff);

		// get firmware revision
		char minor, major;
		ts5_fflush_cedrusbox(i);
		ts5_write_cedrusbox(i, "_d4", 3);
		ts5_read_cedrusbox(i, serial_buff, 6);
		major = serial_buff[0];

		ts5_fflush_cedrusbox(i);
		ts5_write_cedrusbox(i, "_d5", 3);
		ts5_read_cedrusbox(i, serial_buff, 6);
		minor = serial_buff[0];
		ts5_log(TS5_LOGLEVEL_1,"%s: cedrusbox firmware revision: %c.%c\n",
					calling_function, major, minor);

		// reset timers
		ts5_fflush_cedrusbox(i);
		ts5_write_cedrusbox(i, "e1", 2);

		// if we get here mark it as a cedrusbox
		_ts5_status.timer_status.num_cedrusbox++;
	}

	// abort if we could not detect any cedrusboxes
	if (!_ts5_status.timer_status.num_cedrusbox) {
		ts5_fatal("%s: could not install Tscope5 cedrusboxes\n", calling_function);
	}

	// initialize event source
	al_init_user_event_source(&_ts5_status.timer_status.cedrusbox_response_event_source);

	// register event source
	al_register_event_source(_ts5_status.timer_status.response_queue,
								&_ts5_status.timer_status.cedrusbox_response_event_source);

	// initialize thread
	_ts5_status.timer_status.cedrusbox_thread = al_create_thread(ts5_cedrusbox_threadfunc, NULL);
	al_start_thread(_ts5_status.timer_status.cedrusbox_thread);

	// assign response codes
	_ts5_status.timer_status.cedrusbox = (TS5_CEDRUSBOX_STATUS *)al_malloc(sizeof(TS5_CEDRUSBOX_STATUS)*
											_ts5_status.timer_status.num_cedrusbox);

	int j=0;
	for (i=0; i<num_serial_devices; i++) {

		if (device_type[i]==TS5_CEDRUSBOX_NODEVICE) {
			continue;
		}

		_ts5_status.timer_status.cedrusbox[j].port_num = i;
		_ts5_status.timer_status.cedrusbox[j].type = device_type[i];

		switch (_ts5_status.timer_status.cedrusbox[j].type) {

			case TS5_CEDRUSBOX_LUMINA:
				_ts5_status.timer_status.cedrusbox[j].num_buttons = 4;
				break;

			case TS5_CEDRUSBOX_SV1:
				_ts5_status.timer_status.cedrusbox[j].num_buttons = 1;
				break;

			case TS5_CEDRUSBOX_RB530:
				_ts5_status.timer_status.cedrusbox[j].num_buttons = 5;
				break;

			case TS5_CEDRUSBOX_RB730:
				_ts5_status.timer_status.cedrusbox[j].num_buttons = 7;
				break;

			case TS5_CEDRUSBOX_RB830:
				_ts5_status.timer_status.cedrusbox[j].num_buttons = 8;
				break;

			case TS5_CEDRUSBOX_RB834:
				_ts5_status.timer_status.cedrusbox[j].num_buttons = 8;
				break;

			default:
				ts5_fatal("%s: unknown cedrus device type\n", calling_function);
		}

		_ts5_status.timer_status.cedrusbox[j].num_defined_buttons = 0;
		_ts5_status.timer_status.cedrusbox[j].num_active_buttons = 0;

		_ts5_status.timer_status.cedrusbox[j].button_press_defined =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.cedrusbox[j].num_buttons);

		_ts5_status.timer_status.cedrusbox[j].button_press_active =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.cedrusbox[j].num_buttons);

		_ts5_status.timer_status.cedrusbox[j].button_release_defined =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.cedrusbox[j].num_buttons);

		_ts5_status.timer_status.cedrusbox[j].button_release_active =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.cedrusbox[j].num_buttons);

		int k;
		for (k=0; k<_ts5_status.timer_status.cedrusbox[j].num_buttons; k++) {
			_ts5_status.timer_status.cedrusbox[j].button_press_defined[k] = 0;
			_ts5_status.timer_status.cedrusbox[j].button_press_active[k] = 0;
			_ts5_status.timer_status.cedrusbox[j].button_release_defined[k] = 0;
			_ts5_status.timer_status.cedrusbox[j].button_release_active[k] = 0;
		}
		j++;
	}
	atexit(ts5_uninstall_cedrusbox);

	cedrusbox_is_installed:
	;

}


//////////////////////////////////////////////////////////////////////////////
/// Uninstall the cedrusbox subsystem.
///
/// This function is called automatically at the end of the program.
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_cedrusbox()
{
	if (!_ts5_is_cedrusbox_installed) {
		goto cedrusbox_is_uninstalled;
	}

	ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 cedrusbox\n");

	int i;

	// reset voice keys
	for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

		ts5_fflush_cedrusbox(i);

		if (_ts5_status.timer_status.cedrusbox[i].type==TS5_CEDRUSBOX_SV1) {

			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_LOCKINGLEVEL, TS5_CEDRUSBOX_LOCK);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_MIN_TRESHOLD, 0);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_MAX_TRESHOLD, 255);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_TRESHOLD, 128);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_MIN_RISE_DELAY, 0);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_MAX_RISE_DELAY, 100);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_RISE_DELAY, 0);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_DROP_DELAY, 0);
			ts5_cedrusbox_set_parameter(i, TS5_CEDRUSBOX_LOCKINGLEVEL, TS5_CEDRUSBOX_UNLOCK);

		}
	}

	// remove thread
	al_join_thread(_ts5_status.timer_status.cedrusbox_thread, NULL);
	al_destroy_thread(_ts5_status.timer_status.cedrusbox_thread);

	// unregister event source
	al_unregister_event_source(_ts5_status.timer_status.response_queue,
								&_ts5_status.timer_status.cedrusbox_response_event_source);

	// remove event source
	al_destroy_user_event_source(&_ts5_status.timer_status.cedrusbox_response_event_source);

	// close devices
	#ifdef ALLEGRO_WINDOWS
		ts5_uninstall_cedrusbox_windows();

	#else
		ts5_uninstall_cedrusbox_posix();

	#endif

	// remove key definitions
	for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

		// free data structure
		for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

			al_free(_ts5_status.timer_status.cedrusbox[i].button_press_defined);
			_ts5_status.timer_status.cedrusbox[i].button_press_defined = NULL;

			al_free(_ts5_status.timer_status.cedrusbox[i].button_press_active);
			_ts5_status.timer_status.cedrusbox[i].button_press_active = NULL;

			al_free(_ts5_status.timer_status.cedrusbox[i].button_release_defined);
			_ts5_status.timer_status.cedrusbox[i].button_release_defined = NULL;

			al_free(_ts5_status.timer_status.cedrusbox[i].button_release_active);
			_ts5_status.timer_status.cedrusbox[i].button_release_active = NULL;
		}

		al_free(_ts5_status.timer_status.cedrusbox);
		_ts5_status.timer_status.cedrusbox=NULL;
		_ts5_status.timer_status.num_cedrusbox = 0;
		_ts5_status.timer_status.cedrusbox_is_response_device = 0;

		_ts5_is_cedrusbox_installed = 0;
	}

	cedrusbox_is_uninstalled:
	;
}


//////////////////////////////////////////////////////////////////////////////
/// Read text from a cedrusbox.
///
/// \param port Port number the cedrusbox is attached to.
/// \param buff Character buffer.
/// \param bytes_to_read Number of bytes to be written.
///
/// \return Length of the string that is sent to the port.
///
/// \warning This is an internal function, no checks are performed.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_read_cedrusbox(int port, char *buff, unsigned long bytes_to_read)
{
	ts5_log(TS5_LOGLEVEL_5, "ts5_read_cedrusbox(%d,%s,%lu)\n", port, buff, bytes_to_read);

	unsigned int retval;

	#ifdef ALLEGRO_WINDOWS
		retval = ts5_read_cedrusbox_windows(port, buff, bytes_to_read);

	#else
		retval = ts5_read_cedrusbox_posix(port, buff, bytes_to_read);

	#endif

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Write text to a cedrusbox.
///
/// \param port Port number the cedrusbox is attached to.
/// \param buff Character buffer.
/// \param bytes_to_write Number of bytes to be written.
///
/// \return Length of the string that is read.
///
/// \warning This is an internal function, no checks are performed.
//////////////////////////////////////////////////////////////////////////////
unsigned int ts5_write_cedrusbox(int port, char *buff, unsigned long bytes_to_write)
{
	ts5_log(TS5_LOGLEVEL_5, "ts5_write_cedrusbox(%d,%s,%lu)\n", port, buff, bytes_to_write);

	unsigned int retval;

	#ifdef ALLEGRO_WINDOWS
		retval = ts5_write_cedrusbox_windows(port, buff, bytes_to_write);

	#else
		retval = ts5_write_cedrusbox_posix(port, buff, bytes_to_write);

	#endif

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Flush the output of a cedrusbox.
///
/// \param port Port number the cedrusbox is attached to.
///
/// \warning This is an internal function, no checks are performed.
//////////////////////////////////////////////////////////////////////////////
void ts5_fflush_cedrusbox(int port)
{
	ts5_log(TS5_LOGLEVEL_5, "ts5_fflush_cedrusbox(%d)\n", port);

	#ifdef ALLEGRO_WINDOWS
		ts5_fflush_cedrusbox_windows(port);

	#else
		ts5_fflush_cedrusbox_posix(port);

	#endif
}


//////////////////////////////////////////////////////////////////////////////
/// Cedrusbox thread function.
///
/// \warning This is an internal function, no checks are performed.
//////////////////////////////////////////////////////////////////////////////
void *ts5_cedrusbox_threadfunc()
{
	ALLEGRO_EVENT cedrusbox_event;
	double looptime = al_get_time();
	double now;

	while (!al_get_thread_should_stop(_ts5_status.timer_status.cedrusbox_thread)) {

		now = al_get_time();

		char serial_buff[6] = {0};

		int i;
		for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

			unsigned int num_bytes =
				ts5_read_cedrusbox(_ts5_status.timer_status.cedrusbox[i].port_num, serial_buff, 6);

			if (num_bytes==6 && serial_buff[0]=='k') {

				int action =  ((unsigned char)(serial_buff[1]) & 16) >> 4;

				if (action==0) {
					cedrusbox_event.user.type = TS5_EVENT_CEDRUS_BUTTON_UP;
				}

				if (action==1) {
					cedrusbox_event.user.type = TS5_EVENT_CEDRUS_BUTTON_DOWN;
				}

				cedrusbox_event.user.data1 = i;
				cedrusbox_event.user.data2 = ((unsigned char)serial_buff[1] >> 5) - 1;
				cedrusbox_event.user.data3 = (intptr_t)((now-looptime)*1000000.0);

				al_emit_user_event(&_ts5_status.timer_status.cedrusbox_response_event_source,
									&cedrusbox_event, NULL);
			}
		}
		looptime = now;
	}

	return NULL;
}

