//////////////////////////////////////////////////////////////////////////////
/// \file display_internal.c
/// Definitions of internal display functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/display.h"
#include "../include/tscope5/display_internal.h"
#include "../include/tscope5/system_internal.h"

/// Is the display subsystem installed?
int  _ts5_is_display_installed = 0;


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each display function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether is Tscope5 installed. If not, Tscope5 is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_display(char *calling_function)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5(calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_display\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each display function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// \param display Index of the display that has to be checked.
///
/// Checks whether display 'display' is a valid display. If not, the program is aborted.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_display2(char *calling_function, const int display)
{
	ts5_check_display(calling_function);

	if (display < 0 || display > _ts5_status.num_displays) {
		ts5_fatal("%s: display index not valid\n", calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_display2\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Close all open displays.
///
/// Currently all displays are closed at once,
/// there is no option to close only one display
/// and leave the others open.
///
/// This function is called automatically at the end of the program.
///
/// \warning Do not call this function yourself and try to reopen a new window in the same program.
/// This will result in undefined behavior (i.e. sometimes it will work fine, sometimes it will crash).
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_displays()
{
	ts5_check_display("ts5_uninstall_displays");
	ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 displays\n");
//	ts5_uninstall_mouse();
//	ts5_uninstall_keyboard();

	#ifdef ALLEGRO_MACOSX
		int i;
		for (i = 0; i < _ts5_status.num_displays; i++) {

			if (_ts5_status.display_status[i].display_mode == TS5_FULLSCREEN_WINDOW) {
				al_toggle_display_flag(_ts5_status.display[i], ALLEGRO_FULLSCREEN_WINDOW, 0);
			}

			al_destroy_display(_ts5_status.display[i]);
			al_destroy_bitmap(_ts5_status.display_buffer[i]);
		}

		al_free(_ts5_status.display);
		al_free(_ts5_status.display_buffer);
		al_free(_ts5_status.display_status);

		_ts5_status.display_status = NULL;
		_ts5_status.display = NULL;
		_ts5_status.display_buffer = NULL;
		_ts5_status.target = NULL;

	#endif

	_ts5_status.num_displays = 0;
	_ts5_status.active_display = -1;
}

