//////////////////////////////////////////////////////////////////////////////
/// \file joystick.c
/// Definitions of joystick functions
//////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/joystick.h"
#include "../include/tscope5/timer.h"
#include "../include/tscope5/joystick_internal.h"
#include "../include/tscope5/system_internal.h"


//////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// Joysticks can only be used as response devices. 
/// See timer.c for more information about response registration.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Define a joystick key as a response key.
///
/// \param device Number of the joystick (devices are numbered from 1).
/// \param key Number of the key (keys are counted from 1).
///
/// \return The reponse number associated with the key.
///
/// Give a positive number if you want to monitor key press events,
/// a negative number if you want to monitor key release events.
//////////////////////////////////////////////////////////////////////////////
int ts5_define_joystick_key(int device, int key)
{
	ts5_check_joystick("ts5_define_joystick_key");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_joystick_key(%d,%d)\n", device, key);

	if (device<1 || device>_ts5_status.timer_status.num_joystick) {
		ts5_fatal("ts5_define_joystick_key: device argument is %d, number of joysticks is %d\n",
					device, _ts5_status.timer_status.num_joystick);
	}

	if (key==0) {
		ts5_fatal("ts5_define_joystick_key: key argument is 0, response keys are numbered from 1\n");
	}

	if (abs(key)>_ts5_status.timer_status.joystick[device-1].num_buttons) {
		ts5_fatal("ts5_define_joystick_key: key argument is %d, number of joystick buttons for device",
					" %d is %d\n", key, device, _ts5_status.timer_status.joystick[device-1].num_buttons);
	}

	if (key>0 && _ts5_status.timer_status.joystick[device-1].button_press_defined[key-1]!=0) {
		ts5_fatal("ts5_define_joystick_key: key press %d is already defined\n", key);
	}

	if (key<0 && _ts5_status.timer_status.joystick[device-1].button_release_defined[-key-1]!=0) {
		ts5_fatal("ts5_define_joystick_key: key release %d is already defined\n", key);
	}

	_ts5_status.timer_status.joystick_is_response_device = 1;
	_ts5_status.timer_status.num_defined_buttons++;
	_ts5_status.timer_status.num_active_buttons++;
	_ts5_status.timer_status.joystick[device-1].num_defined_buttons++;
	_ts5_status.timer_status.joystick[device-1].num_active_buttons++;

	if (key>0) {
		_ts5_status.timer_status.joystick[device-1].button_press_defined[key-1] =
			_ts5_status.timer_status.num_defined_buttons;

		_ts5_status.timer_status.joystick[device-1].button_press_active[key-1] =
			_ts5_status.timer_status.num_defined_buttons;
	}
	else {
		_ts5_status.timer_status.joystick[device-1].button_release_defined[-key-1] =
			_ts5_status.timer_status.num_defined_buttons;

		_ts5_status.timer_status.joystick[device-1].button_release_active[-key-1] =
			_ts5_status.timer_status.num_defined_buttons;
	}

	return _ts5_status.timer_status.num_defined_buttons;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////
