//////////////////////////////////////////////////////////////////////////////
/// \file keyboard_internal.c
/// Definitions of internal keyboard functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/keyboard_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"

/// Is the keyboard subsystem installed?
int  _ts5_is_keyboard_installed = 0;


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each keyboard function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether the keyboard subsystem is installed. If not, the keyboard subsystem is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_keyboard(char *calling_function)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_fatal("%s: tscope5 not installed\n", calling_function);
	}

	if (_ts5_status.active_display == -1) {
		ts5_fatal("%s: no active display available\n", calling_function);
	}

	if (!_ts5_is_keyboard_installed) {
		ts5_install_keyboard(calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_keyboard\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Install the keyboard subsystem.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// This function is called automatically if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_install_keyboard(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_keyboard_installed) {

		_ts5_is_keyboard_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 keyboard\n", calling_function);

		if (!al_install_keyboard()) {
			ts5_fatal("%s: could not install tscope5 keyboard\n", calling_function);
		}

		al_register_event_source(_ts5_status.timer_status.response_queue,
									al_get_keyboard_event_source());

		_ts5_status.timer_status.keyboard.num_buttons = TS5_KEY_MAX;
		_ts5_status.timer_status.keyboard.num_defined_buttons = 0;
		_ts5_status.timer_status.keyboard.num_active_buttons = 0;

		_ts5_status.timer_status.keyboard.button_press_defined =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.keyboard.num_buttons);

		_ts5_status.timer_status.keyboard.button_press_active =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.keyboard.num_buttons);

		_ts5_status.timer_status.keyboard.button_release_defined =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.keyboard.num_buttons);

		_ts5_status.timer_status.keyboard.button_release_active =
			(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.keyboard.num_buttons);

		int i;
		for (i=0; i<_ts5_status.timer_status.keyboard.num_buttons; i++) {
			_ts5_status.timer_status.keyboard.button_press_defined[i] = 0;
			_ts5_status.timer_status.keyboard.button_press_active[i] = 0;
			_ts5_status.timer_status.keyboard.button_release_defined[i] = 0;
			_ts5_status.timer_status.keyboard.button_release_active[i] = 0;
		}

		atexit(ts5_uninstall_keyboard);
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Uninstall the keyboard subsystem.
///
/// This function is called automatically at the end of the program.
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_keyboard()
{
	if (_ts5_is_keyboard_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 keyboard\n");

		_ts5_status.timer_status.keyboard.num_buttons = 0;
		_ts5_status.timer_status.keyboard.num_defined_buttons = 0;
		_ts5_status.timer_status.keyboard.num_active_buttons = 0;

		al_free(_ts5_status.timer_status.keyboard.button_press_defined);
		_ts5_status.timer_status.keyboard.button_press_defined = NULL;

		al_free(_ts5_status.timer_status.keyboard.button_press_active);
		_ts5_status.timer_status.keyboard.button_press_active = NULL;

		al_free(_ts5_status.timer_status.keyboard.button_release_defined);
		_ts5_status.timer_status.keyboard.button_release_defined = NULL;

		al_free(_ts5_status.timer_status.keyboard.button_release_active);
		_ts5_status.timer_status.keyboard.button_release_active = NULL;

		_ts5_status.timer_status.keyboard_is_response_device = 0;

		al_unregister_event_source(_ts5_status.timer_status.response_queue,
									al_get_keyboard_event_source());

		al_uninstall_keyboard();

		_ts5_is_keyboard_installed = 0;
	}
}

