//////////////////////////////////////////////////////////////////////////////
/// \file mouse.c
/// Definitions of mouse functions
//////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/mouse.h"
#include "../include/tscope5/timer.h"
#include "../include/tscope5/graphics.h"
#include "../include/tscope5/display.h"
#include "../include/tscope5/primitives.h"
#include "../include/tscope5/mouse_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/graphics_internal.h"


//////////////////////////////////////////////////////////////////////////////
/// @name Mouse GUI functions
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Make the mouse cursor visible on the active display.
//////////////////////////////////////////////////////////////////////////////
void ts5_show_mouse()
{
	ts5_check_mouse("ts5_show_mouse");
	ts5_log(TS5_LOGLEVEL_5, "ts5_show_mouse()\n");

	al_show_mouse_cursor(_ts5_status.display[_ts5_status.active_display]);
}


//////////////////////////////////////////////////////////////////////////////
/// Make the mouse cursor invisible on the active display.
//////////////////////////////////////////////////////////////////////////////
void ts5_hide_mouse()
{
	ts5_check_mouse("ts5_hide_mouse");
	ts5_log(TS5_LOGLEVEL_5, "ts5_hide_mouse()\n");

	al_hide_mouse_cursor(_ts5_status.display[_ts5_status.active_display]);
}


//////////////////////////////////////////////////////////////////////////////
/// Move the mouse pointer to position (x,y).
///
/// \param x Horizontal position of the mouse on the active display.
/// \param y Vertical position of the mouse on the active display.
//////////////////////////////////////////////////////////////////////////////
void ts5_set_mouse_position(double x, double y)
{
	ts5_check_mouse("ts5_set_mouse_position");
	ts5_log(TS5_LOGLEVEL_5, "ts5_set_mouse_position(%f,%f)\n", x, y);

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		x = ts5_relative_to_absolute_coordinate_x(x);
		y = ts5_relative_to_absolute_coordinate_y(y);
	}

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		x = ts5_cartesian_to_display_coordinate_x(x);
		y = ts5_cartesian_to_display_coordinate_y(y);
	}

	al_set_mouse_xy(_ts5_status.display[_ts5_status.active_display], (int)x, (int)y);
}

//////////////////////////////////////////////////////////////////////////////
/// Get the horizontal and vertical position of the mouse pointer
///
/// \param x Variable that will store the horizontal position.
/// \param y Variable that will store the vertical position.
///
/// You can pass NULL for values you are not interested in.
//////////////////////////////////////////////////////////////////////////////
void ts5_get_mouse_position(double *x, double *y)
{
	ts5_check_mouse("ts5_get_mouse_position");
	ts5_log(TS5_LOGLEVEL_6, "ts5_get_mouse_position(%p,%p)\n", x, y);

	ALLEGRO_MOUSE_STATE status;
	al_get_mouse_state(&status);

	double xx, yy;

	xx = al_get_mouse_state_axis(&status, 0);
	yy = al_get_mouse_state_axis(&status, 1);

	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		xx = ts5_display_to_cartesian_coordinate_x(xx);
		yy = ts5_display_to_cartesian_coordinate_y(yy);
	}

	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		xx = ts5_absolute_to_relative_coordinate_x(xx);
		yy = ts5_absolute_to_relative_coordinate_y(yy);
	}

	if (x) {
		*x = xx;
	}

	if (y) {
		*y = yy;
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Get the horizontal position of the mouse pointer
///
/// \return The horizontal position of the mouse pointer.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_mouse_x()
{
	ts5_check_mouse("ts5_get_mouse_x");
	ts5_log(TS5_LOGLEVEL_6, "ts5_get_mouse_x()\n");

	double x;
	ts5_get_mouse_position(&x, NULL);

	return x;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the vertical position of the mouse pointer
///
/// \return The vertical position of the mouse pointer.
//////////////////////////////////////////////////////////////////////////////
double ts5_get_mouse_y()
{
	ts5_check_mouse("ts5_get_mouse_y");
	ts5_log(TS5_LOGLEVEL_6, "ts5_get_mouse_y()\n");

	double y;
	ts5_get_mouse_position(NULL, &y);

	return y;
}


//////////////////////////////////////////////////////////////////////////////
/// Draw a mouse button and wait for a click.
///
/// \param x Horizontal position of the button on the active display.
/// \param y Vertical position of the button on the active display.
///
/// The button has an ugly green color. This color can not be changed.
/// This is not a bug. It is a Deliberately Annoying Feature (DAF).
//////////////////////////////////////////////////////////////////////////////
void ts5_draw_mouse_button(double x, double y)
{
	ts5_check_mouse("ts5_draw_mouse_button");
	ts5_log(TS5_LOGLEVEL_5, "ts5_draw_mouse_button(%f,%f)\n", x, y);

	// always write to the display
	ts5_set_active_display(_ts5_status.active_display+1);
	TS5_BITMAP *oldmap = ts5_set_drawing_target(_ts5_status.display_buffer[_ts5_status.active_display]);

	TS5_COLOR oldcolor = ts5_set_foreground_color(ts5_make_named_color("chartreuse", 1.0));
	double oldthick = ts5_set_drawing_thickness(4.0);
	int oldfill = ts5_set_fill_mode(TS5_FILL_OFF);

	double w_abs_disp = 40, h_abs_disp = 40;
	double x1_abs_disp, x2_abs_disp, y1_abs_disp, y2_abs_disp;

	// covert to absolute scale
	if (_ts5_status.graphics_status.coordinate_scale == TS5_RELATIVE_COORDINATES) {
		x1_abs_disp = ts5_relative_to_absolute_coordinate_x(x);
		y1_abs_disp = ts5_relative_to_absolute_coordinate_y(y);
	}
	else {
		x1_abs_disp = x;
		y1_abs_disp = y;
	}

	// convert to display coordinates.
	// for cartesian display coordinates x and y are the center
	// for display coordinates x and y are
	if (_ts5_status.graphics_status.coordinate_system == TS5_CARTESIAN_COORDINATES) {
		x1_abs_disp = ts5_cartesian_to_display_coordinate_x(x1_abs_disp);
		y1_abs_disp = ts5_cartesian_to_display_coordinate_y(y1_abs_disp);

		x1_abs_disp -= w_abs_disp/2;
		y1_abs_disp -= h_abs_disp/2;
	}

	x2_abs_disp = x1_abs_disp + w_abs_disp;
	y2_abs_disp = y1_abs_disp + h_abs_disp;

	int oldsystem = ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	int oldscale = ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);

	// check if the button will be on the screen
	double dispw = _ts5_status.display_status[_ts5_status.active_display].w;
	double disph = _ts5_status.display_status[_ts5_status.active_display].h;

	if (x1_abs_disp < 0.0 || x2_abs_disp > dispw || y1_abs_disp < 0.0 || y2_abs_disp > disph) {
		ts5_fatal("ts5_draw_mouse_button: button will be off screen (%f,%f)\n", x1_abs_disp, y1_abs_disp);
	}

	ts5_draw_rounded_rectangle(x1_abs_disp, y1_abs_disp, x2_abs_disp, y2_abs_disp, 3.0, 3.0);
	ts5_set_fill_mode(TS5_FILL_ON);
	ts5_draw_triangle(x2_abs_disp - 5.0, (y1_abs_disp+y2_abs_disp)/2.0,
		x1_abs_disp+5.0, y1_abs_disp+5.0, x1_abs_disp+5.0, y2_abs_disp-5.0);
	ts5_flip_display();
	ts5_show_mouse();

	ALLEGRO_MOUSE_STATE status;
	int down;
	int ok = 0;
	do {
		al_rest(0.01);
		double xp, yp;
		ts5_get_mouse_position(&xp, &yp);
		al_get_mouse_state(&status);
		down = al_mouse_button_down(&status, 1);
		if (down && xp > x1_abs_disp && xp < x2_abs_disp && yp > y1_abs_disp && yp < y2_abs_disp) {
			ok = 1;
		}
	} while (!ok);

	ts5_set_foreground_color(ts5_make_named_color("black", 0.5));
	ts5_draw_triangle(x2_abs_disp - 5, (y1_abs_disp+y2_abs_disp)/2,
		x1_abs_disp+5, y1_abs_disp+5, x1_abs_disp+5, y2_abs_disp-5);
	ts5_flip_display();

	ok = 0;
	do {
		al_rest(0.01);
		al_get_mouse_state(&status);
		down = al_mouse_button_down(&status, 1);
		if (!down) {
			ok = 1;
		}
	} while (!ok);

	ts5_set_foreground_color(ts5_make_named_color("chartreuse", 1.0));
	ts5_draw_triangle(x2_abs_disp - 5, (y1_abs_disp+y2_abs_disp)/2,
		x1_abs_disp+5, y1_abs_disp+5, x1_abs_disp+5, y2_abs_disp-5);
	ts5_flip_display();

	ts5_set_drawing_target(oldmap);
	ts5_set_foreground_color(oldcolor);
	ts5_set_drawing_thickness(oldthick);
	ts5_set_fill_mode(oldfill);
	ts5_set_coordinate_system(oldsystem);
	ts5_set_coordinate_scale(oldscale);
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// The mouse can also be used as a response device. 
/// See timer.c for more information about response registration.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Define a mouse key as a response key.
///
/// \param key Number of the key (keys are counted from 1).
///
/// \return The reponse number associated with the key.
///
/// Give a positive number if you want to monitor key press events,
/// a negative number if you want to monitor key release events.
//////////////////////////////////////////////////////////////////////////////
int ts5_define_mouse_key(int key)
{
	ts5_check_mouse("ts5_define_mouse_key");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_mouse_key(%d)\n", key);

	if (key==0) {
		ts5_fatal("ts5_define_mouse_key: key argument is 0, response keys are numbered from 1\n");
	}

	if (abs(key)>_ts5_status.timer_status.mouse.num_buttons) {
		ts5_fatal("ts5_define_mouse_key: key argument is %d, number of mouse buttons is %d\n",
					key, _ts5_status.timer_status.mouse.num_buttons);
	}

	if (key>0 && _ts5_status.timer_status.mouse.button_press_defined[key-1]!=0) {
		ts5_fatal("ts5_define_mouse_key: key press %d is already defined\n", key);
	}

	if (key<0 && _ts5_status.timer_status.mouse.button_release_defined[-key-1]!=0) {
		ts5_fatal("ts5_define_mouse_key: key release %d is already defined\n", key);
	}

	_ts5_status.timer_status.mouse_is_response_device = 1;
	_ts5_status.timer_status.num_defined_buttons++;
	_ts5_status.timer_status.num_active_buttons++;
	_ts5_status.timer_status.mouse.num_defined_buttons++;
	_ts5_status.timer_status.mouse.num_active_buttons++;

	if (key>0) {
		_ts5_status.timer_status.mouse.button_press_defined[key-1] =
			_ts5_status.timer_status.num_defined_buttons;

		_ts5_status.timer_status.mouse.button_press_active[key-1] =
			_ts5_status.timer_status.num_defined_buttons;
	}
	else {
		_ts5_status.timer_status.mouse.button_release_defined[-key-1] =
			_ts5_status.timer_status.num_defined_buttons;

		_ts5_status.timer_status.mouse.button_release_active[-key-1] =
			_ts5_status.timer_status.num_defined_buttons;
	}

	return _ts5_status.timer_status.num_defined_buttons;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////
