//////////////////////////////////////////////////////////////////////////////
/// \file parport.c
/// Definitions of parallel port functions
//////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/parport.h"
#include "../include/tscope5/timer.h"
#include "../include/tscope5/parport_internal.h"
#include "../include/tscope5/system_internal.h"

#ifdef ALLEGRO_WINDOWS
	#include <windows.h>
	HINSTANCE hLib;
	typedef short (_stdcall *inbptr)(short portaddr);
	typedef short (_stdcall *outbptr)(short portaddr, short value);
	inbptr inb;
	outbptr outb;

#elif defined ALLEGRO_UNIX
	#include <unistd.h>
	#include <sys/io.h>

#endif


//////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// Parallel port boxes can be used as response devices. 
/// See timer.c for more information about response registration.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Define a parallel port key as a response key.
///
/// \param device Number of the parallel port (devices are numbered from 1).
/// \param key Number of the key (keys are counted from 1).
///
/// \return The reponse number associated with the key.
///
/// Give a positive number for the key if you want to monitor key press events,
/// a negative number if you want to monitor key release events.
//////////////////////////////////////////////////////////////////////////////
int ts5_define_parport_key(int device, int key)
{
	ts5_check_parport("ts5_define_parport_key");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_parport_key(%d,%d)\n", device, key);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_define_parport_key: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	if (_ts5_status.timer_status.parport[device-1].is_trigger_input_device) {
		ts5_fatal("ts5_define_parport_key: parallel port %d is already defined as a trigger input device\n",
					device);
	}

	if (_ts5_status.timer_status.parport[device-1].is_trigger_output_device) {
		ts5_fatal("ts5_define_parport_key: parallel port %d is already defined as a trigger output device\n",
					device);
	}

	if (key==0) {
		ts5_fatal("ts5_define_parport_key: key argument is 0, response keys are numbered from 1\n");
	}

	if (abs(key)>_ts5_status.timer_status.parport[device-1].num_buttons) {
		ts5_fatal("ts5_define_parport_key: key argument is %d",
					", number of parport buttons for device %d is %d\n",
					key, device, _ts5_status.timer_status.parport[device-1].num_buttons);
	}

	if (key>0 && _ts5_status.timer_status.parport[device-1].button_press_defined[key-1]!=0) {
		ts5_fatal("ts5_define_parport_key: key press %d is already defined\n", key);
	}

	if (key<0 && _ts5_status.timer_status.parport[device-1].button_release_defined[-key-1]!=0) {
		ts5_fatal("ts5_define_parport_key: key release %d is already defined\n", key);
	}

	_ts5_status.timer_status.parport_is_response_device = 1;
	_ts5_status.timer_status.num_defined_buttons++;
	_ts5_status.timer_status.num_active_buttons++;
	_ts5_status.timer_status.parport[device-1].num_defined_buttons++;
	_ts5_status.timer_status.parport[device-1].num_active_buttons++;

	if (key>0) {
		_ts5_status.timer_status.parport[device-1].button_press_defined[key-1] =
			_ts5_status.timer_status.num_defined_buttons;

		_ts5_status.timer_status.parport[device-1].button_press_active[key-1] =
			_ts5_status.timer_status.num_defined_buttons;
	}
	else {
		_ts5_status.timer_status.parport[device-1].button_release_defined[-key-1] =
			_ts5_status.timer_status.num_defined_buttons;

		_ts5_status.timer_status.parport[device-1].button_release_active[-key-1] =
			_ts5_status.timer_status.num_defined_buttons;
	}

	return _ts5_status.timer_status.num_defined_buttons;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the button debounce time of a parallel port response box.
///
/// \param device Number of the parallel port
/// \param button_debounce_time Debounce time in seconds
///
/// \return The previous setting
///
///
/// The debounce time is the minimum amount of time between two state changes of
/// the buttons of the box before Tscope5 really counts it as a state change.
/// On some button boxes it is necessary to set this value because the press or release
/// of a button causes a ripple in the electrical current sent to the parallel port
/// (instead of a single state change between on and off).
///
/// The debounce time is the amount of time between the press and the release of a button
/// (or vice versa). So the minimum amount of time between two presses (or two releases)
/// will be twice the debounce time set here.
///
/// The default debounce time is set to 0.0
/// (because the button boxes at our department have an internal debouncing circuit).
//////////////////////////////////////////////////////////////////////////////
double ts5_set_parport_button_debounce_time(int device, double button_debounce_time)
{
	ts5_check_parport("ts5_set_parport_button_debounce_time");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_parport_button_debounce_time(%d,%f)\n", device, button_debounce_time);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_set_parport_button_debounce_time: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	if (button_debounce_time < 0.0) {
		ts5_fatal("ts5_set_parport_button_debounce_time: debounce time should be positive (is %f)\n", button_debounce_time);
	}

	double retval = _ts5_status.timer_status.parport[device-1].button_debounce_time;

	_ts5_status.timer_status.parport[device-1].button_debounce_time = button_debounce_time;

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_parport_button_debounce_time: set debounce time of device %d to %f (was %f)\n",
								device, _ts5_status.timer_status.parport[device-1].button_debounce_time, retval);

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the button debounce time of a parallel port response box.
///
/// \param device Number of the parallel port
///
/// \return The debounce time
//////////////////////////////////////////////////////////////////////////////
double ts5_get_parport_button_debounce_time(int device)
{
	ts5_check_parport("ts5_get_parport_button_debounce_time");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_parport_button_debounce_time(%d)\n", device);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_get_parport_button_debounce_time: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	return _ts5_status.timer_status.parport[device-1].button_debounce_time;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Trigger input/output
/// The parallel ports can be used as trigger input/output devices. 
/// See timer.c for more information about trigger input/output.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Define a parallel port as a trigger input device
///
/// \param device Number of the parallel port (devices are numbered from 1).
//////////////////////////////////////////////////////////////////////////////
void ts5_define_parport_trigger_input(int device)
{
	ts5_check_parport("ts5_define_parport_trigger_input");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_parport_trigger_input(%d)\n", device);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_define_parport_trigger_input: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	ts5_set_parport_trigger_input("ts5_define_parport_trigger_input", device);
}


//////////////////////////////////////////////////////////////////////////////
/// Turn on parallel port input simulation.
///
/// \param device Number of the parallel port (devices are numbered from 1).
/// \param value Input value that has to be simulated (0-255, 0 turns the simulation off).
/// \param interval Interval in seconds between two triggers.
//////////////////////////////////////////////////////////////////////////////
void ts5_simulate_parport_trigger_input(int device, unsigned char value, double interval)
{
	ts5_check_parport("ts5_simulate_parport_trigger_input");
	ts5_log(TS5_LOGLEVEL_3, "ts5_simulate_parport_trigger_input(%d,%d,%f)\n", device, value, interval);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_simulate_parport_trigger_input: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	_ts5_status.timer_status.parport[device-1].simulate_trigger_input = value;
	_ts5_status.timer_status.parport[device-1].trigger_simulation_interval = interval;
}


//////////////////////////////////////////////////////////////////////////////
/// Set the trigger debounce time of a parallel port.
///
/// \param device Number of the parallel port
/// \param trigger_debounce_time Debounce time in seconds
///
/// \return The previous setting
//////////////////////////////////////////////////////////////////////////////
double ts5_set_parport_trigger_debounce_time(int device, double trigger_debounce_time)
{
	ts5_check_parport("ts5_set_parport_trigger_debounce_time");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_parport_trigger_debounce_time(%d,%f)\n", device, trigger_debounce_time);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_set_parport_trigger_debounce_time: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	if (trigger_debounce_time < 0.0) {
		ts5_fatal("ts5_set_parport_trigger_debounce_time: debounce time should be positive (is %f)\n", trigger_debounce_time);
	}

	double retval = _ts5_status.timer_status.parport[device-1].trigger_debounce_time;

	_ts5_status.timer_status.parport[device-1].trigger_debounce_time = trigger_debounce_time;

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_parport_trigger_debounce_time: set debounce time of device %d to %f (was %f)\n",
								device, _ts5_status.timer_status.parport[device-1].trigger_debounce_time, retval);

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the trigger debounce time of a parallel port.
///
/// \param device Number of the parallel port
///
/// \return The debounce time
//////////////////////////////////////////////////////////////////////////////
double ts5_get_parport_trigger_debounce_time(int device)
{
	ts5_check_parport("ts5_get_parport_trigger_debounce_time");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_parport_trigger_debounce_time(%d)\n", device);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_get_parport_trigger_debounce_time: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	return _ts5_status.timer_status.parport[device-1].trigger_debounce_time;
}


//////////////////////////////////////////////////////////////////////////////
/// Send a trigger trough a parallel port
///
/// \param device Number of the parallel port (devices are numbered from 1).
/// \param value The trigger value that has to be sent (0-255).
//////////////////////////////////////////////////////////////////////////////
void ts5_send_parport_trigger(int device, unsigned char value)
{
	ts5_check_parport("ts5_send_parport_trigger");
	ts5_log(TS5_LOGLEVEL_3, "ts5_send_parport_trigger(%d,%d)\n", device, value);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_send_parport_trigger: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	ts5_set_parport_trigger_output("ts5_send_parport_trigger", device);

	#ifndef ALLEGRO_MACOSX
		short data_register[] = {TS5_PARPORT_1_DATA, TS5_PARPORT_2_DATA, TS5_PARPORT_3_DATA};

	#endif

	#ifdef ALLEGRO_WINDOWS
		outb((short)data_register[device-1], value);
		ts5_wait(_ts5_status.timer_status.parport[device-1].trigger_output_time);
		outb((short)data_register[device-1], 0);

	#elif defined ALLEGRO_UNIX
		outb(value, (short)data_register[device-1]);
		ts5_wait(_ts5_status.timer_status.parport[device-1].trigger_output_time);
		outb(0, (short)data_register[device-1]);

	#endif
}


//////////////////////////////////////////////////////////////////////////////
/// Set the trigger output time of a parallel port.
///
/// \param device Number of the parallel port
/// \param trigger_output_time Output time in seconds
///
/// \return The previous setting
///
/// The default output time is 0.002 seconds.
//////////////////////////////////////////////////////////////////////////////
double ts5_set_parport_trigger_output_time(int device, double trigger_output_time)
{
	ts5_check_parport("ts5_set_parport_trigger_output_time");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_parport_trigger_output_time(%d,%f)\n", device, trigger_output_time);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_set_parport_trigger_output_time: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	if (trigger_output_time < 0.0) {
		ts5_fatal("ts5_set_parport_trigger_output_time: output time should be positive (is %f)\n", trigger_output_time);
	}

	double retval = _ts5_status.timer_status.parport[device-1].trigger_output_time;

	_ts5_status.timer_status.parport[device-1].trigger_output_time = trigger_output_time;

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_parport_trigger_output_time: set output time of device %d to %f (was %f)\n",
								device, _ts5_status.timer_status.parport[device-1].trigger_output_time, retval);

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get the trigger output time of a parallel port.
///
/// \param device Number of the parallel port
///
/// \return The output time
//////////////////////////////////////////////////////////////////////////////
double ts5_get_parport_trigger_output_time(int device)
{
	ts5_check_parport("ts5_get_parport_trigger_output_time");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_parport_trigger_output_time(%d)\n", device);

	if (device<1 || device>_ts5_status.timer_status.num_parport) {
		ts5_fatal("ts5_get_parport_trigger_output_time: device argument is %d, number of parallel ports is %d\n",
					device, _ts5_status.timer_status.num_parport);
	}

	return _ts5_status.timer_status.parport[device-1].trigger_output_time;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////
