//////////////////////////////////////////////////////////////////////////////
/// \file parport_internal.c
/// Definitions of internal parallel port functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/parport_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"

/// Is the parport subsystem installed?
int  _ts5_is_parport_installed = 0;

#ifdef ALLEGRO_WINDOWS
	#include <windows.h>
	HINSTANCE hLib;
	typedef short (_stdcall *inbptr)(short portaddr);
	typedef short (_stdcall *outbptr)(short portaddr, short value);
	inbptr inb;
	outbptr outb;

#elif defined ALLEGRO_UNIX
	#include <unistd.h>
	#include <sys/io.h>

#endif


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each parport function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether the parport subsystem is installed. If not, the parport subsystem is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_parport(char *calling_function)
{
	if (!_ts5_is_parport_installed) {
		ts5_install_parport(calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_parport\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Install the parport subsystem.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// This function is called automatically if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_install_parport(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_parport_installed) {

		_ts5_is_parport_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 parport\n", calling_function);

		#ifdef ALLEGRO_MACOSX
		//	ts5_fatal("%s: parallel port boxes not supported on mac os x\n", calling_function);

		#endif

		#ifdef ALLEGRO_WINDOWS
			if (!(hLib = LoadLibrary("inpout32.dll"))) {
				ts5_fatal("%s: could not load inpout32.dll\n", calling_function);
			}

			if (!(inb = (inbptr)GetProcAddress(hLib, "Inp32"))) {
				ts5_fatal("%s: could not get address for Inp32\n", calling_function);
			}

			if (!(outb = (outbptr)GetProcAddress(hLib, "Out32"))) {
				ts5_fatal("%s: could not get address for Out32\n", calling_function);
			}

			outb((short)TS5_PARPORT_1_CONTROL, 0);
			outb((short)TS5_PARPORT_2_CONTROL, 0);
			outb((short)TS5_PARPORT_3_CONTROL, 0);

		#endif

		#ifdef ALLEGRO_UNIX
			if (geteuid() != 0) {
				ts5_fatal("%s: only root can open the parallel port.", calling_function);
			}

			if (ioperm(TS5_PARPORT_1_DATA, 3, 1)) {
				ts5_fatal("%s: error opening parallell port 1.", calling_function );
			}

			if (ioperm(TS5_PARPORT_2_DATA, 3, 1)) {
				ts5_fatal("%s: error opening parallell port 2.", calling_function);
			}

			if (ioperm(TS5_PARPORT_3_DATA, 3, 1)) {
				ts5_fatal("%s: error opening parallell port 3.", calling_function);
			}

			outb(0, (short)TS5_PARPORT_1_CONTROL);
			outb(0, (short)TS5_PARPORT_2_CONTROL);
			outb(0, (short)TS5_PARPORT_3_CONTROL);

		#endif

		// unitialize event source
		al_init_user_event_source(&_ts5_status.timer_status.parport_response_event_source);

		al_init_user_event_source(&_ts5_status.timer_status.parport_trigger_event_source);

		// register event source
		al_register_event_source(_ts5_status.timer_status.response_queue,
									&_ts5_status.timer_status.parport_response_event_source);

		al_register_event_source(_ts5_status.timer_status.trigger_queue,
									&_ts5_status.timer_status.parport_trigger_event_source);

		al_register_event_source(_ts5_status.timer_status.trigger_log,
									&_ts5_status.timer_status.parport_trigger_event_source);

		// initialize thread for parport
		_ts5_status.timer_status.parport_thread = al_create_thread(ts5_parport_threadfunc, NULL);
		al_start_thread(_ts5_status.timer_status.parport_thread);

		// assign respons codes
		_ts5_status.timer_status.num_parport = 3;
		_ts5_status.timer_status.parport =
			(TS5_PARPORT_STATUS *)al_malloc(sizeof(TS5_PARPORT_STATUS)*
			_ts5_status.timer_status.num_parport);

		int i;
		for (i=0; i<_ts5_status.timer_status.num_parport; i++) {

			_ts5_status.timer_status.parport[i].num_buttons = 5;
			_ts5_status.timer_status.parport[i].num_defined_buttons = 0;
			_ts5_status.timer_status.parport[i].num_active_buttons = 0;
			_ts5_status.timer_status.parport[i].button_debounce_time = 0.0;

			_ts5_status.timer_status.parport[i].status_register_current_state =
				(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.parport[i].num_buttons);

			_ts5_status.timer_status.parport[i].status_register_previous_state =
				(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.parport[i].num_buttons);

			_ts5_status.timer_status.parport[i].status_register_last_state_toggle_time =
				(double *)al_malloc(sizeof(double)*_ts5_status.timer_status.parport[i].num_buttons);

			_ts5_status.timer_status.parport[i].button_press_defined =
				(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.parport[i].num_buttons);

			_ts5_status.timer_status.parport[i].button_press_active =
				(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.parport[i].num_buttons);

			_ts5_status.timer_status.parport[i].button_release_defined =
				(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.parport[i].num_buttons);

			_ts5_status.timer_status.parport[i].button_release_active =
				(int *)al_malloc(sizeof(int)*_ts5_status.timer_status.parport[i].num_buttons);

			int j;
			for (j=0; j<_ts5_status.timer_status.parport[i].num_buttons; j++) {
				_ts5_status.timer_status.parport[i].status_register_current_state[j] = 0;
				_ts5_status.timer_status.parport[i].status_register_previous_state[j] = 0;
				_ts5_status.timer_status.parport[i].status_register_last_state_toggle_time[j] = al_get_time();
				_ts5_status.timer_status.parport[i].button_press_defined[j] = 0;
				_ts5_status.timer_status.parport[i].button_press_active[j] = 0;
				_ts5_status.timer_status.parport[i].button_release_defined[j] = 0;
				_ts5_status.timer_status.parport[i].button_release_active[j] = 0;
			}

			_ts5_status.timer_status.parport[i].is_trigger_input_device = 0;
			_ts5_status.timer_status.parport[i].is_trigger_output_device = 0;
			_ts5_status.timer_status.parport[i].trigger_debounce_time = 0.0;

			_ts5_status.timer_status.parport[i].simulate_trigger_input = 0;
			_ts5_status.timer_status.parport[i].trigger_simulation_interval = 2.0;
			_ts5_status.timer_status.parport[i].last_trigger_simulation = al_get_time();

			_ts5_status.timer_status.parport[i].data_register_current_state = 0;
			_ts5_status.timer_status.parport[i].data_register_previous_state = 0;
			_ts5_status.timer_status.parport[i].trigger_output_time = 0.002;
		}

		atexit(ts5_uninstall_parport);
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Uninstall the parport subsystem.
///
/// This function is called automatically at the end of the program.
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_parport()
{
	if (_ts5_is_parport_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 parport\n");

		// remove thread
		ts5_log(TS5_LOGLEVEL_1, "Remove thread\n");
		al_join_thread(_ts5_status.timer_status.parport_thread, NULL);
		al_destroy_thread(_ts5_status.timer_status.parport_thread);

		// unregister event source
		ts5_log(TS5_LOGLEVEL_1, "unregister event source\n");
		al_unregister_event_source(_ts5_status.timer_status.response_queue,
									&_ts5_status.timer_status.parport_response_event_source);

		al_unregister_event_source(_ts5_status.timer_status.trigger_queue,
									&_ts5_status.timer_status.parport_trigger_event_source);

		al_unregister_event_source(_ts5_status.timer_status.trigger_log,
									&_ts5_status.timer_status.parport_trigger_event_source);

		// remove event source
		ts5_log(TS5_LOGLEVEL_1, "Remove event source parport\n");
		al_destroy_user_event_source(&_ts5_status.timer_status.parport_response_event_source);
		al_destroy_user_event_source(&_ts5_status.timer_status.parport_trigger_event_source);

		#ifdef ALLEGRO_WINDOWS
			outb((short)TS5_PARPORT_1_CONTROL, 0);
			outb((short)TS5_PARPORT_2_CONTROL, 0);
			outb((short)TS5_PARPORT_3_CONTROL, 0);

			FreeLibrary(hLib);

		#endif

		#ifdef ALLEGRO_UNIX
			outb(0, (short)TS5_PARPORT_1_CONTROL);
			outb(0, (short)TS5_PARPORT_2_CONTROL);
			outb(0, (short)TS5_PARPORT_3_CONTROL);

			if (geteuid() != 0) {
				ts5_fatal("ts5_uninstall_parport: only root can close the parallel port.");
			}

			if (ioperm(TS5_PARPORT_1_DATA, 3, 0)) {
				ts5_fatal("ts5_uninstall_parport: error closing parallell port 1.");
			}

			if (ioperm(TS5_PARPORT_2_DATA, 3, 0)) {
				ts5_fatal("ts5_uninstall_parport: error closing parallell port 2.");
			}

			if (ioperm(TS5_PARPORT_3_DATA, 3, 0)) {
				ts5_fatal("ts5_uninstall_parport: error closing parallell port 3.");
			}

		#endif

		ts5_log(TS5_LOGLEVEL_1, "Free memory\n");

		int i;
		for (i=0; i<_ts5_status.timer_status.num_parport; i++) {

			al_free(_ts5_status.timer_status.parport[i].status_register_current_state);
			_ts5_status.timer_status.parport[i].status_register_current_state = NULL;

			al_free(_ts5_status.timer_status.parport[i].status_register_previous_state);
			_ts5_status.timer_status.parport[i].status_register_previous_state = NULL;

			al_free(_ts5_status.timer_status.parport[i].status_register_last_state_toggle_time);
			_ts5_status.timer_status.parport[i].status_register_last_state_toggle_time = NULL;

			al_free(_ts5_status.timer_status.parport[i].button_press_defined);
			_ts5_status.timer_status.parport[i].button_press_defined = NULL;

			al_free(_ts5_status.timer_status.parport[i].button_press_active);
			_ts5_status.timer_status.parport[i].button_press_active = NULL;

			al_free(_ts5_status.timer_status.parport[i].button_release_defined);
			_ts5_status.timer_status.parport[i].button_release_defined = NULL;

			al_free(_ts5_status.timer_status.parport[i].button_release_active);
			_ts5_status.timer_status.parport[i].button_release_active = NULL;
		}

		al_free(_ts5_status.timer_status.parport);
		_ts5_status.timer_status.parport=NULL;

		_ts5_status.timer_status.num_parport = 0;
		_ts5_status.timer_status.parport_is_response_device = 0;

		_ts5_is_parport_installed = 0;
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Set the DATA register of the parallel port to input mode.
///
/// \param calling_function Name the function that calls for this check or installation.
/// \param device Number of the parallel port.
//////////////////////////////////////////////////////////////////////////////
void ts5_set_parport_trigger_input(char *calling_function, int device)
{
	ts5_check_parport(calling_function);

	if (!_ts5_status.timer_status.parport[device-1].is_trigger_input_device) {

		ts5_log(TS5_LOGLEVEL_1, "%s: Set parallel port %d to input mode\n", calling_function, device);

		if (device<1 || device>_ts5_status.timer_status.num_parport) {
			ts5_fatal("%s: device argument is %d, number of parallel ports is %d\n",
						calling_function, device, _ts5_status.timer_status.num_parport);
		}

		if (_ts5_status.timer_status.parport[device-1].num_defined_buttons) {
			ts5_fatal("%s: parallel port %d is already defined as an response device\n",
						calling_function, device);
		}

		if (_ts5_status.timer_status.parport[device-1].is_trigger_output_device) {
			ts5_fatal("%s: parallel port %d is already defined as a trigger output device\n",
						calling_function, device);
		}

		_ts5_status.timer_status.parport_is_trigger_device = 1;
		_ts5_status.timer_status.parport[device-1].is_trigger_input_device = 1;

		#ifndef ALLEGRO_MACOSX
			short control_register[] = {TS5_PARPORT_1_CONTROL, TS5_PARPORT_2_CONTROL, TS5_PARPORT_3_CONTROL};

		#endif

		#ifdef ALLEGRO_WINDOWS
			outb((short)control_register[device-1], 32);

		#elif defined ALLEGRO_UNIX
			outb(32, (short)control_register[device-1]);

		#endif

	}
}


//////////////////////////////////////////////////////////////////////////////
/// Set the DATA register of the parallel port to output mode.
///
/// \param calling_function Name the function that calls for this check or installation.
/// \param device Number of the parallel port.
//////////////////////////////////////////////////////////////////////////////
void ts5_set_parport_trigger_output(char *calling_function, int device)
{
	ts5_check_parport(calling_function);

	if (!_ts5_status.timer_status.parport[device-1].is_trigger_output_device) {

		ts5_log(TS5_LOGLEVEL_1, "%s: Set parallel port %d to output mode\n", calling_function, device);

		if (device<1 || device>_ts5_status.timer_status.num_parport) {
			ts5_fatal("%s: device argument is %d, number of parallel ports is %d\n",
						calling_function, device, _ts5_status.timer_status.num_parport);
		}

		if (_ts5_status.timer_status.parport[device-1].num_defined_buttons) {
			ts5_fatal("%s: parallel port %d is already defined as an response device\n",
						calling_function, device);
		}

		if (_ts5_status.timer_status.parport[device-1].is_trigger_input_device) {
			ts5_fatal("%s: parallel port %d is already defined as a trigger input device\n",
						calling_function, device);
		}

		_ts5_status.timer_status.parport_is_trigger_device = 1;
		_ts5_status.timer_status.parport[device-1].is_trigger_output_device = 1;

		#ifndef ALLEGRO_MACOSX
			short control_register[] = {TS5_PARPORT_1_CONTROL, TS5_PARPORT_2_CONTROL, TS5_PARPORT_3_CONTROL};

		#endif

		#ifdef ALLEGRO_WINDOWS
			outb((short)control_register[device-1], 0);

		#elif defined ALLEGRO_UNIX
			outb(0, (short)control_register[device-1]);

		#endif

	}
}


//////////////////////////////////////////////////////////////////////////////
/// Parallel port thread function.
///
/// \warning This is an internal function, no checks are performed.
//////////////////////////////////////////////////////////////////////////////
void *ts5_parport_threadfunc()
{
	ALLEGRO_EVENT parport_event;
	double looptime = al_get_time();
	double now;

	while (!al_get_thread_should_stop(_ts5_status.timer_status.parport_thread)) {

		int i;
		for (i=0; i<_ts5_status.timer_status.num_parport; i++) {

			now = al_get_time();

			// monitor STATUS register for responses
			short input = 0;

			#ifndef ALLEGRO_MACOSX
				short status_register[] = {TS5_PARPORT_1_STATUS, TS5_PARPORT_2_STATUS, TS5_PARPORT_3_STATUS};
				input = inb((short)status_register[i]);

			#endif

			_ts5_status.timer_status.parport[i].status_register_current_state[0] = (input &	8) >> 3;
			_ts5_status.timer_status.parport[i].status_register_current_state[1] = (input &  16) >> 4;
			_ts5_status.timer_status.parport[i].status_register_current_state[2] = (input &  32) >> 5;
			_ts5_status.timer_status.parport[i].status_register_current_state[3] = (input &  64) >> 6;
			_ts5_status.timer_status.parport[i].status_register_current_state[4] = (input & 128) >> 7;

			int button = 0;
			int action = 0;

			int j;
			for (j=0; j<_ts5_status.timer_status.parport[i].num_buttons; j++) {

				if (_ts5_status.timer_status.parport[i].status_register_current_state[j] !=
					_ts5_status.timer_status.parport[i].status_register_previous_state[j]) {

					if (now - _ts5_status.timer_status.parport[i].status_register_last_state_toggle_time[j] >
						_ts5_status.timer_status.parport[i].button_debounce_time) {

						button = j+1;
						action = _ts5_status.timer_status.parport[i].status_register_current_state[j];

						_ts5_status.timer_status.parport[i].status_register_previous_state[j] =
							_ts5_status.timer_status.parport[i].status_register_current_state[j];

						_ts5_status.timer_status.parport[i].status_register_last_state_toggle_time[j] = now;
					}
				}
			}

			if (button) {

				if (action==0) {
					parport_event.user.type = TS5_EVENT_PARPORT_BUTTON_UP;
				}

				if (action==1) {
					parport_event.user.type = TS5_EVENT_PARPORT_BUTTON_DOWN;
				}

				parport_event.user.data1 = i;
				parport_event.user.data2 = button;
				parport_event.user.data3 =  (intptr_t)((now-looptime)*1000000.0);

				al_emit_user_event(&_ts5_status.timer_status.parport_response_event_source,
									&parport_event, NULL);
			}


			// monitor DATA register for triggers
			input = 0;

			#ifndef ALLEGRO_MACOSX
				short data_register[] = {TS5_PARPORT_1_DATA, TS5_PARPORT_2_DATA, TS5_PARPORT_3_DATA};
				input = inb((short)data_register[i]);

			#endif

			_ts5_status.timer_status.parport[i].data_register_current_state = input;

			if (_ts5_status.timer_status.parport[i].is_trigger_input_device &&
					_ts5_status.timer_status.parport[i].data_register_current_state &&
					_ts5_status.timer_status.parport[i].data_register_current_state !=
					_ts5_status.timer_status.parport[i].data_register_previous_state) {

				if (now - _ts5_status.timer_status.parport[i].data_register_last_state_toggle_time >
						_ts5_status.timer_status.parport[i].trigger_debounce_time) {

					parport_event.user.type = TS5_EVENT_PARPORT_TRIGGER;

					parport_event.user.data1 = i;
					parport_event.user.data2 = (int)_ts5_status.timer_status.parport[i].data_register_current_state;
					parport_event.user.data3 = (intptr_t)((now-looptime)*1000000.0);

					al_emit_user_event(&_ts5_status.timer_status.parport_trigger_event_source,
										&parport_event, NULL);

					_ts5_status.timer_status.parport[i].data_register_previous_state =
						_ts5_status.timer_status.parport[i].data_register_current_state;

					_ts5_status.timer_status.parport[i].data_register_last_state_toggle_time = now;
				}
			}

			// simulate triggers
			if (_ts5_status.timer_status.parport[i].simulate_trigger_input) {

				if (now - _ts5_status.timer_status.parport[i].last_trigger_simulation
						>= _ts5_status.timer_status.parport[i].trigger_simulation_interval) {

					parport_event.user.type = TS5_EVENT_PARPORT_TRIGGER;

					parport_event.user.data1 = i;
					parport_event.user.data2 = _ts5_status.timer_status.parport[i].simulate_trigger_input;
					parport_event.user.data3 = (intptr_t)((now-looptime)*1000000.0);

					al_emit_user_event(&_ts5_status.timer_status.parport_trigger_event_source,
										&parport_event, NULL);

					_ts5_status.timer_status.parport[i].last_trigger_simulation = now;
				}
			}
		}
		
		looptime = now;
	}

	return NULL;
}


