//////////////////////////////////////////////////////////////////////////////
/// \file serialport.c
/// Definitions of serial port functions
//////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/serialport.h"
#include "../include/tscope5/timer.h"
#include "../include/tscope5/serialport_internal.h"
#include "../include/tscope5/system_internal.h"


//////////////////////////////////////////////////////////////////////////////
/// @name Trigger input/output
/// The serial ports can be used as trigger input/output devices. 
/// See timer.c for more information about trigger input/output.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Define a serial port as a trigger input device
///
/// \param portname Name of the serial port (COM1, ... on Windows, /etc/tty... on Posix systems).
///
/// \return The device number of the port (the first serial port opened by Tscope5 gets number 1, etc.).
//////////////////////////////////////////////////////////////////////////////
int ts5_define_serialport_trigger_input(char *portname)
{
	ts5_check_serialport("ts5_define_serialport_trigger_input");
	int portnum = ts5_check_serialport2("ts5_define_serialport_trigger_input", portname);
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_serialport_trigger_input(%s)\n", portname);

	_ts5_status.timer_status.serialport_is_trigger_device = 1;
	_ts5_status.timer_status.serialport[portnum].is_trigger_input_device = 1;	
	return portnum+1;
}


//////////////////////////////////////////////////////////////////////////////
/// Turn on serial port input simulation.
///
/// \param device Number of the serial port.
/// \param value Input value that has to be simulated (0-255, 0 turns the simulation off).
/// \param interval Interval in seconds between two triggers.
//////////////////////////////////////////////////////////////////////////////
void ts5_simulate_serialport_trigger_input(int device, unsigned char value, double interval)
{
	ts5_check_serialport("ts5_simulate_serialport_trigger_input");
	ts5_log(TS5_LOGLEVEL_3, "ts5_simulate_serialport_trigger_input(%d,%d,%f)\n", device, value, interval);

	if (device<1 || device>_ts5_status.timer_status.num_serialport) {
		ts5_fatal("ts5_simulate_serialport_trigger_input: device argument is %d, number of serial ports is %d\n",
					device, _ts5_status.timer_status.num_serialport);
	}

	_ts5_status.timer_status.serialport[device-1].simulate_trigger_input = value;
	_ts5_status.timer_status.serialport[device-1].trigger_simulation_interval = interval;
}


//////////////////////////////////////////////////////////////////////////////
/// Define a serial port as a trigger output device
///
/// \param portname Name of the serial port (COM1, ... on Windows, /etc/tty... on Posix systems).
///
/// \return The device number of the port (the first serial port opened by Tscope5 gets number 1, etc.).
//////////////////////////////////////////////////////////////////////////////
int ts5_define_serialport_trigger_output(char *portname)
{
	ts5_check_serialport("ts5_define_serialport_trigger_output");
	int portnum = ts5_check_serialport2("ts5_define_serialport_trigger_output", portname);
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_serialport_trigger_output(%s)\n", portname);

	_ts5_status.timer_status.serialport_is_trigger_device = 1;
	_ts5_status.timer_status.serialport[portnum].is_trigger_output_device = 1;	
	return portnum+1;
}


//////////////////////////////////////////////////////////////////////////////
/// Send a trigger trough a serial port
///
/// \param device Number of the serial port (devices are numbered from 1).
/// \param value The trigger value that has to be sent (0-255).
//////////////////////////////////////////////////////////////////////////////
void ts5_send_serialport_trigger(int device, unsigned char value)
{
	ts5_check_serialport("ts5_send_serialport_trigger");
	ts5_log(TS5_LOGLEVEL_3, "ts5_send_serialport_trigger(%d,%d)\n", device, value);

	if (device<1 || device>_ts5_status.timer_status.num_serialport) {
		ts5_fatal("ts5_send_serialport_trigger: device argument is %d, number of serial ports is %d\n",
					device, _ts5_status.timer_status.num_serialport);
	}

	char serial_buff[TS5_MAX_CHAR];
	sprintf(serial_buff, "%c", value);

	#ifdef ALLEGRO_WINDOWS
		
		unsigned long bytes_to_write = strlen(serial_buff);
		unsigned long bytes_written;

		if (!WriteFile(*(_ts5_status.timer_status.serialport[device-1].port), serial_buff, bytes_to_write, &bytes_written, NULL)) {
			ts5_fatal("ts5_send_serialport_trigger: write error 1\n");
		}

		if (bytes_written != bytes_to_write) {
			ts5_fatal("ts5_send_serialport_trigger: write error 2\n");
		}
	

	#else

		unsigned int bytes_to_write = strlen(serial_buff);
		unsigned int bytes_written;

		bytes_written = write(*(_ts5_status.timer_status.serialport[device-1].port), serial_buff, bytes_to_write);

		if (bytes_written != bytes_to_write) {
			ts5_fatal("ts5_send_serialport_trigger: write error\n");
		}
	

	#endif
}


//////////////////////////////////////////////////////////////////////////////
/// Get a pointer to a serial port
///
/// \param device Number of the serial port (devices are numbered from 1).
///
/// \return pointer to the serial port.
//////////////////////////////////////////////////////////////////////////////
TS5_SERIALPORT*	ts5_get_serialport(int device)
{
	ts5_check_serialport("ts5_get_serialport");
	ts5_log(TS5_LOGLEVEL_3, "ts5_get_serialport(%d)\n", device);
	
	if (device < 1 || device>_ts5_status.timer_status.num_serialport) {
		ts5_fatal("%s: device argument is %d, number of opened serial ports is %d\n", device, _ts5_status.timer_status.num_serialport);
	}
	
	return _ts5_status.timer_status.serialport[device-1].port;
}



//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////
