//////////////////////////////////////////////////////////////////////////////
/// \file serialport_internal.c
/// Definitions of internal serial port functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/serialport_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"

#ifndef ALLEGRO_WINDOWS
	#include <sys/types.h>
	#include <sys/ioctl.h>
	#include <dirent.h>
	#include <termios.h>
	#include <errno.h>

#endif

/// Is the serialport subsystem installed?
int  _ts5_is_serialport_installed = 0;


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each serialport function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether the serialport subsystem is installed. If not, the serialport subsystem is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_serialport(char *calling_function)
{
	if (!_ts5_is_serialport_installed) {
		ts5_install_serialport(calling_function);
	}
	
	_ts5_status.timer_status.serialport_is_trigger_device = 1;
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_serialport\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each serialport function.
///
/// \param calling_function Name the function that calls for this check or installation.
/// \param portname Name of the port.
///
/// Checks whether this serial port is already opened. If so, return the port number. 
/// If not, open it and return the port number. 
//////////////////////////////////////////////////////////////////////////////
int ts5_check_serialport2(char *calling_function, char *portname)
{
	int i;
	int portnum = -1;
	
	// is the port already open? if so return the port index
	ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 serialport: check if port is open\n", calling_function);
	for (i=0; i<_ts5_status.timer_status.num_serialport; i++) {
		if (!strcmp(portname, _ts5_status.timer_status.serialport[i].portname)) {
			portnum = i;
		} 		
	}	

	// if the port is not open try to open it
	if (portnum == -1) {

		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 serialport: trying to open port\n", calling_function);
		TS5_SERIALPORT *port;
		port = (TS5_SERIALPORT *)al_malloc(sizeof(TS5_SERIALPORT));
		
		#ifdef ALLEGRO_WINDOWS
			*port = CreateFile(portname,
						GENERIC_READ | GENERIC_WRITE,
						0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
						
			if (&port == INVALID_HANDLE_VALUE) {
				ts5_fatal("%s: could not open serial port %s\n", calling_function, portname);
			}
										
		#else
			*port = open(portname, O_RDWR | O_NOCTTY | O_NDELAY);

			if (*port == -1) {
				ts5_fatal("%s: could not open serial port %s\n", calling_function, portname);
			}

			// get exclusive io
			if (ioctl(*port, TIOCEXCL) == -1) {
				ts5_fatal("%s: could not get exclusivity serial port %s\n",
					calling_function, portname);
			}		
		
		#endif
		
		// if we get here we apparently managed to open the port
		_ts5_status.timer_status.num_serialport++;
		portnum = _ts5_status.timer_status.num_serialport - 1;
		
		// update status pointer
		TS5_SERIALPORT_STATUS *serialport_status;
		serialport_status = al_realloc(_ts5_status.timer_status.serialport, _ts5_status.timer_status.num_serialport * sizeof(TS5_SERIALPORT_STATUS));

		if (serialport_status == NULL) {
			ts5_fatal("%s: failed to allocate memory for serial port settings\n", calling_function);
		}

		_ts5_status.timer_status.serialport = (TS5_SERIALPORT_STATUS *) serialport_status;
		
		// fill in status values
		strcpy(_ts5_status.timer_status.serialport[portnum].portname, portname);
		_ts5_status.timer_status.serialport[portnum].port = port;
		_ts5_status.timer_status.serialport[portnum].is_trigger_input_device = 0;
		_ts5_status.timer_status.serialport[portnum].is_trigger_output_device = 0;
		_ts5_status.timer_status.serialport[portnum].simulate_trigger_input = 0;
		_ts5_status.timer_status.serialport[portnum].trigger_simulation_interval = 2.0;
		_ts5_status.timer_status.serialport[portnum].last_trigger_simulation = al_get_time();
		
	}
		
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_serialport2\n", calling_function);
	return portnum;
}


//////////////////////////////////////////////////////////////////////////////
/// Install the serialport subsystem.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// This function is called automatically if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_install_serialport(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_serialport_installed) {

		_ts5_is_serialport_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 serialport\n", calling_function);

		// unitialize event source
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 serialport: initializing event source\n", calling_function);
		al_init_user_event_source(&_ts5_status.timer_status.serialport_trigger_event_source);

		// register event source
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 serialport: registering event source\n", calling_function);
		al_register_event_source(_ts5_status.timer_status.trigger_queue,
									&_ts5_status.timer_status.serialport_trigger_event_source);

		al_register_event_source(_ts5_status.timer_status.trigger_log,
									&_ts5_status.timer_status.serialport_trigger_event_source);

		// initialize thread for serialport
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 serialport: initializing thread\n", calling_function);
		_ts5_status.timer_status.serialport_thread = al_create_thread(ts5_serialport_threadfunc, NULL);
		al_start_thread(_ts5_status.timer_status.serialport_thread);

		atexit(ts5_uninstall_serialport);
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Uninstall the serialport subsystem.
///
/// This function is called automatically at the end of the program.
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_serialport()
{
	if (_ts5_is_serialport_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 serialport\n");

		// remove thread
		ts5_log(TS5_LOGLEVEL_1, "Remove thread\n");
		al_join_thread(_ts5_status.timer_status.serialport_thread, NULL);
		al_destroy_thread(_ts5_status.timer_status.serialport_thread);

		// unregister event source
		ts5_log(TS5_LOGLEVEL_1, "unregister event source\n");
		al_unregister_event_source(_ts5_status.timer_status.trigger_queue,
									&_ts5_status.timer_status.serialport_trigger_event_source);

		al_unregister_event_source(_ts5_status.timer_status.trigger_log,
									&_ts5_status.timer_status.serialport_trigger_event_source);

		// remove event source
		ts5_log(TS5_LOGLEVEL_1, "Remove event source serialport\n");
		al_destroy_user_event_source(&_ts5_status.timer_status.serialport_trigger_event_source);

		// close ports
		int i;
		for (i=0; i<_ts5_status.timer_status.num_serialport; i++) {
		
			#ifdef ALLEGRO_WINDOWS
				if (!CloseHandle(*(_ts5_status.timer_status.serialport[i].port))) {
					ts5_log(TS5_LOGLEVEL_1,"ts5_uninstall_serialport:\n");
					ts5_log(TS5_LOGLEVEL_1,"\tcould not uninstall serial port %d\n", i);
				}

			#else
				tcdrain(*(_ts5_status.timer_status.serialport[i].port));

				if (close(*(_ts5_status.timer_status.serialport[i].port)) == -1) {
					ts5_log(TS5_LOGLEVEL_1, "ts5_uninstall_serialport:\n");
					ts5_log(TS5_LOGLEVEL_1, "\tcould not uninstall serial port %d\n", i);
				}
						
			#endif
			
			al_free(_ts5_status.timer_status.serialport[i].port);
		}
		
		al_free(_ts5_status.timer_status.serialport);
		
		_ts5_status.timer_status.serialport_is_trigger_device = 0;
	
		_ts5_is_serialport_installed = 0;
	}
}



//////////////////////////////////////////////////////////////////////////////
/// Serial port thread function.
///
/// \warning This is an internal function, no checks are performed.
//////////////////////////////////////////////////////////////////////////////
void *ts5_serialport_threadfunc()
{
	ALLEGRO_EVENT serialport_event;
	double looptime = al_get_time();
	double now;

	while (!al_get_thread_should_stop(_ts5_status.timer_status.serialport_thread)) {

		int i;
		for (i=0; i<_ts5_status.timer_status.num_serialport; i++) {

			now = al_get_time();

			// monitor triggers
			if (_ts5_status.timer_status.serialport[i].is_trigger_input_device) {
				
				char serial_buff[TS5_MAX_CHAR];
				int num_bytes;
				
				#ifdef ALLEGRO_WINDOWS
					unsigned long bytes_to_read=1, bytes_read;
					
					if (!ReadFile(*(_ts5_status.timer_status.serialport[i].port), serial_buff, bytes_to_read, &bytes_read, NULL)) {
						ts5_fatal("ts5_serialport_threadfunc: read error\n");
					}
					
					num_bytes = bytes_read;

				#else
					unsigned int bytes_to_read=1, bytes_read;
					
					bytes_read = read(*(_ts5_status.timer_status.serialport[i].port), serial_buff, bytes_to_read);

					if ((int)bytes_read == -1) {
						ts5_fatal("ts5_serialport_threadfunc: read error\n");
					}
				
					num_bytes = bytes_read;

				#endif
			
				if (num_bytes) {
					
					serialport_event.user.type = TS5_EVENT_SERIALPORT_TRIGGER;

					serialport_event.user.data1 = i;
					serialport_event.user.data2 = serial_buff[0];
					serialport_event.user.data3 = (intptr_t)((now-looptime)*1000000.0);				

					al_emit_user_event(&_ts5_status.timer_status.serialport_trigger_event_source,
										&serialport_event, NULL);
				}
			}

			// simulate triggers
			if (_ts5_status.timer_status.serialport[i].simulate_trigger_input) {

				if (now - _ts5_status.timer_status.serialport[i].last_trigger_simulation
						>= _ts5_status.timer_status.serialport[i].trigger_simulation_interval) {

					serialport_event.user.type = TS5_EVENT_SERIALPORT_TRIGGER;

					serialport_event.user.data1 = i;
					serialport_event.user.data2 = _ts5_status.timer_status.serialport[i].simulate_trigger_input;
					serialport_event.user.data3 = (intptr_t)((now-looptime)*1000000.0);

					al_emit_user_event(&_ts5_status.timer_status.serialport_trigger_event_source,
										&serialport_event, NULL);

					_ts5_status.timer_status.serialport[i].last_trigger_simulation = now;
				}
			}
	
		}
		looptime = now;
	}

	return NULL;
}


