//////////////////////////////////////////////////////////////////////////////
/// \mainpage
/// Tscope5 
/// \tableofcontents
/// 
/// is a C/C++ experiment programming library for cognitive scientists.
///
///	It is mainly targeted at the Windows platform but also runs on Mac OS X and Linux.
///
/// It provides functions for randomization, graphics, sound, timing, response registration and 
/// synchronisation between computers. 
///
/// Tscope5 is the successor of <a href="http://users.ugent.be/~masteven/tscope/" target="_blank"> Tscope</a>.
/// It is entirely rewritten from scatch and is therefore not source compatible with the old version,
/// even though the same functionality (and more) is provided.
///
/// - See <a href="news.html"> the news page </a> for news about Tscope5 updates
///   or follow us on <a href="http://www.facebook.com/pages/Tscope/347744695089?ref=hl" target="_blank"> facebook</a>.
/// - See <a href="install.html"> the installation manual </a> for information on how to install Tscope5 on your computer.
/// - See <a href="compile.html"> the compilation manual </a> for information on how to compile Tscope5 programs on your computer.
/// 
/// The Tscope5 functions are split into different groups. These groups are listed here:
///
/// \section system System functions
/// Central to Tscope5 there is a small set of functions that allow you to
/// - Log information about the sequence of events in your experiment.
/// - Write out information about the computer you  are running on.
/// - Exit your program gracefully in case of an unexpected event.
///
/// - The functions are documented in system.c
///
/// \section randomizer Randomizer functions
/// Tscope5 provides an interface to the randomization routines from the GNU Scientific Library (GSL).
/// - It can generate random lists with or without replacement.
/// - It can generate numbers from uniform, normal or exponential distributions.
/// - Other distributions are quite easy to implement, please feel free to ask if you need one.
/// - Internally the RANLUX algorithm is used, which has a period of about 10^171. 
///
/// - The functions are documented in randomizer.c
///
/// \section stimuli Stimulus output functions
///
/// \subsection display Display functions
/// This group of functions is used for opening/closing Tscope5 displays
/// and for setting display parameters.
///
/// Tscope5 supports multiple displays, so you can
/// - Have two participants do interaction experiments.
/// - Add an experimenter interface with progress and performance statistics.
///
/// - All Tscope5 drawing operations are performed in sync with the vertical refresh rate of your display.
/// - Fullscreen and windowed modes are supported.
///
/// The graphics functions, primitives drawing functions, bitmap drawing functions,
/// text input/output functions, mouse and keyboard functions require a display to be opened.
///
/// The system functions, randomizer functions, audio functions and timer functions
/// do not require a display to be opened.
///
/// - The functions are documented in display.c
///
/// \subsection graphics Graphics functions
/// This set of functions contains three groups:
///	- Helper functions for computing coordinates on the display.
/// 	- A coordinate system with the origin in the center of the screen is the default.
///		- A relative coordinate system that makes your experiment resolution-independent is available.
///	- Functions for defining colors.
///		- RGB, HSL and HSL color definitions with (partially) transparant colors are supported.
///	- Parameter functions that control the appearance of graphical output and text.
///
/// Graphics parameters influence the primitives drawing functions, bitmap drawing functions
/// and text input/output functions documented below.
///
/// - The functions are documented in graphics.c
///
/// \subsection primitives Primitives drawing functions
/// This group of functions is used for drawing primitive objects on the display.
///
/// - Primitive objects are lines, rectangles, triangles, circles, etc. 
///
/// - The functions are documented in primitives.c
///
/// \subsection bitmaps Bitmap drawing functions
/// This group of functions is used for drawing bitmap objects on the display.
///
/// Bitmaps can either be
/// - Loaded from a file in most popular bitmap formats (bmp, pcx, tga, jpeg, png, mabye some others, ...).
/// - Prepared in advanced using the primitives and text output functions.
///
/// - The functions are documented in bitmaps.c
///
/// \subsection textio Text input/output functions
/// This group of functions is used for drawing text on the display and for reading text from the keyboard.
///
/// - Functions for both C strings and for Unicode strings are available.
///	- TrueType fonts (TTF) are supported. 
///	- A small set of fonts is built in, any font that is available on your system can be imported very easily.
///
/// - The functions are documented in textio.c
///
/// \subsection audio Audio functions
/// This group of functions provides audio playback support.
///
/// - Supported audio file formats are wav, flac and ogg (no mp3 support, that's a commercial format).
/// - Audio recording functions and a software voice key are not implemented yet but are on top of the todo list.
/// - Functions for generating sounds on the fly are also on the todo list.
///
/// - The functions are documented in audio.c
///
/// \section timer Timer functions
/// This group of functions provides timer support.
///
/// The timer functions can
/// - Alter the program's priority to increase timing accuracy (tbd).
/// - Wait for a given time to elapse (stimulus timing).
/// - Wait for a response from the participant (response registration).
/// - Wait for a trigger from another computer (synchronisation between computers).
///
/// The valid reponse buttons or trigger signals have to be defined first
/// using the mouse, keyboard, joystick, cedrusbox, parallel or serial port functions
/// documented below.
///
/// On all response registration devices Tscope5 can detect both key press events and key release events.
///
/// On devices with sub-millisecond timing accuracy (cedrusboxes and parallel port response boxes)
/// Tscope5 gives an estimate of the timing error. On other devices this estimate is not available (set to 0.0).
///
/// Trigger signals from another computer can be simulated internally so you don't need the other
/// computer attached while programming your experiment.
///
/// - The functions are documented in timer.c
///
/// \subsection mouse Mouse functions
/// This group of functions provides mouse (and touch screen) support.
///
/// The mouse can be used
/// - As a response device with low timing resolution.
/// - As a simple gui.
/// - Currenlty only key presses and releases can be defined as responses.
///   Adding mouse movements as responses should be quite easy, feel free to ask if you need it.
/// - Touch screens have the same functionality as a mouse and are programmed in the same way. 
///   If you need touch screen input just use the mouse functions and it will work.
/// - Multiple mice can be attached to a computer, but to Tscope5 it is as if there is only one mouse 
///   (i.e. pressing the left button on the first mouse has the same effect as pressing 
///   the left button on the second mouse. Tscope5 has no idea on which mouse the button was pressed).
///
/// - The functions are documented in mouse.c
///
/// \subsection keyboard Keyboard functions
/// This group of functions provides keyboard support.
///
/// The keyboard can be used
/// - As a response device with low timing resolution.
/// - For scanf type input.
/// - Multiple keyboards can be attached to a computer, but to Tscope5 it is as if there is only one keyboard 
///   (i.e. pressing e.g. the 'A' on the first keyboard has the same effect as pressing 
///   the 'A' on the second keyboard. Tscope5 has no idea on which keyboard the button was pressed).
///
/// - The functions are documented in keyboard.c
///
/// \subsection joystick Joystick functions
/// This group of functions provides joystick support.
///
/// - The joystick can be used as a response device with low timing resolution.
/// - Multiple joysticks can be used simultaneously as independent response devices.
/// - Currenlty only key presses and releases can be defined as responses.
/// - Adding axis movements as responses should be quite easy, feel free to ask if you need it.
///
/// - The functions are documented in joystick.c
///
/// \subsection cedrusbox Cedrusbox functions
/// This group of functions provides cedrusbox support.
///
/// - Cedrusboxes can be used as response devices with sub-millisecond timing resolution.
/// - Multiple cedrusboxes can be used simultaneously as independent response devices.
/// - Cedrusboxes require a driver available from www.cedrus.com (Windows, Mac and Linux).
/// - Tscope5 communicates with Cedrusboxes using the Cedrus XID protocol (flip dip switches 1 and 2 down, 3 up).
///
/// - The functions are documented in cedrusbox.c
///
/// \subsection parport Parallel port functions
/// This group of functions provides support for parallel port communication.
///
/// The parallel port can be used
/// - For attaching home-made response boxes and voice keys with sub-millisecond timing resolution 
///   (pins 10-15, the status register).
/// - For sending and receiving trigger signals between computers with sub-millisecond timing resolution 
///   (pins 2-9, the data register).
/// - Up to three parallel ports can be used simultaneously  as independent response or trigger devices, 
///   but each port can only be used for one job:
///   registering responses, registering triggers or sending triggers.
/// 
/// - The functions are documented in parport.c
///
/// \subsection serialport Serial port functions
/// This group of functions provides support for communication over the serial port.
///
/// Serial ports can be used for sending and receiving trigger signals between computers.
///
/// - Multiple serial ports can be used simultaneously as independent trigger devices.
/// - Each port can be used for both sending and receiving trigger signals 
///   with sub-millisecond timing resolution (you don't need two cables for bidirectional communication).
/// - Modern computers often don't have a serial port but serial communication can be simulated over usb 
///   (like the cedruxboxes or arduino's). These devices will also have sub-millisecond timing resolution.
/// - Serial protocol settings have to be set directly by the user (see examples).
///
/// - The functions are documented in serialport.c
///
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// \page news News about Tscope5 updates
///
/// - 2012-10-24 Tscope5 0.1 released!
///		- download the source <a href="http://users.ugent.be/~masteven/tscope5/download/tscope5-0.1.tar.bz2"> here </a>
///		- download the windows installer <a href="http://users.ugent.be/~masteven/tscope5/download/tscope5-0.1.exe"> here </a>
///
/// - 2012-10-24 Mingw4Tscope5 0.1 released!
///		- download the windows installer <a href="http://users.ugent.be/~masteven/tscope5/download/MinGW4Tscope5-0.1.exe"> here </a>
///
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// \page install Tscope5 installation manual
///
///	\tableofcontents
///
///	\section wininstall Windows installation
/// We bundled all the tools (compiler, libraries, fonts, custom scripts) that Tscope5 relies on into a single installer package
/// called Mingw4Tscope5.
///
/// Tscope5 itself is not part of MinGW4Tscope5, as it is updated more frequently.
///	After the installation of MinGW4Tscope5, you will still need to install Tscope5 as described in the next section.
/// 
/// \subsection minginstall Mingw4Tscope5 installation
/// - Download the installer <a href="http://users.ugent.be/~masteven/tscope5/download/MinGW4Tscope5-0.1.exe"> here</a>. It is about 150 megabytes.
/// - Execute the installer. You will need administrative rights to run the installer.
///   A power user account is not enough. If you don't have administrative rights, the installer will tell you and abort.
/// - There are not many installation options. 
/// Mingw4Tscope5 will not install in the standard Program Files directory because of the space in that directory's name. 
/// It will be installed in C:\\Mingw4Tscope5 instead. 
/// (Actually, Mingw4Tscope5 will be installed on the same drive as your Windows system - so if Windows is installed on the D: drive,
///  Mingw4Tscope5 will be installed in D:\\Mingw4Tscope5, etc. 
/// All further information on this site assumes Mingw4Tscope5 is installed in C:\\Mingw4Tscope5, 
/// but please note that on your system this also might be D:\\Mingw4Tscope5 or any other drive).
/// - If everything is ok, the installer will extract all the necessary files. 
/// This can take a while - the fully extracted system takes about 500 megabytes of disk space.
///
/// After installation an icon called 'MinGW4Tscope5 will be added to your desktop.
/// Double clicking this icon will open a bash shell where you can enter your compilation commands.
/// 
/// \subsection tscopewinstall Tscope5 installation
/// - For Windows we provide an automatic installer that can be downloaded 
/// <a href="http://users.ugent.be/~masteven/tscope5/download/tscope5-0.1.exe"> here</a>.
/// - Execute the installer in the same way as the MinGW4Tscope5 installer. Again there are no installation options.
///
/// After installation you are ready to start developing Tscope5 programs. 
///
/// Please not that the distribution currently does not contain a source code editor.
/// You will have to download and install one yourself. 
/// For Windows we recommend <a href="http://notepad-plus-plus.org" target="_blank"> notepad++</a>. 
/// In the future we might provide a plug-in that allows you to compile and run programs from notepad++,
/// but for now you will need to use the shell commands described in <a href="compile.html"> the compilation manual </a> .
///
///	\section macinstall Mac OS X installation
///	To be done
///
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// \page compile Tscope5 compilation manual
///
/// Compile foo.c with
///
/// tcc foo.c
///
/// This will produce foo.exe, if there are no errors in your code.
///
/// You can run foo.exe with
///
/// ./foo.exe
///
/// If necessary, kill foo.exe by pressing CTRL-C in the shell.
///
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// \file system.c
/// Definitions of system functions
/// \example system1.c Log level and status
/// \example system2.c write log and error exit
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/system.h"
#include "../include/tscope5/display.h"
#include "../include/tscope5/system_internal.h"


//////////////////////////////////////////////////////////////////////////////
/// @name Logging functions
/// Tscope5 automatically sends information about its status to a logging window and a file called tscope5.log.
/// 
/// With the functions below you control the amount of information that is logged.
/// It is also possible to send your own log messages.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Send info to a logging window.
///
/// \param level The importance of this message (TS5_LOGLEVEL_0 - TS5_LOGLEVEL_6). Smaller values are more important.
/// \param format Printf-style format string that contains the log message.
///
/// The log message is printed in a separate text log window.
///
/// Messages with a level above the level specified with ts5_set_log_level are omitted.
//////////////////////////////////////////////////////////////////////////////
void ts5_log(const unsigned int level, const char *format, ...)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_log");
	}

	if (level <= _ts5_status.loglevel) {

		if (!_ts5_is_textlog_installed) {
			_ts5_status.textlog = fopen("tscope5.log", "w+");
			if (!_ts5_status.textlog) {
				ts5_fatal("ts5_log: could not open log file\n");
			}
			_ts5_is_textlog_installed = 1;
		}

		char outbuf[TS5_MAX_CHAR];
		va_list args;
		va_start(args, format);
		vsnprintf(outbuf, TS5_MAX_CHAR*sizeof(char), format, args);
		va_end(args);

		double logtime = al_get_time();
		fprintf(stderr, "%8.3f: %s", logtime, outbuf);
		fprintf(_ts5_status.textlog, "%8.3f: %s", logtime, outbuf);
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Set log level
///
/// \param loglevel How much stuff is logged.
///
/// \return The previous log level.
///
/// This parameter function controls what information is printed to the log window.
///
/// There are six possible settings:
/// - TS5_LOGLEVEL_0: no debugging info.
/// - TS5_LOGLEVEL_1: report when Tscope5 subsystems are loaded.
/// - TS5_LOGLEVEL_2: report when memory is allocated/freed.
/// - TS5_LOGLEVEL_3: report when response keys are defined.
/// - TS5_LOGLEVEL_4: report when parameter are set or queried.
/// - TS5_LOGLEVEL_5: report each function call.
/// - TS5_LOGLEVEL_6: report internal check and coordinate functions.
///
/// If other values are passed to this function the program is aborted.
///
/// The default loglevel is TS5_LOGLEVEL_3.
//////////////////////////////////////////////////////////////////////////////
int ts5_set_log_level(const unsigned int loglevel)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_set_log_level");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_log_level(%d)\n", loglevel);

	int retval;
	retval = _ts5_status.loglevel;

	if (loglevel <= TS5_LOGLEVEL_6) {
		_ts5_status.loglevel = loglevel;
		ts5_log(TS5_LOGLEVEL_4, "ts5_set_log_level: changed loglevel from %d to %d\n",
									retval, loglevel);
	}
	else {
		ts5_fatal("ts5_log: requested loglevel not supported.");
	}

	return retval;
}


//////////////////////////////////////////////////////////////////////////////
/// Get log level
///
/// \return The log level.
//////////////////////////////////////////////////////////////////////////////
int ts5_get_log_level()
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_get_log_level");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_get_log_level\n");

	return _ts5_status.loglevel;
}


//////////////////////////////////////////////////////////////////////////////
/// Write the status of all Tscope5 subsystem to the log window.
///
/// Information is printed about:
/// - The display adapter(s) of the computer and the available fullscreen modes.
/// - The displays opened by Tscope5.
/// - The graphics parameters.
//////////////////////////////////////////////////////////////////////////////
void ts5_print_status()
{
	// display adapters + supported fullscreen modes
	int i;
	ts5_log(TS5_LOGLEVEL_0, "\n");
	ts5_log(TS5_LOGLEVEL_0, "========================================\n");
	ts5_log(TS5_LOGLEVEL_0, "tscope5 status:\n");
	ts5_log(TS5_LOGLEVEL_0, "\n");

	ts5_log(TS5_LOGLEVEL_0, "Loglevel: %d\n", _ts5_status.loglevel);
	ts5_log(TS5_LOGLEVEL_0, "Number of display adapters: %d\n", _ts5_status.num_display_adapters);

	for (i = 0; i < _ts5_status.num_display_adapters; i++) {

		ts5_log(TS5_LOGLEVEL_0, "Display adapter %d: w: %d (%d:%d), h: %d (%d:%d)\n", i+1,
				_ts5_status.display_adapter_status[i].w,
				_ts5_status.display_adapter_status[i].x1,
				_ts5_status.display_adapter_status[i].x2,
				_ts5_status.display_adapter_status[i].h,
				_ts5_status.display_adapter_status[i].y1,
				_ts5_status.display_adapter_status[i].y2);
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");

	int oldadapter = ts5_nextdisplay.adapter;
	ALLEGRO_DISPLAY_MODE *mode;
	mode = (ALLEGRO_DISPLAY_MODE *)al_malloc(sizeof(ALLEGRO_DISPLAY_MODE));

	for (i = 0; i < _ts5_status.num_display_adapters; i++) {

		al_set_new_display_adapter(i);
		int oldrate = al_get_new_display_refresh_rate();
		int oldflags = al_get_new_display_flags();
		al_set_new_display_refresh_rate(0);
		al_set_new_display_flags(0);

		int nmodes = al_get_num_display_modes();

		int j;
		for (j=0; j<nmodes; j++) {

			mode = al_get_display_mode(j, mode);
			ts5_log(TS5_LOGLEVEL_0, "adapter %d mode %d: w=%d, h=%d, format=%d, rate=%d\n",
						i+1, j+1, mode->width, mode->height, mode->format, mode->refresh_rate);
		}

		al_set_new_display_refresh_rate(oldrate);
		al_set_new_display_flags(oldflags);
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");
	al_set_new_display_adapter(oldadapter);
	al_free(mode);

	// display settings + settings for next display
	ts5_log(TS5_LOGLEVEL_0, "========================================\n");
	ts5_log(TS5_LOGLEVEL_0, "Parameters for next display:\n");
	ts5_log(TS5_LOGLEVEL_0, "\tadapter %d\n", ts5_nextdisplay.adapter+1);
	ts5_log(TS5_LOGLEVEL_0, "\tw=%d, h=%d\n", ts5_nextdisplay.w, ts5_nextdisplay.h);
	ts5_log(TS5_LOGLEVEL_0, "\tx=%d, y=%d\n", ts5_nextdisplay.x, ts5_nextdisplay.y);
	ts5_log(TS5_LOGLEVEL_0, "\trefreshrate=%d\n", ts5_nextdisplay.refreshrate);
	ts5_log(TS5_LOGLEVEL_0, "\tdisplay mode=%d\n", ts5_nextdisplay.display_mode);
	ts5_log(TS5_LOGLEVEL_0, "\tvsync mode=%d\n", ts5_nextdisplay.vsync_mode);
	ts5_log(TS5_LOGLEVEL_0, "\n");

	ts5_log(TS5_LOGLEVEL_0, "Number of displays: %d\n", _ts5_status.num_displays);
	ts5_log(TS5_LOGLEVEL_0, "Active display: %d\n", _ts5_status.active_display);

	for (i = 0; i < _ts5_status.num_displays; i++) {

		ts5_log(TS5_LOGLEVEL_0,
				"Display %d: adapter %d, w=%d, h=%d, x=%d, y=%d, rate=%d, display_mode=%d, vsync=%d\n",
				i+1, _ts5_status.display_status[i].adapter+1,
				_ts5_status.display_status[i].w, _ts5_status.display_status[i].h,
				_ts5_status.display_status[i].x, _ts5_status.display_status[i].y,
				_ts5_status.display_status[i].refreshrate,
				_ts5_status.display_status[i].display_mode,
				_ts5_status.display_status[i].vsync_mode);
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");

	// graphics parameters
	float r, g, b, a;
	ts5_log(TS5_LOGLEVEL_0, "========================================\n");
	ts5_log(TS5_LOGLEVEL_0, "Coordinate system: %d\n", _ts5_status.graphics_status.coordinate_system);
	ts5_log(TS5_LOGLEVEL_0, "Coordinate scale: %d\n", _ts5_status.graphics_status.coordinate_scale);
	ts5_log(TS5_LOGLEVEL_0, "Target bitmap size: %d x %d\n",
				_ts5_status.graphics_status.target_width, _ts5_status.graphics_status.target_height);

	al_unmap_rgba_f(_ts5_status.graphics_status.foreground_color, &r, &g, &b, &a);
	ts5_log(TS5_LOGLEVEL_0, "Foreground color: r: %f  g: %f b: %f alpha: %f\n", r, g, b, a);
	al_unmap_rgba_f(_ts5_status.graphics_status.background_color, &r, &g, &b, &a);
	ts5_log(TS5_LOGLEVEL_0, "Background color: r: %f  g: %f b: %f alpha: %f\n", r, g, b, a);
	ts5_log(TS5_LOGLEVEL_0, "Line drawing thickness: %f\n", _ts5_status.graphics_status.drawing_thickness);
	ts5_log(TS5_LOGLEVEL_0, "Fill mode: %d\n", _ts5_status.graphics_status.fill_mode);
	ts5_log(TS5_LOGLEVEL_0, "Font type: %d\n", _ts5_status.graphics_status.font_type);
	ts5_log(TS5_LOGLEVEL_0, "Font style: %d\n", _ts5_status.graphics_status.font_style);
	ts5_log(TS5_LOGLEVEL_0, "Font size: %d\n", _ts5_status.graphics_status.font_size);
	ts5_log(TS5_LOGLEVEL_0, "Font number: %d\n", _ts5_status.graphics_status.font_index);
	ts5_log(TS5_LOGLEVEL_0, "Text alignment: %d\n", _ts5_status.graphics_status.text_alignment);
	ts5_log(TS5_LOGLEVEL_0, "\n");

	// response keys
	ts5_log(TS5_LOGLEVEL_0, "========================================\n");
	ts5_log(TS5_LOGLEVEL_0, "Number of defined response keys: %d\n",
								_ts5_status.timer_status.num_defined_buttons);

	ts5_log(TS5_LOGLEVEL_0, "Number of active response keys: %d\n",
								_ts5_status.timer_status.num_active_buttons);

	ts5_log(TS5_LOGLEVEL_0, "Mouse is response device: %d\n",
								_ts5_status.timer_status.mouse_is_response_device);

	if (_ts5_status.timer_status.mouse_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.mouse.num_buttons; i++) {

			if (_ts5_status.timer_status.mouse.button_press_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "Mouse button press %d has response value %d\n",
											i+1, _ts5_status.timer_status.mouse.button_press_defined[i]);
			}

			if (_ts5_status.timer_status.mouse.button_release_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "Mouse button release %d has response value %d\n",
											i+1, _ts5_status.timer_status.mouse.button_release_defined[i]);
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Keyboard is response device: %d\n",
								_ts5_status.timer_status.keyboard_is_response_device);

	if (_ts5_status.timer_status.keyboard_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.keyboard.num_buttons; i++) {

			if (_ts5_status.timer_status.keyboard.button_press_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "Keyboard button press %d has response value %d\n",
							i+1, _ts5_status.timer_status.keyboard.button_press_defined[i]);
			}

			if (_ts5_status.timer_status.keyboard.button_release_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "Keyboard button release %d has response value %d\n",
							i+1, _ts5_status.timer_status.keyboard.button_release_defined[i]);
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Number of joysticks: %d\n", _ts5_status.timer_status.num_joystick);
	ts5_log(TS5_LOGLEVEL_0, "Joystick is response device: %d\n",
								_ts5_status.timer_status.joystick_is_response_device);

	if (_ts5_status.timer_status.joystick_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.num_joystick; i++) {
			int j;

			for (j=0; j<_ts5_status.timer_status.joystick[i].num_buttons; j++) {

				if (_ts5_status.timer_status.joystick[i].button_press_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "Joystick %d button press %d has response value %d\n",
								i+1, j+1, _ts5_status.timer_status.joystick[i].button_press_defined[j]);
				}

				if (_ts5_status.timer_status.joystick[i].button_release_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "Joystick %d button release %d has response value %d\n",
								i+1, j+1, _ts5_status.timer_status.joystick[i].button_release_defined[j]);
				}
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Number of cedrusboxes: %d\n", _ts5_status.timer_status.num_cedrusbox);
	ts5_log(TS5_LOGLEVEL_0, "Cedrusbox is response device: %d\n",
								_ts5_status.timer_status.cedrusbox_is_response_device);

	if (_ts5_status.timer_status.cedrusbox_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

			ts5_log(TS5_LOGLEVEL_0, "Cedrusbox %d is mapped to serial device %d\n",
										i+1, _ts5_status.timer_status.cedrusbox[i].port_num+1);

			ts5_log(TS5_LOGLEVEL_0, "Cedrusbox %d is type %d\n",
										i+1, _ts5_status.timer_status.cedrusbox[i].type);

			int j;
			for (j=0; j<_ts5_status.timer_status.cedrusbox[i].num_buttons; j++) {

				if (_ts5_status.timer_status.cedrusbox[i].button_press_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "Cedrusbox %d button press %d has response value %d\n",
								i+1, j+1, _ts5_status.timer_status.cedrusbox[i].button_press_defined[j]);
				}

				if (_ts5_status.timer_status.cedrusbox[i].button_release_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "Cedrusbox %d button release %d has response value %d\n",
								i+1, j+1, _ts5_status.timer_status.cedrusbox[i].button_release_defined[j]);
				}
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Number of parallel ports: %d\n", _ts5_status.timer_status.num_parport);
	ts5_log(TS5_LOGLEVEL_0, "Parallel port is response device: %d\n",
				_ts5_status.timer_status.parport_is_response_device);

	if (_ts5_status.timer_status.parport_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.num_parport; i++) {

			int j;
			for (j=0; j<_ts5_status.timer_status.parport[i].num_buttons; j++) {

				if (_ts5_status.timer_status.parport[i].button_press_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "Parport %d button press %d has response value %d\n",
								i+1, j+1, _ts5_status.timer_status.parport[i].button_press_defined[j]);
				}

				if (_ts5_status.timer_status.cedrusbox[i].button_release_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "Parport %d button release %d has response value %d\n",
								i+1, j+1, _ts5_status.timer_status.parport[i].button_release_defined[j]);
				}
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");


	// audio parameters
	ts5_log(TS5_LOGLEVEL_0, "========================================\n");
	ts5_log(TS5_LOGLEVEL_0, "Number of audio channels: %d\n", _ts5_status.audio_status.channels);
	ts5_log(TS5_LOGLEVEL_0, "Sample rate: %d\n", _ts5_status.audio_status.frequency);
	ts5_log(TS5_LOGLEVEL_0, "Sample format: %d\n", _ts5_status.audio_status.depth);
	ts5_log(TS5_LOGLEVEL_0, "Gain: %f\n", _ts5_status.audio_status.gain);


	ts5_log(TS5_LOGLEVEL_0, "\n");
	ts5_log(TS5_LOGLEVEL_0, "========================================\n");
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Error exit
/// During an experiment lots of unexpected things can happen.
/// (e.g. you forgot to copy stimulus bitmaps to the experiment computer).
///
/// It is a good habit to check for such unexpected events and exit your
/// program safely before an ugly crash happens. 
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Exit safely with an error message.
///
/// \param format Printf-style format string that contains the error message.
///
/// Closes Tscope5, all subsystems and the log window.
///
/// A message box appears with the error message supplied by the user.
///
/// This function can be called before Tscope5 is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_fatal(const char *format, ...)
{
	ts5_log(TS5_LOGLEVEL_1, "ts5_fatal\n");

	char outbuf[TS5_MAX_CHAR];
	va_list args;
	va_start(args, format);
	vsnprintf(outbuf, TS5_MAX_CHAR*sizeof(char), format, args);
	va_end(args);

	fprintf(stderr, "Tscope5 error: %s", outbuf);

	if (_ts5_status.num_displays>0) {

		if (_ts5_status.display_status[_ts5_status.active_display].display_mode == TS5_FULLSCREEN_WINDOW) {
			al_toggle_display_flag(_ts5_status.display[_ts5_status.active_display],
									ALLEGRO_FULLSCREEN_WINDOW, 0);
			_ts5_status.display_status[_ts5_status.active_display].display_mode = TS5_WINDOWED;
		}

		al_show_native_message_box(_ts5_status.display[_ts5_status.active_display],
									"Tscope5 error", "Tscope5 error", outbuf, NULL, ALLEGRO_MESSAGEBOX_ERROR);
	}
	else {
		al_show_native_message_box(NULL, "Tscope5 error", "Tscope5 error",
									outbuf, NULL, ALLEGRO_MESSAGEBOX_ERROR);
	}
	exit(1);
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////
