//////////////////////////////////////////////////////////////////////////////
/// \file system_internal.c
/// Definitions of internal system functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/system.h"
#include "../include/tscope5/system_internal.h"

/// Is Tscope5 installed?
int  _ts5_is_tscope5_installed = 0;

/// Is text log installed?
int  _ts5_is_textlog_installed = 0;

/// Settings for the next display that will be opened.
TS5_DISPLAY_STATUS ts5_nextdisplay = { 0, 640, 360, 0, 0, 60, TS5_WINDOWED, TS5_VSYNC_ON };


//////////////////////////////////////////////////////////////////////////////
/// Install Tscope5.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Gathers information about the display adapter(s) of the
/// computer and sets the default parameters for displays, graphics and audio.
///
/// This function is called automatically if necessary.
///
/// If Tscope5 can't be installed the program is aborted.
///
/// This function does not open a display, that is done by ts5_install_display().
//////////////////////////////////////////////////////////////////////////////
void ts5_install_tscope5(char *calling_function)
{
	int i;

	if (!_ts5_is_tscope5_installed) {

		_ts5_is_tscope5_installed = 1;
		_ts5_status.loglevel = TS5_LOGLEVEL_3;
		_ts5_status.textlog = NULL;

		if (!al_install_system(ALLEGRO_VERSION_INT, NULL)) {
			ts5_fatal("%s: could not install Tscope5\n", calling_function);
		}

		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5\n", calling_function);
		al_inhibit_screensaver(1);
		al_set_separate_blender(ALLEGRO_ADD, ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA,
									ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ONE);
//		al_set_new_display_option(ALLEGRO_SAMPLE_BUFFERS, 1, ALLEGRO_REQUIRE);
//		al_set_new_display_option(ALLEGRO_SAMPLES, 4, ALLEGRO_SUGGEST);

		// display adapters
		_ts5_status.num_display_adapters = al_get_num_video_adapters();

		_ts5_status.display_adapter_status =
			al_malloc(_ts5_status.num_display_adapters * sizeof(TS5_DISPLAY_ADAPTER_STATUS));

		for (i = 0; i < _ts5_status.num_display_adapters; i++) {

			ALLEGRO_MONITOR_INFO monitorinfo;
			al_get_monitor_info(i, &monitorinfo);
			_ts5_status.display_adapter_status[i].x1 = monitorinfo.x1;
			_ts5_status.display_adapter_status[i].x2 = monitorinfo.x2;
			_ts5_status.display_adapter_status[i].y1 = monitorinfo.y1;
			_ts5_status.display_adapter_status[i].y2 = monitorinfo.y2;
			_ts5_status.display_adapter_status[i].w = monitorinfo.x2 - monitorinfo.x1;
			_ts5_status.display_adapter_status[i].h = monitorinfo.y2 - monitorinfo.y1;
		}

		ts5_nextdisplay.x = 0;
		ts5_nextdisplay.y = 0;

		// displays
		_ts5_status.num_displays = 0;
		_ts5_status.active_display = -1;
		_ts5_status.display_status = NULL;
		_ts5_status.display = NULL;
		_ts5_status.target = NULL;

		// graphics parameters
		_ts5_status.graphics_status.coordinate_system = TS5_CARTESIAN_COORDINATES;
		_ts5_status.graphics_status.coordinate_scale = TS5_ABSOLUTE_COORDINATES;

		_ts5_status.graphics_status.foreground_color = al_map_rgba_f(1.0, 1.0, 1.0, 1.0);
		_ts5_status.graphics_status.background_color = al_map_rgba_f(0.0, 0.0, 0.0, 1.0);
		_ts5_status.graphics_status.drawing_thickness = 0.0;
		_ts5_status.graphics_status.fill_mode = TS5_FILL_OFF;

		_ts5_status.graphics_status.font_type = TS5_COURIER;
		_ts5_status.graphics_status.font_style = TS5_BOLD;
		_ts5_status.graphics_status.font_size = 20;
		_ts5_status.graphics_status.font_index =
			_ts5_status.graphics_status.font_type * TS5_FONTSTYLES * TS5_FONTSIZES +
			_ts5_status.graphics_status.font_style * TS5_FONTSIZES +
			_ts5_status.graphics_status.font_size - 8;

		_ts5_status.graphics_status.text_alignment = TS5_ALIGN_CENTER;


		for (i=0; i<TS5_NFONTS; i++) {
			_ts5_status.graphics_status.fonts[i] = NULL;
		}
		_ts5_status.graphics_status.font = NULL;

		// randomizer
		_ts5_status.randomizer = NULL;

		// response keys
		_ts5_status.timer_status.num_defined_buttons = 0;
		_ts5_status.timer_status.num_active_buttons = 0;

		_ts5_status.timer_status.mouse_is_response_device = 0;

		_ts5_status.timer_status.keyboard_is_response_device = 0;

		_ts5_status.timer_status.num_joystick = 0;
		_ts5_status.timer_status.joystick = NULL;
		_ts5_status.timer_status.joystick_is_response_device = 0;

		_ts5_status.timer_status.num_cedrusbox = 0;
		_ts5_status.timer_status.cedrusbox = NULL;
		_ts5_status.timer_status.cedrusbox_thread = NULL;
		_ts5_status.timer_status.cedrusbox_is_response_device = 0;

		_ts5_status.timer_status.num_parport = 0;
		_ts5_status.timer_status.parport = NULL;
		_ts5_status.timer_status.parport_thread = NULL;
		_ts5_status.timer_status.parport_is_response_device = 0;

		_ts5_status.timer_status.num_serialport = 0;
		_ts5_status.timer_status.serialport = NULL;
		_ts5_status.timer_status.serialport_thread = NULL;
		_ts5_status.timer_status.serialport_is_trigger_device = 0;

		_ts5_status.timer_status.response_queue = NULL;
		_ts5_status.timer_status.trigger_queue = NULL;
		_ts5_status.timer_status.trigger_log = NULL;

		// audio parameters
		_ts5_status.audio_status.frequency = TS5_44100;
		_ts5_status.audio_status.channels = TS5_STEREO;
		_ts5_status.audio_status.depth = TS5_FLOAT;
		_ts5_status.audio_status.gain = 1.0;

		_ts5_status.audio_status.voice = NULL;
		_ts5_status.audio_status.mixer = NULL;

		atexit(ts5_uninstall_tscope5);
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Uninstall Tscope5.
///
/// Closes Tscope5, all subsystems and the log window.
///
/// This function is called automatically at the end of the program.
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_tscope5()
{
	if (_ts5_is_tscope5_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5\n");
		al_inhibit_screensaver(0);

		if (_ts5_is_textlog_installed) {
			fflush(_ts5_status.textlog);
			fclose(_ts5_status.textlog);
			_ts5_is_textlog_installed=0;
		}

		al_free(_ts5_status.display_adapter_status);

		#ifdef ALLEGRO_MACOSX
			al_uninstall_system();

		#endif

		_ts5_is_tscope5_installed = 0;
	}
}

