//////////////////////////////////////////////////////////////////////////////
/// \file timer.c
/// Definitions of timer functions
/// \example timer1.c wait functions
/// \example timer2.c stimulus timing
/// \example timer3.c timing with maximum response time
/// \example timer4.c double keypress response
/// \example timer5.c changing the stimulus while waiting for a response
/// \example timer6.c measuring key press time
//////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/timer.h"
#include "../include/tscope5/timer_internal.h"
#include "../include/tscope5/system_internal.h"


//////////////////////////////////////////////////////////////////////////////
/// @name General timing functions
/// The general timing functions are mainly used to control stimulus timing.
/// 
/// You can either wait for a given amount of time, 
/// or until some time in the future. 
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Get the number of seconds since the program started.
///
/// \return The number of seconds since the program started (as a double precision float).
//////////////////////////////////////////////////////////////////////////////
double ts5_get_time()
{
	ts5_check_timer("ts5_get_time");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_time()\n");

	return al_get_time();
}


//////////////////////////////////////////////////////////////////////////////
/// Wait for a number of seconds.
//////////////////////////////////////////////////////////////////////////////
void ts5_wait(double waittime)
{
	ts5_check_timer("ts5_wait");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait(%f)\n", waittime);

	// only use up cpu during the last 20 ms.
	double starttime = al_get_time();

	if (waittime > 0.02) {
		al_rest(waittime - 0.02);
	}

	while (al_get_time() < starttime+waittime) {

	}
}


//////////////////////////////////////////////////////////////////////////////
/// Wait until a deadline.
//////////////////////////////////////////////////////////////////////////////
void ts5_wait_until(double deadline)
{
	ts5_check_timer("ts5_wait_until");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_until(%f)\n", deadline);

	// only use up cpu during the last 20 ms.
	double starttime = al_get_time();

	if (deadline - starttime > 0.02) {
		al_rest(deadline - starttime - 0.02);
	}

	while (al_get_time() < deadline) {

	}
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// The response registration functions require you to define response keys first
/// using ts5_define_mouse_key(), ts5_define_keyboard_key(), ts5_define_joystick_key(),
/// ts5_define_cedrusbox_key() or ts5_define_parport_key(). 
/// 
/// Each response key will get a response value. 
/// The first key that is defined will get number 1, etc.
///
/// Registration of responses is done asynchronously. 
/// All key presses that are defined will be buffered continuously during the program.
///
/// Waiting for a response to a stimulus will typically involve flushing all buffered responses
/// that occurred befor stimulus onset using ts5_flush_responses() and then checking for
/// new responses using ts5_check_response(), ts5_wait_for_response(),
/// ts5_wait_for_response_timed() or ts5_wait_for_response_until().
///
/// The response definition can be altered during the program using
/// ts5_hide_response_key() or ts5_remove_response_keys().
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Temporarily deactivates a response key.
///
/// \param key The key that has to be deactivated.
///
/// \return The number of active keys left.
///
/// This function also checks whether there are any active response left. Aborts if necessary.
///
/// Reactivating all defined keys is possible by calling ts5_hide_response_key(0).
//////////////////////////////////////////////////////////////////////////////
int ts5_hide_response_key(int key)
{
	ts5_check_timer("ts5_hide_response_key");
	ts5_check_timer2("ts5_hide_response_key");
	ts5_log(TS5_LOGLEVEL_3, "ts5_hide_response_key(%d)\n", key);

	if (key<0 || key>_ts5_status.timer_status.num_defined_buttons) {
		ts5_fatal("ts5_hide_response_key: key argument is %d, there are %d reponse buttons defined\n",
					key, _ts5_status.timer_status.num_defined_buttons);
	}

	// reactivate
	int found = 1;

	if (key==0) {

		int i, j;

		if (_ts5_status.timer_status.mouse_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.mouse.num_buttons; i++) {

				_ts5_status.timer_status.mouse.button_press_active[i] =
					_ts5_status.timer_status.mouse.button_press_defined[i];

				_ts5_status.timer_status.mouse.button_release_active[i] =
					_ts5_status.timer_status.mouse.button_release_defined[i];
			}

			_ts5_status.timer_status.mouse.num_active_buttons =
				_ts5_status.timer_status.mouse.num_defined_buttons;
		}

		if (_ts5_status.timer_status.keyboard_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.keyboard.num_buttons; i++) {

				_ts5_status.timer_status.keyboard.button_press_active[i] =
					_ts5_status.timer_status.keyboard.button_press_defined[i];

				_ts5_status.timer_status.keyboard.button_release_active[i] =
					_ts5_status.timer_status.keyboard.button_release_defined[i];
			}

			_ts5_status.timer_status.keyboard.num_active_buttons =
				_ts5_status.timer_status.keyboard.num_defined_buttons;
		}

		if (_ts5_status.timer_status.joystick_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.num_joystick; i++) {

				for (j=0; j<_ts5_status.timer_status.joystick[i].num_buttons; j++) {

					_ts5_status.timer_status.joystick[i].button_press_active[j] =
						_ts5_status.timer_status.joystick[i].button_press_defined[j];

					_ts5_status.timer_status.joystick[i].button_release_active[j] =
						_ts5_status.timer_status.joystick[i].button_release_defined[j];
				}
				_ts5_status.timer_status.joystick[i].num_active_buttons =
					_ts5_status.timer_status.joystick[i].num_defined_buttons;
			}
		}

		if (_ts5_status.timer_status.cedrusbox_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

				for (j=0; j<_ts5_status.timer_status.cedrusbox[i].num_buttons; j++) {

					_ts5_status.timer_status.cedrusbox[i].button_press_active[j] =
						_ts5_status.timer_status.cedrusbox[i].button_press_defined[j];

					_ts5_status.timer_status.cedrusbox[i].button_release_active[j] =
						_ts5_status.timer_status.cedrusbox[i].button_release_defined[j];
				}
				_ts5_status.timer_status.cedrusbox[i].num_active_buttons =
					_ts5_status.timer_status.cedrusbox[i].num_defined_buttons;
			}
		}
		_ts5_status.timer_status.num_active_buttons = _ts5_status.timer_status.num_defined_buttons;
	}
	else {
		int i, j;
		found=0;

		if (_ts5_status.timer_status.mouse_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.mouse.num_buttons; i++) {

				if (key==_ts5_status.timer_status.mouse.button_press_active[i]) {
					_ts5_status.timer_status.mouse.button_press_active[i] = 0;
					found = 1;
				}
				else if (key==_ts5_status.timer_status.mouse.button_release_active[i]) {
					_ts5_status.timer_status.mouse.button_release_active[i] = 0;
					found = 1;
				}

				if (found) {
					_ts5_status.timer_status.mouse.num_active_buttons--;
					_ts5_status.timer_status.num_active_buttons--;
					goto end;
				}
			}
		}

		if (_ts5_status.timer_status.keyboard_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.keyboard.num_buttons; i++) {

				if (key==_ts5_status.timer_status.keyboard.button_press_active[i]) {
					_ts5_status.timer_status.keyboard.button_press_active[i] = 0;
					found = 1;
				}
				else if (key==_ts5_status.timer_status.keyboard.button_release_active[i]) {
					_ts5_status.timer_status.keyboard.button_release_active[i] = 0;
					found = 1;
				}

				if (found) {
					_ts5_status.timer_status.keyboard.num_active_buttons--;
					_ts5_status.timer_status.num_active_buttons--;
					goto end;
				}
			}
		}

		if (_ts5_status.timer_status.joystick_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.num_joystick; i++) {

				for (j=0; j<_ts5_status.timer_status.joystick[i].num_buttons; j++) {

					if (key==_ts5_status.timer_status.joystick[i].button_press_active[j]) {
						_ts5_status.timer_status.joystick[i].button_press_active[j] = 0;
						found = 1;
					}
					else if (key==_ts5_status.timer_status.joystick[i].button_release_active[j]) {
						_ts5_status.timer_status.joystick[i].button_release_active[j] = 0;
						found = 1;
					}

					if (found) {
						_ts5_status.timer_status.joystick[i].num_active_buttons--;
						_ts5_status.timer_status.num_active_buttons--;
						goto end;
					}
				}
			}
		}

		if (_ts5_status.timer_status.cedrusbox_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

				for (j=0; j<_ts5_status.timer_status.cedrusbox[i].num_buttons; j++) {

					if (key==_ts5_status.timer_status.cedrusbox[i].button_press_active[j]) {
						_ts5_status.timer_status.cedrusbox[i].button_press_active[j] = 0;
						found = 1;
					}
					else if (key==_ts5_status.timer_status.cedrusbox[i].button_release_active[j]) {
						_ts5_status.timer_status.cedrusbox[i].button_release_active[j] = 0;
						found = 1;
					}

					if (found) {
						_ts5_status.timer_status.cedrusbox[i].num_active_buttons--;
						_ts5_status.timer_status.num_active_buttons--;
						goto end;
					}
				}
			}
		}
	}

	end:

	if (!found) {
		ts5_fatal("ts5_hide_response_key: response key %d not found\n", key);
	}

	if (_ts5_status.timer_status.num_active_buttons<1) {
		ts5_fatal("ts5_hide_response_key: there are no active response buttons left\n");
	}

	return _ts5_status.timer_status.num_active_buttons;
}


//////////////////////////////////////////////////////////////////////////////
/// Completely removes the response key definition.
//////////////////////////////////////////////////////////////////////////////
void ts5_remove_response_keys()
{
	ts5_check_timer("ts5_remove_response_keys");
	ts5_log(TS5_LOGLEVEL_3, "ts5_remove_response_keys()\n");

	int i, j;

	// mouse
	if (_ts5_status.timer_status.mouse_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.mouse.num_buttons; i++) {

			_ts5_status.timer_status.mouse.button_press_defined[i] = 0;
			_ts5_status.timer_status.mouse.button_press_active[i] = 0;
			_ts5_status.timer_status.mouse.button_release_defined[i] = 0;
			_ts5_status.timer_status.mouse.button_release_active[i] = 0;
		}

		_ts5_status.timer_status.mouse_is_response_device = 0;
	}

	// keyboard
	if (_ts5_status.timer_status.keyboard_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.keyboard.num_buttons; i++) {

			_ts5_status.timer_status.keyboard.button_press_defined[i] = 0;
			_ts5_status.timer_status.keyboard.button_press_active[i] = 0;
			_ts5_status.timer_status.keyboard.button_release_defined[i] = 0;
			_ts5_status.timer_status.keyboard.button_release_active[i] = 0;
		}

		_ts5_status.timer_status.keyboard_is_response_device = 0;
	}

	// joystick
	if (_ts5_status.timer_status.joystick_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.num_joystick; i++) {

			for (j=0; j<_ts5_status.timer_status.joystick[i].num_buttons; j++) {

				_ts5_status.timer_status.joystick[i].button_press_defined[j] = 0;
				_ts5_status.timer_status.joystick[i].button_press_active[j] = 0;
				_ts5_status.timer_status.joystick[i].button_release_defined[j] = 0;
				_ts5_status.timer_status.joystick[i].button_release_active[j] = 0;
			}
		}

		_ts5_status.timer_status.joystick_is_response_device = 0;
	}

	// joystick
	if (_ts5_status.timer_status.joystick_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.num_joystick; i++) {

			for (j=0; j<_ts5_status.timer_status.joystick[i].num_buttons; j++) {

				_ts5_status.timer_status.joystick[i].button_press_defined[j] = 0;
				_ts5_status.timer_status.joystick[i].button_press_active[j] = 0;
				_ts5_status.timer_status.joystick[i].button_release_defined[j] = 0;
				_ts5_status.timer_status.joystick[i].button_release_active[j] = 0;
			}
		}

		_ts5_status.timer_status.joystick_is_response_device = 0;
	}

	// cedrusbox
	if (_ts5_status.timer_status.cedrusbox_is_response_device) {

		for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

			for (j=0; j<_ts5_status.timer_status.cedrusbox[i].num_buttons; j++) {

				_ts5_status.timer_status.cedrusbox[i].button_press_defined[j] = 0;
				_ts5_status.timer_status.cedrusbox[i].button_press_active[j] = 0;
				_ts5_status.timer_status.cedrusbox[i].button_release_defined[j] = 0;
				_ts5_status.timer_status.cedrusbox[i].button_release_active[j] = 0;
			}
		}
		_ts5_status.timer_status.cedrusbox_is_response_device = 0;
	}

	_ts5_status.timer_status.num_defined_buttons=0;
	_ts5_status.timer_status.num_active_buttons=0;
}


//////////////////////////////////////////////////////////////////////////////
/// Flush the response queue.
//////////////////////////////////////////////////////////////////////////////
void ts5_flush_responses()
{
	ts5_check_timer("ts5_flush_responses");
	ts5_log(TS5_LOGLEVEL_5, "ts5_flush_responses()\n");

	al_flush_event_queue(_ts5_status.timer_status.response_queue);
}


//////////////////////////////////////////////////////////////////////////////
/// Check for a reponse.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error (works only for cedrusboxes).
///
/// \return The key that was pressed (0 if no key was pressed within the deadline).
///
/// This function checks for a response and returns immediately.
//////////////////////////////////////////////////////////////////////////////
int ts5_check_response(double *resptime, double *timing_error)
{
	ts5_check_timer("ts5_check_response");
	ts5_check_timer2("ts5_check_response");
	ts5_log(TS5_LOGLEVEL_6, "ts5_check_response(%p,%p)\n", resptime, timing_error);

	int resp=0;
	ALLEGRO_EVENT event;

	if (!al_get_next_event(_ts5_status.timer_status.response_queue, &event)) {

		if (resptime!=NULL) {
			*resptime = al_get_time();
		}
		if (timing_error!=NULL) {
			*timing_error = 0.0;
		}
	}
	else {

		if (resptime!=NULL) {
			*resptime = event.any.timestamp;
		}

		if (timing_error!=NULL) {
			*timing_error = 0.0;
		}

		int i;

		// mouse (driver response codes are counted from 1)
		if (_ts5_status.timer_status.mouse_is_response_device) {

			if (_ts5_status.timer_status.mouse.num_active_buttons) {

				if (event.type==ALLEGRO_EVENT_MOUSE_BUTTON_DOWN) {

					resp = _ts5_status.timer_status.mouse.button_press_active[event.mouse.button-1];
				}

				if (event.type==ALLEGRO_EVENT_MOUSE_BUTTON_UP) {
					resp = _ts5_status.timer_status.mouse.button_release_active[event.mouse.button-1];
				}
			}
		}

		// keyboard (driver response codes are counted from 1)
		if (_ts5_status.timer_status.keyboard_is_response_device) {

			if (_ts5_status.timer_status.keyboard.num_active_buttons) {

				if (event.type==ALLEGRO_EVENT_KEY_DOWN) {
					resp = _ts5_status.timer_status.keyboard.
								button_press_active[event.keyboard.keycode-1];
				}

				if (event.type==ALLEGRO_EVENT_KEY_UP) {
					resp = _ts5_status.timer_status.keyboard.
								button_release_active[event.keyboard.keycode-1];
				}
			}
		}

		// joystick (driver response codes are counted from 0)
		if (_ts5_status.timer_status.joystick_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.num_joystick; i++) {

				if (_ts5_status.timer_status.joystick[i].num_active_buttons) {

					if (event.type==ALLEGRO_EVENT_JOYSTICK_BUTTON_DOWN) {

						ALLEGRO_JOYSTICK *stick = al_get_joystick(i);

						if (event.joystick.id == stick) {
							resp = _ts5_status.timer_status.joystick[i].
										button_press_active[event.joystick.button];
						}
					}

					if (event.type==ALLEGRO_EVENT_JOYSTICK_BUTTON_UP) {

						ALLEGRO_JOYSTICK *stick = al_get_joystick(i);

						if (event.joystick.id == stick) {
							resp = _ts5_status.timer_status.joystick[i].
										button_release_active[event.joystick.button];
						}
					}
				}
			}
		}

		// cedrusbox (driver response codes are counted from 0)
		if (_ts5_status.timer_status.cedrusbox_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.num_cedrusbox; i++) {

				if (_ts5_status.timer_status.cedrusbox[i].num_active_buttons) {

					if (event.type==TS5_EVENT_CEDRUS_BUTTON_DOWN) {

						if (event.user.data1 == i) {
							resp = _ts5_status.timer_status.cedrusbox[i].
										button_press_active[event.user.data2];
						}
					}

					if (event.type==TS5_EVENT_CEDRUS_BUTTON_UP) {

						if (event.user.data1 == i) {
							resp = _ts5_status.timer_status.cedrusbox[i].
										button_release_active[event.user.data2];
						}
					}

					if (resp && timing_error!=NULL) {
						*timing_error = ((double)event.user.data3)/1000000.0;
					}
				}
			}
		}

		// parallel port (driver response codes are counted from 1)
		if (_ts5_status.timer_status.parport_is_response_device) {

			for (i=0; i<_ts5_status.timer_status.num_parport; i++) {

				if (_ts5_status.timer_status.parport[i].num_active_buttons) {

					if (event.type==TS5_EVENT_PARPORT_BUTTON_DOWN) {

						if (event.user.data1 == i) {
							resp = _ts5_status.timer_status.parport[i].
										button_press_active[event.user.data2-1];
						}
					}

					if (event.type==TS5_EVENT_PARPORT_BUTTON_UP) {

						if (event.user.data1 == i) {
							resp = _ts5_status.timer_status.parport[i].
										button_release_active[event.user.data2-1];
						}
					}

					if (resp && timing_error!=NULL) {
						*timing_error = ((double)event.user.data3)/1000000.0;
					}
				}
			}
		}
	}

	return resp;
}


//////////////////////////////////////////////////////////////////////////////
/// Wait for a response.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error (works only for cedrusboxes).
///
/// \return The key that was pressed.
//////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_response(double *resptime, double *timing_error)
{
	ts5_check_timer("ts5_wait_for_response");
	ts5_check_timer2("ts5_wait_for_response");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_response(%p,%p)\n", resptime, timing_error);

	int resp=0;

	while (!resp) {
		resp = ts5_check_response(resptime, timing_error);
	}

	return resp;
}


//////////////////////////////////////////////////////////////////////////////
/// Wait for a response for a given time.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error (works only for cedrusboxes).
/// \param maxtime Maximum time for a response.
///
/// \return The key that was pressed (0 if no key was pressed within the maximum time).
//////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_response_timed(double *resptime, double *timing_error, double maxtime)
{
	ts5_check_timer("ts5_wait_for_response_timed");
	ts5_check_timer2("ts5_wait_for_response_timed");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_response_timed(%p,%p,%f)\n", resptime, timing_error, maxtime);

	if (maxtime<0) {
		ts5_fatal("ts5_wait_for_response_timed: maxtime is negative (%f)\n", maxtime);
	}

	int resp=0, deadline=0;
	double starttime = al_get_time();

	while (!resp && !deadline) {

		resp = ts5_check_response(resptime, timing_error);

		if (al_get_time()-starttime>=maxtime) {
			deadline=1;
		}
	}

	return resp;
}


//////////////////////////////////////////////////////////////////////////////
/// Wait for a response until a deadline.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error (works only for cedrusboxes).
/// \param deadline Deadline for a response.
///
/// \return The key that was pressed (0 if no key was pressed within the deadline).
//////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_response_until(double *resptime, double *timing_error, double deadline)
{
	ts5_check_timer("ts5_wait_for_response_until");
	ts5_check_timer2("ts5_wait_for_response_until");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_response_until(%p,%p,%f)\n", resptime, timing_error, deadline);

	if (deadline<al_get_time()) {
		ts5_fatal("ts5_wait_for_response_until: response deadline is in the past (%f)\n",
					deadline-al_get_time());
	}

	int resp=0, maxtime=0;

	while (!resp && !maxtime) {

		resp = ts5_check_response(resptime, timing_error);

		if (al_get_time()>=deadline) {
			maxtime=1;
		}
	}

	return resp;
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// @name Trigger input/output
/// The trigger input functions require you to define trigger devices
/// using ts5_define_parport_trigger_input() or ts5_define_serialport_trigger_input().
///
/// Registration of trigger input is done asynchronously. All triggers received
/// on the parallel and/or serial port(s) will be buffered continuously during the program.
///
/// Waiting for a trigger is similar to waiting for a reponse.
/// Old triggers that were not processed yet can be flushed using ts5_flush_triggers().
/// One can either wait for a trigger indefinitely, for a given amount of time or until 
/// some deadline has passed.
///
/// The value of the trigger is defined by the computer that sends the trigger.
///
/// During the development of the program it is not necessary to connect a computer that sends
/// the triggers. Trigger input can be simulated using ts5_simulate_parport_trigger_input() 
/// or ts5_simulate_serialport_trigger_input().
///
/// At the end of the program all triggers can be written to a logfile using ts5_write_all_triggers().
///
/// Trigger output is simple, call ts5_send_parport_trigger() or ts5_send_serialport_trigger()
/// each time you want to send a trigger.
//@{
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
/// Flush the trigger queue.
//////////////////////////////////////////////////////////////////////////////
void ts5_flush_triggers()
{
	ts5_check_timer("ts5_flush_triggers");
	ts5_log(TS5_LOGLEVEL_5, "ts5_flush_triggers()\n");

	al_flush_event_queue(_ts5_status.timer_status.trigger_queue);
}


//////////////////////////////////////////////////////////////////////////////
/// Check for a trigger.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error.
///
/// \return The value of the trigger
///
/// This function checks for a trigger and returns immediately.
//////////////////////////////////////////////////////////////////////////////
int ts5_check_trigger(double *trigtime, double *timing_error)
{
	ts5_check_timer("ts5_check_trigger");
	ts5_check_timer3("ts5_check_trigger");
	ts5_log(TS5_LOGLEVEL_6, "ts5_check_trigger(%p,%p)\n", trigtime, timing_error);

	int trigger=0;
	ALLEGRO_EVENT event;

	if (trigtime!=NULL) {
		*trigtime = al_get_time();
	}
	if (timing_error!=NULL) {
		*timing_error = 0.0;
	}

	if (al_get_next_event(_ts5_status.timer_status.trigger_queue, &event)) {

		if(_ts5_status.timer_status.parport_is_trigger_device) {

			int i;
			for (i=0; i<_ts5_status.timer_status.num_parport; i++) {

				if (_ts5_status.timer_status.parport[i].is_trigger_input_device) {

					if (event.type==TS5_EVENT_PARPORT_TRIGGER && event.user.data1 == i) {

						if (trigtime!=NULL) {
							*trigtime = event.any.timestamp;
						}

						if (timing_error!=NULL) {
							*timing_error = ((double)event.user.data3)/1000000.0;
						}

						trigger = event.user.data2;
					}
				}
			}
		}

		if(_ts5_status.timer_status.serialport_is_trigger_device) {

			int i;
			for (i=0; i<_ts5_status.timer_status.num_serialport; i++) {

				if (_ts5_status.timer_status.serialport[i].is_trigger_input_device) {

					if (event.type==TS5_EVENT_SERIALPORT_TRIGGER && event.user.data1 == i) {

						if (trigtime!=NULL) {
							*trigtime = event.any.timestamp;
						}

						if (timing_error!=NULL) {
							*timing_error = ((double)event.user.data3)/1000000.0;
						}

						trigger = event.user.data2;
					}
				}
			}
		}
	}

	return trigger;
}


//////////////////////////////////////////////////////////////////////////////
/// Wait for a trigger.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error (works only for cedrusboxes).
///
/// \return The key that was pressed.
//////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_trigger(double *trigtime, double *timing_error)
{
	ts5_check_timer("ts5_wait_for_trigger");
	ts5_check_timer2("ts5_wait_for_trigger");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_trigger(%p,%p)\n", trigtime, timing_error);

	int trigger=0;

	while (!trigger) {
		trigger = ts5_check_trigger(trigtime, timing_error);
	}

	return trigger;
}


//////////////////////////////////////////////////////////////////////////////
/// Wait for a trigger for a given time.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error (works only for cedrusboxes).
/// \param maxtime Maximum time for a trigger.
///
/// \return The key that was pressed (0 if no key was pressed within the maximum time).
//////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_trigger_timed(double *trigtime, double *timing_error, double maxtime)
{
	ts5_check_timer("ts5_wait_for_trigger_timed");
	ts5_check_timer2("ts5_wait_for_trigger_timed");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_trigger_timed(%p,%p,%f)\n", trigtime, timing_error, maxtime);

	if (maxtime<0) {
		ts5_fatal("ts5_wait_for_trigger_timed: maxtime is negative (%f)\n", maxtime);
	}

	int trigger=0, deadline=0;
	double starttime = al_get_time();

	while (!trigger && !deadline) {

		trigger = ts5_check_trigger(trigtime, timing_error);

		if (al_get_time()-starttime>=maxtime) {
			deadline=1;
		}
	}

	return trigger;
}


//////////////////////////////////////////////////////////////////////////////
/// Wait for a trigger until a deadline.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error (works only for cedrusboxes).
/// \param deadline Deadline for a trigger.
///
/// \return The key that was pressed (0 if no key was pressed within the deadline).
///
//////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_trigger_until(double *trigtime, double *timing_error, double deadline)
{
	ts5_check_timer("ts5_wait_for_trigger_until");
	ts5_check_timer2("ts5_wait_for_trigger_until");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_trigger_until(%p,%p,%f)\n", trigtime, timing_error, deadline);

	if (deadline<al_get_time()) {
		ts5_fatal("ts5_wait_for_trigger_until: trigger deadline is in the past (%f)\n",
					deadline-al_get_time());
	}

	int trigger=0, maxtime=0;

	while (!trigger && !maxtime) {

		trigger = ts5_check_trigger(trigtime, timing_error);

		if (al_get_time()>=deadline) {
			maxtime=1;
		}
	}

	return trigger;
}


//////////////////////////////////////////////////////////////////////////////
/// Write all triggers to an output file
///
/// \param filename Path to the output file
//////////////////////////////////////////////////////////////////////////////
void ts5_write_all_triggers(char *filename)
{
	ts5_check_timer("ts5_write_all_triggers");
	ts5_check_timer3("ts5_write_all_triggers");
	ts5_log(TS5_LOGLEVEL_5, "ts5_write_all_triggers(%s)\n", filename);

	FILE *fp;
	fp = fopen(filename, "a+");
	if (!fp) {
		ts5_fatal("ts5_write_all_triggers: could not open output file %s\n", filename);
	}

	ALLEGRO_EVENT event;

	while (al_get_next_event(_ts5_status.timer_status.trigger_log, &event)) {

		fprintf(fp, "%10.8f ", event.any.timestamp);

		double timing_error = ((double)event.user.data3)/1000000.0;
		fprintf(fp, "%10.8f ", timing_error);

		if (event.type==TS5_EVENT_PARPORT_TRIGGER) {
			fprintf(fp, "PARPORT    ");
		}
		else if (event.type==TS5_EVENT_SERIALPORT_TRIGGER) {
			fprintf(fp, "SERIALPORT ");
		}
		else {
			fprintf(fp, "UNKNOWN    ");
		}

		fprintf(fp, "%d ", (int)event.user.data1+1);
		fprintf(fp, "%d ", (int)event.user.data2);
		fprintf(fp, "\n");
	}

	fclose(fp);
}


//////////////////////////////////////////////////////////////////////////////
//@}
//////////////////////////////////////////////////////////////////////////////




