//////////////////////////////////////////////////////////////////////////////
/// \file timer_internal.c
/// Definitions of internal timer functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/timer_internal.h"
#include "../include/tscope5/system_internal.h"

/// Is the timer subsystem installed?
int  _ts5_is_timer_installed = 0;


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each timer function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether the timer is installed. If not, the timer is installed.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_timer(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_timer\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each timer function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether there are active response keys. Aborts if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_timer2(char *calling_function)
{
	if (_ts5_status.timer_status.num_active_buttons<1) {
		ts5_fatal("%s: there are no active response buttons\n", calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_timer2\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each timer function.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// Checks whether there are active trigger devices. Aborts if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_check_timer3(char *calling_function)
{
	int found=0;

	if (_ts5_status.timer_status.parport_is_trigger_device) {

		int i;
		for (i=0; i<_ts5_status.timer_status.num_parport; i++) {

			if (_ts5_status.timer_status.parport[i].is_trigger_input_device) {
				found=1;
			}
		}
	}

	if (_ts5_status.timer_status.serialport_is_trigger_device) {

		int i;
		for (i=0; i<_ts5_status.timer_status.num_serialport; i++) {

			if (_ts5_status.timer_status.serialport[i].is_trigger_input_device) {
				found=1;
			}
		}
	}

	if (!found) {
		ts5_fatal("%s: there are no trigger input devices\n", calling_function);
	}

	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_timer3\n", calling_function);
}


//////////////////////////////////////////////////////////////////////////////
/// Install the timer subsystem.
///
/// \param calling_function Name the function that calls for this check or installation.
///
/// This function is called automatically if necessary.
//////////////////////////////////////////////////////////////////////////////
void ts5_install_timer(char *calling_function)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5(calling_function);
	}

	if (!_ts5_is_timer_installed) {

		_ts5_is_timer_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 timer\n", calling_function);

		_ts5_status.timer_status.response_queue = al_create_event_queue();

		if (_ts5_status.timer_status.response_queue==NULL) {
			ts5_fatal("%s: could not install timer\n", calling_function);
		}

		_ts5_status.timer_status.trigger_queue = al_create_event_queue();

		if (_ts5_status.timer_status.trigger_queue==NULL) {
			ts5_fatal("%s: could not install timer\n", calling_function);
		}

		_ts5_status.timer_status.trigger_log = al_create_event_queue();

		if (_ts5_status.timer_status.trigger_log==NULL) {
			ts5_fatal("%s: could not install timer\n", calling_function);
		}

		atexit(ts5_uninstall_timer);
	}
}


//////////////////////////////////////////////////////////////////////////////
/// Uninstall the timer subsystem.
///
/// This function is called automatically at the end of the program.
//////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_timer()
{
	if (_ts5_is_timer_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 timer\n");

		al_flush_event_queue(_ts5_status.timer_status.response_queue);
		al_flush_event_queue(_ts5_status.timer_status.trigger_queue);
		al_flush_event_queue(_ts5_status.timer_status.trigger_log);

		al_destroy_event_queue(_ts5_status.timer_status.response_queue);
		al_destroy_event_queue(_ts5_status.timer_status.trigger_queue);
		al_destroy_event_queue(_ts5_status.timer_status.trigger_log);

		_ts5_is_timer_installed = 0;
	}
}

