//////////////////////////////////////////////////////////////////////////////
/// \file bitmaps_unittest.c
/// Unit tests of bitmap functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope/bitmaps.h"

#include "../include/tscope/display.h"
#include "../include/tscope/graphics.h"
#include "../include/tscope/textio.h"
#include "../include/tscope/mouse.h"
#include "../include/tscope/primitives.h"

//////////////////////////////////////////////////////////////////////////////
/// Bitmaps unit tests.
/// Example of:
/// - reading a bitmap from a file
/// - getting bitmap dimensions
/// - regular draw
/// - flipped draw
/// - tinted draw
/// - region draw
/// - rotated draw
/// - scaled draw
/// - freeing bitmap memory
/// - generate a memory bitmap
/// - set a bitmap as the drawing target
/// - save a bitmap to a file
/// - todo: do some graphics without opening the display
//////////////////////////////////////////////////////////////////////////////

double dw, dh, bw, bh;
TS5_BITMAP *test;

void wwait()
{
	int oldsystem = ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	int oldscale = ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_flip_display();
	ts5_draw_mouse_button(dw/2, dh/2);
	ts5_clear_display();
	ts5_flip_display();
	ts5_set_coordinate_system(oldsystem);
	ts5_set_coordinate_scale(oldscale);
}


void draw_example()
{
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(dw/2, 20.0, "regular draw");
	ts5_printf(dw/2, 40.0, "(notice the transparent shade)");
	ts5_draw_bitmap(test, 0.0, 0.0);
	ts5_draw_bitmap(test, 30.0, 30.0);
	ts5_draw_bitmap(test, 60.0, 60.0);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(0.0, dh/2-20, "regular draw");
	ts5_printf(0.0, dh/2-40, "(notice the transparent shade)");
	ts5_draw_bitmap(test, -30.0, -30.0);
	ts5_draw_bitmap(test, 0.0, 0.0);
	ts5_draw_bitmap(test, 30.0, 30.0);
	wwait();


	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.5, 0.05, "regular draw");
	ts5_printf(0.5, 0.10, "(notice the transparent shade)");
	ts5_draw_bitmap(test, 0.0, 0.0);
	ts5_draw_bitmap(test, 0.2, 0.2);
	ts5_draw_bitmap(test, 0.4, 0.4);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.0, 0.90, "regular draw");
	ts5_printf(0.0, 0.80, "(notice the transparent shade)");
	ts5_draw_bitmap(test, -0.1, -0.1);
	ts5_draw_bitmap(test, 0.0, 0.0);
	ts5_draw_bitmap(test, 0.1, 0.1);
	wwait();
}


void draw_flipped_example()
{
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(dw/2, 20.0, "flipped draw");
	ts5_draw_flipped_bitmap(test, 0.0, 0.0, TS5_FLIP_VERTICAL);
	ts5_draw_flipped_bitmap(test, 0.0, bh, TS5_FLIP_HORIZONTAL);
	ts5_draw_flipped_bitmap(test, 0.0, bh*2, TS5_FLIP_BOTH);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(0.0, dh/2-20, "flipped draw");
	ts5_draw_flipped_bitmap(test, 0.0, bh, TS5_FLIP_VERTICAL);
	ts5_draw_flipped_bitmap(test, 0.0, 0.0, TS5_FLIP_HORIZONTAL);
	ts5_draw_flipped_bitmap(test, 0.0, -bh, TS5_FLIP_BOTH);
	wwait();

	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.5, 0.05, "flipped draw");
	ts5_draw_flipped_bitmap(test, 0.0, 0.0, TS5_FLIP_VERTICAL);
	ts5_draw_flipped_bitmap(test, 0.0, 0.3, TS5_FLIP_HORIZONTAL);
	ts5_draw_flipped_bitmap(test, 0.0, 0.6, TS5_FLIP_BOTH);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.0, 0.90, "flipped draw");
	ts5_draw_flipped_bitmap(test, 0.0, -0.4, TS5_FLIP_VERTICAL);
	ts5_draw_flipped_bitmap(test, 0.0, 0.0, TS5_FLIP_HORIZONTAL);
	ts5_draw_flipped_bitmap(test, 0.0, 0.4, TS5_FLIP_BOTH);
	wwait();
}


void draw_tinted_example()
{
	TS5_COLOR tint = ts5_make_rgb_color(0.0, 0.5, 0.0, 0.5);

	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(dw/2, 20.0, "tinted draw");
	ts5_draw_tinted_bitmap(test, tint, 0.0, 0.0);
	ts5_draw_tinted_bitmap(test, tint, 0.0, dh/4.0);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(0.0, dh/2-20, "tinted draw");
	ts5_draw_tinted_bitmap(test, tint, 0.0, dh/8.0);
	ts5_draw_tinted_bitmap(test, tint, 0.0, -dh/8.0);
	wwait();

	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.5, 0.05, "tinted draw");
	ts5_draw_tinted_bitmap(test, tint, 0.0, 0.0);
	ts5_draw_tinted_bitmap(test, tint, 0.0, 0.5);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.0, 0.90, "tinted draw");
	ts5_draw_tinted_bitmap(test, tint, 0.0, 0.25);
	ts5_draw_tinted_bitmap(test, tint, 0.0, -0.25);
	wwait();
}


void draw_region_example()
{
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(dw/2, 20.0, "draw region");
	ts5_draw_bitmap_region(test, 20.0, 20.0, dw-20, dh-20, dw/2, dh/2);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(0.0, dh/2-20, "draw region");
	ts5_draw_bitmap_region(test, -bw / 2 + 20, -bh / 2 + 20, bw / 2 - 20, bh / 2 - 20, 0.0, 0.0);
	wwait();

	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.5, 0.05, "draw region");
	ts5_draw_bitmap_region(test, 0.1, 0.1, 0.9, 0.9, 0.5, 0.5);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.0, 0.90, "draw region");
	ts5_draw_bitmap_region(test, -0.9, -0.9, 0.9, 0.9, 0.0, 0.0);
	wwait();
}


void draw_rotated_example()
{
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(dw/2, 20.0, "rotated draw");
	ts5_draw_rotated_bitmap(test, bw/2, bh/2, -45.0, dw/2, dh/2);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(0.0, dh/2-20, "rotated draw");
	ts5_draw_rotated_bitmap(test, 0.0, 0.0, -45.0, 0.0, 0.0);
	wwait();

	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.5, 0.05, "rotated draw");
	ts5_draw_rotated_bitmap(test, 0.5, 0.5, -45.0, 0.5, 0.5);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.0, 0.90, "rotated draw");
	ts5_draw_rotated_bitmap(test, 0.1, 0.1, -45.0, -0.1, -0.1);
	wwait();
}


void draw_scaled_example()
{
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(dw/2, 20.0, "scaled draw");
	ts5_draw_scaled_bitmap(test, 0.5, 2.0, 0.0, 0.0);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_printf(0.0, dh/2-20, "scaled draw");
	ts5_draw_scaled_bitmap(test, 0.5, 2.0, 0.0, 0.0);
	wwait();

	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.5, 0.05, "scaled draw");
	ts5_draw_scaled_bitmap(test, 0.5, 2.0, 0.0, 0.0);
	wwait();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_RELATIVE_COORDINATES);
	ts5_printf(0.0, 0.90, "scaled draw");
	ts5_draw_scaled_bitmap(test, 0.5, 2.0, 0.0, 0.0);
	wwait();
}


int main(void)
{
	ts5_set_display_size(600.0, 400.0);
	ts5_set_background_color(ts5_make_rgb_color(1.0, 1.0, 1.0, 1.0));
	ts5_set_foreground_color(ts5_make_rgb_color(0.0, 0.0, 0.0, 1.0));
	ts5_install_display();
	ts5_get_target_size(&dw, &dh);

	test = ts5_read_bitmap("green.png");
	ts5_get_bitmap_size(test, &bw, &bh);

	draw_example();
	draw_flipped_example();
	draw_tinted_example();
	draw_region_example();
	draw_rotated_example();
	draw_scaled_example();

	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);


	ts5_clear_display();
	TS5_BITMAP *test2, *scr;
	test2 = ts5_alloc_bitmap(300.0, 200.0);
	scr = ts5_set_drawing_target(test2);
	ts5_set_fill_mode(TS5_FILL_ON);
	ts5_draw_circle(0.0, 0.0, 100.0);

	ts5_set_drawing_target(scr);
	ts5_printf(0.0, dh/2 - 20, "draw a user generated bitmap");
	ts5_draw_bitmap(test2, -dw / 2, 0.0);
	ts5_draw_bitmap(test2, dw / 2, 0.0);
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(-0.9));


	ts5_write_bitmap("bitmapstest.tga", scr);

	ts5_free_bitmap(test);
	ts5_free_bitmap(test2);

	return 0;
}
