//////////////////////////////////////////////////////////////////////////////
/// \file display_unittest.c
/// Unit tests of display functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope/display.h"

#include "../include/tscope/mouse.h"
#include "../include/tscope/graphics.h"

//////////////////////////////////////////////////////////////////////////////
/// Display functions unit tests.
///
/// Example of:
/// - querying the display adapter
/// - setting display parameters
/// - querying display parameters for next display
/// - opening 1 or 2 displays
/// - querying display parameters for active display
//////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
	ts5_set_log_level(TS5_LOGLEVEL_5);

	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);

	if (argc < 2) {
		ts5_fatal("Usage: %s 0:2 (0=windowed, 1=fullscreen, 2=fullscreen_window)\n", argv[0]);
	}

	// query the display adapters
	int nad = ts5_get_num_display_adapters();
	ts5_log(TS5_LOGLEVEL_0, "number of display adapters: %d\n", nad);
	int i;
	double dw, dh;
	double w, h, x, y;
	for (i=0; i<nad; i++) {
		ts5_get_display_adapter_size (i+1, &dw, &dh);
		ts5_log(TS5_LOGLEVEL_0, "adapter %d size: %f x %f\n", i+1, dw, dh);
	}
	ts5_log(TS5_LOGLEVEL_0, "\n");

	// set the display parameters for the next display (functions return the previous setting)
	int old, new;
	new = 1;
	old = ts5_set_display_adapter(new);
	ts5_log(TS5_LOGLEVEL_0, "set display adapter for next display from %d to %d\n", old, new);
	ts5_set_display_size(400.0, 400.0);
	ts5_log(TS5_LOGLEVEL_0, "set display size for next display to 400 x 400\n");
	ts5_set_display_position(0.0, 0.0);
	ts5_log(TS5_LOGLEVEL_0, "set display position for next display to 0 x 0\n");
	new = 60;
	old = ts5_set_refreshrate(new);
	ts5_log(TS5_LOGLEVEL_0, "set refreshrate for next display from %d to %d\n", old, new);
	new = TS5_WINDOWED;
	old = ts5_set_display_mode(new);
	ts5_log(TS5_LOGLEVEL_0, "set display mode for next display from %d to %d\n", old, new);
	new = TS5_VSYNC_ON;
	old = ts5_set_vsync_mode(new);
	ts5_log(TS5_LOGLEVEL_0, "set vsync mode for next display from %d to %d\n", old, new);
	ts5_log(TS5_LOGLEVEL_0, "\n");

	// query the display parameters for the next display
	ts5_log(TS5_LOGLEVEL_0, "display adapter for next display: %d\n", ts5_get_display_adapter(0));
	ts5_get_display_size(0, &w, &h);
	ts5_log(TS5_LOGLEVEL_0, "display size for next display: %f x %f\n", w, h);
	ts5_get_display_position(0, &x, &y);
	ts5_log(TS5_LOGLEVEL_0, "display position for next display: %f x %f\n", x, y);
	ts5_log(TS5_LOGLEVEL_0, "refreshrate for next display: %d\n", ts5_get_refreshrate(0));
	ts5_log(TS5_LOGLEVEL_0, "display mode for next display: %d\n", ts5_get_display_mode(0));
	ts5_log(TS5_LOGLEVEL_0, "vsync mode for next display: %d\n", ts5_get_vsync_mode(0));

	int type = atoi(argv[1]);
	if (type==0) {

		ts5_get_display_adapter_size(ts5_get_display_adapter(0), &dw, &dh);

		ts5_set_display_mode(TS5_WINDOWED);

		ts5_set_display_size(dw/4.0, dh/4.0);

		ts5_set_display_position(dw/4.0, 0.0);
		ts5_install_display();

		ts5_set_display_position(-dw/4.0, 0.0);
		ts5_install_display();


	}
	else if (type==1) {
		ts5_set_display_mode(TS5_FULLSCREEN);
		ts5_get_display_adapter_size(ts5_get_display_adapter(0), &dw, &dh);
		ts5_set_display_size(dw, dh);
		ts5_install_display();
	}
	else if (type==2) {
		ts5_set_display_mode(TS5_FULLSCREEN_WINDOW);
		ts5_install_display();
	}
	else {
		ts5_fatal("Usage: %s 0:2 (0=windowed, 1=fullscreen, 2=fullscreen_window)\n", argv[0]);
	}

	// query the display parameters for the active display
	int dpl = ts5_get_active_display();
	ts5_log(TS5_LOGLEVEL_0, "active display: %d\n", dpl);
	ts5_log(TS5_LOGLEVEL_0, "display adapter for active display: %d\n", ts5_get_display_adapter(dpl));
	ts5_get_display_size(dpl, &w, &h);
	ts5_log(TS5_LOGLEVEL_0, "display size for active display: %f x %f\n", w, h);
	ts5_get_display_position(dpl, &x, &y);
	ts5_log(TS5_LOGLEVEL_0, "display position for active display: %f x %f\n", x, y);
	ts5_log(TS5_LOGLEVEL_0, "refreshrate for active display: %d\n", ts5_get_refreshrate(dpl));
	ts5_log(TS5_LOGLEVEL_0, "display mode for active display: %d\n", ts5_get_display_mode(dpl));
	ts5_log(TS5_LOGLEVEL_0, "vsync mode for active display: %d\n", ts5_get_vsync_mode(dpl));

	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.8), ts5_relative_to_absolute_coordinate_y(-0.8));

	return 0;
}




