//////////////////////////////////////////////////////////////////////////////
/// \file graphics_unittest.c
/// Unit tests of graphics parameter functions
//////////////////////////////////////////////////////////////////////////////

#include "../include/tscope/graphics.h"

#include "../include/tscope/display.h"
#include "../include/tscope/primitives.h"
#include "../include/tscope/mouse.h"
#include "../include/tscope/textio.h"

//////////////////////////////////////////////////////////////////////////////
/// Graphics functions unit tests.
///
/// Example of:
/// - coordinate systems
/// - color definitions
/// - graphics parameters
/// - fonts and text alignment
//////////////////////////////////////////////////////////////////////////////
int main(void)
{
	ts5_set_coordinate_scale(TS5_ABSOLUTE_COORDINATES);
	ts5_set_display_size(600.0, 480.0);
	ts5_install_display();

	double xmax, ymax;
	ts5_get_target_size(&xmax, &ymax);

	// TS5_CARTESIAN_COORDINATES system
	ts5_draw_line(0.0, 0.0, 100.0, 100.0);
	ts5_draw_line(0.0, 0.0, xmax/2.0, ymax/4.0);
	ts5_draw_line(0.0, 0.0, ts5_relative_to_absolute_coordinate_x(0.5), ts5_relative_to_absolute_coordinate_y(0.5));
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(-0.9));

	// TS5_DISPLAY_COORDINATES system
	ts5_clear_display();
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_draw_line(0.0, 0.0, 100.0, 100.0);
	ts5_draw_line(0.0, 0.0, xmax/2.0, ymax/4.0);
	ts5_draw_line(0.0, 0.0, ts5_relative_to_absolute_coordinate_x(0.5), ts5_relative_to_absolute_coordinate_y(0.5));
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(0.9));

	// RGB color definitions
	ts5_clear_display();
	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_fill_mode(TS5_FILL_ON);
	ts5_set_foreground_color(ts5_make_rgb_color(1.0, 0.0, 0.0, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(-0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_rgb_color(0.0, 1.0, 0.0, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.0), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_rgb_color(1.0, 0.0, 0.0, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(-0.9));

	// HSL color definitions
	ts5_clear_display();
	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_fill_mode(TS5_FILL_ON);
	ts5_set_foreground_color(ts5_make_hsl_color(0.0, 0.5, 0.5, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(-0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_hsl_color(180.0, 0.5, 0.5, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.0), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_hsl_color(360.0, 0.5, 0.5, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(-0.9));

	// HSV color definitions
	ts5_clear_display();
	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_fill_mode(TS5_FILL_ON);
	ts5_set_foreground_color(ts5_make_hsv_color(0.0, 0.5, 0.5, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(-0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_hsv_color(180.0, 0.5, 0.5, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.0), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_hsv_color(360.0, 0.5, 0.5, 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(-0.9));

	// named color definitions
	ts5_clear_display();
	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_fill_mode(TS5_FILL_ON);
	ts5_set_foreground_color(ts5_make_named_color("tomato", 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(-0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_named_color("olive", 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.0), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_foreground_color(ts5_make_named_color("tomato", 0.5));
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(-0.9));

	// graphics parameters
	ts5_clear_display();
	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_fill_mode(TS5_FILL_ON);
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(-0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_fill_mode(TS5_FILL_OFF);
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.0), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_set_drawing_thickness(5.0);
	ts5_draw_circle(ts5_relative_to_absolute_coordinate_x(0.25), 0.0, ts5_relative_to_absolute_coordinate_x(0.25));
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(-0.9));

	// font types
	ts5_clear_display();
	ts5_set_foreground_color(ts5_make_named_color("tomato", 1.0));
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_text_alignment(TS5_ALIGN_LEFT);
	ts5_printf(10.0, 20.0, "courier");
	int oldtype = ts5_set_font_type(TS5_ARIAL);
	ts5_printf(10.0, 40.0, "arial");
	ts5_set_font_type(TS5_TIMES);
	ts5_printf(10.0, 60.0, "times");
	ts5_set_font_type(oldtype);
	ts5_printf(10.0, 80.0, "default");
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(0.9));

	// font styles
	ts5_clear_display();
	ts5_set_foreground_color(ts5_make_named_color("tomato", 1.0));
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_text_alignment(TS5_ALIGN_LEFT);
	ts5_printf(10.0, 20.0, "bold");
	int oldstyle = ts5_set_font_style(TS5_ITALIC);
	ts5_printf(10.0, 40.0, "italic");
	ts5_set_font_style(TS5_BOLD_ITALIC);
	ts5_printf(10.0, 60.0, "bold italic");
	ts5_set_font_style(TS5_REGULAR);
	ts5_printf(10.0, 80.0, "regular");
	ts5_set_font_style(oldstyle);
	ts5_printf(10.0, 100.0, "default");
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(0.9));

	// font sizes
	ts5_clear_display();
	ts5_set_foreground_color(ts5_make_named_color("tomato", 1.0));
	ts5_set_coordinate_system(TS5_DISPLAY_COORDINATES);
	ts5_set_text_alignment(TS5_ALIGN_LEFT);
	ts5_printf(10.0, 20.0, "size 20");
	int oldsize = ts5_set_font_size(30);
	ts5_printf(10.0, 50.0, "size 30");
	ts5_set_font_size(40);
	ts5_printf(10.0, 80.0, "size 40");
	ts5_set_font_size(oldsize);
	ts5_printf(10.0, 110.0, "default size");
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), ts5_relative_to_absolute_coordinate_y(0.9));


	// text alignment
	ts5_clear_display();
	ts5_set_foreground_color(ts5_make_named_color("tomato", 1.0));
	ts5_set_coordinate_system(TS5_CARTESIAN_COORDINATES);
	ts5_set_text_alignment(TS5_ALIGN_CENTER);
	int oldalign = ts5_set_text_alignment(TS5_ALIGN_LEFT);
	ts5_printf(0.0, 30.0, "left aligned text");
	ts5_set_text_alignment(TS5_ALIGN_CENTER);
	ts5_printf(0.0, 10.0, "centered text");
	ts5_set_text_alignment(TS5_ALIGN_RIGHT);
	ts5_printf(0.0, -10.0, "right aligned text");
	ts5_set_text_alignment(oldalign);
	ts5_printf(0.0, -30.0, "default alignment");
	ts5_flip_display();
	ts5_draw_mouse_button(ts5_relative_to_absolute_coordinate_x(0.9), -ts5_relative_to_absolute_coordinate_y(0.9));




	return 0;
}
